/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDevice.h"

/*
	This file controls the creation of a new device in the Palm OS
	Emulator.  In order to add a new device to the set of supported
	devices:

	* Add a new kDeviceFoo value to DeviceType in EmDevice.h

	* Add a "case" statement for that device in GetMenuItemList,
	  GetStartupItemList, and GetPrefItemList.

	* Add information about the new device to the kDeviceInfo array.

	* Define and add a sub-class of EmRegs to describe the hardware
	  registers for the device (e.g., EmRegsEZPalmV).  This sub-class
	  can be a typedef of another sub-class if the hardware is the
	  same (e.g., EmRegsEZPalmVx).

	* Define and add any other EmRegs sub-classes as needed if the
	  device has any special hardware needs (e.g., EmRegsUSBVisor).

	* Add a new "case" statement in CreateRegs that creates the
	  appropriate EmRegs sub-classes.

	* If your needs are extensive and creating a sub-class of EmRegs
	  isn't sufficient, then consider creating a new EmBankFoo and
	  managing that in EmMemory.cpp along with all the other EmBankFoos.

	* Modify SupportsROM to make sure that your device appears on the
	  device list when your ROM is selected, and try to make sure that
	  it does not appear for other ROMs.
*/

#include "EmBankRegs.h"
#include "EmROMReader.h"

#include "PalmPack.h"
#define NON_PORTABLE
	#include "HwrMiscFlags.h"		// hwrMiscFlagIDTouchdown, etc.

	#ifndef hwrMiscFlagIDPalmIIIc
	#define	hwrMiscFlagIDPalmIIIc			(hwrMiscFlagID4 | hwrMiscFlagID1)
	#endif

	// The m100 ID has changed
	#undef hwrMiscFlagIDm100
	#define hwrMiscFlagIDm100	(hwrMiscFlagID3 | hwrMiscFlagID1)

#undef NON_PORTABLE
#include "PalmPackPop.h"

#include "EmRegs328Pilot.h"
#include "EmRegs328PalmPilot.h"
#include "EmRegs328PalmIII.h"
#include "EmRegs328PalmVII.h"
#include "EmRegs328Symbol1700.h"

#include "EmRegsEZm100.h"
#include "EmRegsEZPalmIIIc.h"
#include "EmRegsEZPalmIIIe.h"
#include "EmRegsEZPalmIIIx.h"
#include "EmRegsEZPalmV.h"
#include "EmRegsEZPalmVx.h"
#include "EmRegsEZPalmVII.h"
#include "EmRegsEZPalmVIIx.h"
#include "EmRegsEZVisor.h"
#include "EmRegsEZTemp.h"

#include "EmRegsVZTemp.h"
#include "EmRegsSED1376.h"


#include "EmRegsASICSymbol1700.h"
#include "EmRegsFrameBuffer.h"
#include "EmRegsPLDPalmVIIEZ.h"
#include "EmRegsSED1375.h"
#include "EmRegsUSBVisor.h"

#include "EmTRG.h"


struct DeviceInfo
{
	DeviceType	fDeviceType;
	uint32		fFlags;
	long		fHardwareID;
	long		fHardwareSubID;
};


// (None of these values are (is?) exposed, so they can be freely reordered.)

static const uint32	kSupports68328			= 0x00000001;
static const uint32	kSupports68EZ328		= 0x00000002;
static const uint32	kSupports68VZ328		= 0x00000004;

static const uint32	kHasFlash				= 0x00000010;

static const long	hwrMiscFlagIDNone		= 0xFF;
static const long	hwrMiscFlagExtSubIDNone	= 0xFF;

#define UNSUPPORTED			0


/* Miscellaneous symbols used to discriminate between ROMs. Do not assume that
   these are meant to convey actual ownership or documented interfaces. */
   
#define PALM_MANUF			"Palm Computing"
#define HANDSPRING_MANUF	"Handspring, Inc."
#define SYMBOL_1700_DB		"S24PowerDown"
#define PALM_VII_DB			"WMessaging"
#define PALM_VII_IF1_DB		"RAM NetIF"
#define PALM_VII_IF2_DB		"RAM NetIFRM"
#define PALM_VII_EZ_NEW_DB	"TuneUp"
#define PALM_VIIX_DB		"iMessenger"
#define PALM_m100_DB		"cclkPalmClock"

static const DeviceInfo kDeviceInfo[] =
{
	{
		kDeviceUnspecified,
		UNSUPPORTED
	},
	{
		kDevicePilot1000,
		kSupports68328,
		hwrMiscFlagIDTouchdown, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePilot5000,
		kSupports68328,
		hwrMiscFlagIDTouchdown, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePalmPilotPersonal,
		kSupports68328,
		hwrMiscFlagIDPalmPilot, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePalmPilotProfessional,
		kSupports68328,
		hwrMiscFlagIDPalmPilot, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePalmPilotWithUpgrade,
		UNSUPPORTED
	},
	{
		kDevicePalmIII,
		kSupports68328 + kHasFlash,
		hwrMiscFlagIDRocky, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePalmVII,
		kSupports68328 + kHasFlash,
		hwrMiscFlagIDJerry, hwrMiscFlagExtSubIDNone,
	},
	{
		kDeviceUnused1,
		UNSUPPORTED
	},
	{
		kDeviceQualComm,
		UNSUPPORTED
	},
	{
		kDevicePalmV,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDSumo, hwrMiscFlagExtSubIDSumo,
	},
	{
		kDevicePalmIIIx,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDBrad, hwrMiscFlagExtSubIDBrad,
	},
	{
		kDeviceUnused2,
		UNSUPPORTED
	},
	{
		kDevicePalmIIIc,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDPalmIIIc, hwrMiscFlagExtSubIDNone,
	},
	{
		kDevicePalmVIIEZ,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDJerryEZ, 0,		// sub-id *should* be hwrMiscFlagExtSubIDNone, but
										// a bug in Rangoon (OS 3.5 for Palm VII EZ) has it
										// accidentally looking for and using the sub-id. In
										// order to get the hard keys to work, we need to
										// return zero here.
	},
	{
		kDevicePalmIIIe,
		kSupports68EZ328,
		hwrMiscFlagIDBrad, hwrMiscFlagExtSubIDBrad,
	},
	{
		kDevicePalmVx,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDSumo, hwrMiscFlagExtSubIDCobra,
	},
	{
		kDeviceSymbol1700,
		kSupports68328 + kHasFlash,
		hwrMiscFlagIDRocky, hwrMiscFlagExtSubIDNone,
	},
	{
		kDeviceTRGpro,
		kSupports68EZ328 + kHasFlash,
		hwrOEMCompanyIDTRG, hwrTRGproID,
	},
	{
		kDeviceVisor,
		kSupports68EZ328,
		0, 0,	// Hardware ID for first Visor
	},
	{
		kDevicem100,
		kSupports68EZ328,
		hwrMiscFlagIDm100, hwrMiscFlagExtSubIDBrad,
	},
	{
		kDevicePalmVIIx,
		kSupports68EZ328 + kHasFlash,
		hwrMiscFlagIDJerryEZ, hwrMiscFlagExtSubIDBrad,	// David Mai says that he *guesses* that the sub-id is zero
	},
	{
		kDeviceReserved1,
		UNSUPPORTED
	},
	{
		kDeviceReserved2,
		UNSUPPORTED
	},
	{
		kDeviceReserved3,
		UNSUPPORTED
	},
	{
		kDeviceReserved4,
		UNSUPPORTED
	},
	{
		kDeviceReserved5,
		UNSUPPORTED
	},
	{
		kDeviceVZGeneric,
		kSupports68VZ328,
		hwrMiscFlagIDNone, hwrMiscFlagExtSubIDNone,
	},
	{
		kDeviceReserved7,
		UNSUPPORTED
	},
	{
		kDeviceReserved8,
		UNSUPPORTED
	},
	{
		kDeviceLast,
		UNSUPPORTED
	}
};


/***********************************************************************
 *
 * FUNCTION:    EmDevice::EmDevice
 *
 * DESCRIPTION: Constructors
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

EmDevice::EmDevice (void) :
	fDeviceType (kDeviceUnspecified)
{
}


EmDevice::EmDevice (DeviceType device) :
	fDeviceType (device)
{
	// Make sure our device information table is built correctly.

	assert (countof (kDeviceInfo) - 1 == kDeviceLast);

	for (DeviceType dt = kDeviceUnspecified; dt <= kDeviceLast; ++dt)
	{
		assert (kDeviceInfo[dt].fDeviceType == dt);
	}

	assert (Supported());
}


EmDevice::EmDevice (const EmDevice& other) :
	fDeviceType (other.fDeviceType)
{
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::~EmDevice
 *
 * DESCRIPTION: Destructor
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

EmDevice::~EmDevice (void)
{
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::Supported
 *
 * DESCRIPTION: Returns whether or not this EmDevice represents a device
 *				we currently support emulating.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    True if so.
 *
 ***********************************************************************/

Bool EmDevice::Supported (void) const
{
	return fDeviceType <= kDeviceLast && kDeviceInfo[fDeviceType].fFlags != 0;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::Supports68328
 *
 * DESCRIPTION: Returns whether or not this device is based on the
 *				MC68328.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    True if so.
 *
 ***********************************************************************/

Bool EmDevice::Supports68328 (void) const
{
	return (kDeviceInfo[fDeviceType].fFlags & kSupports68328) != 0;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::Supports68EZ328
 *
 * DESCRIPTION: Returns whether or not this device is based on the
 *				MC68EZ328.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    True if so.
 *
 ***********************************************************************/

Bool EmDevice::Supports68EZ328 (void) const
{
	return (kDeviceInfo[fDeviceType].fFlags & kSupports68EZ328) != 0;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::Supports68VZ328
 *
 * DESCRIPTION: Returns whether or not this device is based on the
 *				MC68VZ328.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    True if so.
 *
 ***********************************************************************/

Bool EmDevice::Supports68VZ328 (void) const
{
	return (kDeviceInfo[fDeviceType].fFlags & kSupports68VZ328) != 0;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::SupportsROM
 *
 * DESCRIPTION: Returns whether or not this device can support the
 *				execution of a particular ROM.
 *
 * PARAMETERS:  A reference to an EmROMReader object. Hopefully one
 *				that has successfully acquired the details of a ROM.
 *
 * RETURNED:    True if ROM is supported by device.
 *
 ***********************************************************************/

Bool EmDevice::SupportsROM (const EmROMReader<LAS> & ROM) const
{
	bool is7 = ROM.ContainsDB(PALM_VII_IF1_DB) || ROM.ContainsDB(PALM_VII_IF2_DB);

	/* If the ROM is recent enough, use the embedded hardware ID information
	   to determine whether it will work with this device */

	if (ROM.GetCardVersion() >= 5) {

		switch (fDeviceType)
		{
	
		/* These are the only '328 devices, and are slightly ambiguous: */
		case kDevicePalmIII:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'rcky'))
				return true;
			break;

		case kDevicePalmVII:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'jery') &&
				is7)
				return true;
			break;

		/* The rest are EZ or VZ, and straight-forward */
		case kDevicePalmV:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'sumo'))
				return true;
			break;

		case kDevicePalmIIIx:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'sumo'))
				return true;
			break;

		case kDevicePalmIIIc:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'astn'))
				return true;
			break;

		case kDevicePalmIIIe:
			break;

		case kDevicePalmVIIEZ:
			break;

		case kDevicePalmVx:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'sumo'))
				return true;
			break;

		case kDevicem100:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'clvn'))
				return true;
			break;

		case kDevicePalmVIIx:
			break;

		case kDeviceReserved1:
			break;

		case kDeviceReserved2:
			break;

		case kDeviceReserved3:
			break;

		case kDeviceReserved4:
			break;

		case kDeviceReserved5:
			break;

		case kDeviceVZGeneric:
			if ((ROM.GetCompanyID() == 'palm') &&
				(ROM.GetHalID() == 'ecty'))
				return true;
			break;

		case kDeviceReserved7:
			break;

		case kDeviceReserved8:
			break;

		/* These devices don't have post-OS 3.5 ROMs */

		case kDevicePilot1000:
			break;

		case kDevicePilot5000:
			break;

		case kDevicePalmPilotPersonal:
			break;

		case kDevicePalmPilotProfessional:
			break;

		case kDevicePalmPilotWithUpgrade:
			break;

		case kDeviceUnused1:
			break;

		case kDeviceQualComm:
			break;

		case kDeviceUnused2:
			break;

		case kDeviceUnspecified:
			break;

		case kDeviceSymbol1700:
			break;

		case kDeviceTRGpro:
			break;

		case kDeviceVisor:
			break;

		case kDeviceLast:
			break;

		}
		
		return false;
	}
	
	/* Otherwise, we will use miscellaneous pieces of data to heuristically
	   discriminate between ROMs. Do not assume that any of these tests are meant to
	   convey ownership or documented interfaces. */

	bool is328 = (ROM.GetCardVersion() <= 2) || ROM.GetFlag328();
	bool isColor = ROM.GetSplashChunk() && (ROM.IsBitmapColor(*ROM.GetSplashChunk(), true) == 1);

	if (Supports68328() && !is328)
		return false;
	else if (Supports68EZ328() && !ROM.GetFlagEZ())
		return false;
	else if (Supports68VZ328() && !ROM.GetFlagVZ())
		return false;

	switch (fDeviceType)
	{
		case kDeviceUnspecified:
			return false;
			break;

		case kDevicePilot1000:
			if ((ROM.Version() < 0x020000) &&
				(ROM.GetCardManufacturer() == PALM_MANUF)
				)
				return true;
			break;

		case kDevicePilot5000:
			if ((ROM.Version() < 0x020000) &&
				(ROM.GetCardManufacturer() == PALM_MANUF)
				)
				return true;
			break;

		case kDevicePalmPilotPersonal:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				(ROM.Version() >= 0x020000) &&
				(ROM.Version() < 0x030000)
				)
				return true;
			break;

		case kDevicePalmPilotProfessional:
			break;

		case kDevicePalmPilotWithUpgrade:
			break;

		case kDevicePalmIII:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				(ROM.Version() >= 0x030000) &&
				!ROM.ContainsDB(SYMBOL_1700_DB) &&
				!is7
				)
				return true;
			break;

		case kDevicePalmVII:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				(ROM.Version() >= 0x030000) &&
				!ROM.ContainsDB(SYMBOL_1700_DB) &&
				is7
				)
				return true;
			break;

		case kDeviceUnused1:
			break;

		case kDeviceQualComm:
			break;

		case kDevicePalmV:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				!ROM.ContainsDB(PALM_m100_DB) &&
				!is7 &&
				!isColor
				)
				return true;
			break;

		case kDevicePalmIIIx:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				!ROM.ContainsDB(PALM_m100_DB) &&
				!is7 &&
				!isColor
				)
				return true;
			break;

		case kDeviceUnused2:
			break;

		case kDevicePalmIIIc:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				!ROM.ContainsDB(PALM_m100_DB) &&
				!is7 &&
				isColor
				)
				return true;
			break;

		case kDevicePalmVIIEZ:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				is7 &&
				(!ROM.ContainsDB(PALM_VIIX_DB) ||
				 ROM.ContainsDB(PALM_VII_EZ_NEW_DB)) &&
				!isColor
				)
				return true;
			break;

		case kDevicePalmIIIe:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				!ROM.ContainsDB(PALM_m100_DB) &&
				!is7 &&
				!isColor &&
				(ROM.Version() == 0x030100)
				)
				return true;
			break;

		case kDevicePalmVx:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				!ROM.ContainsDB(PALM_m100_DB) &&
				!is7 &&
				!isColor &&
				(ROM.Version() >= 0x030300)
				)
				return true;
			break;

		case kDeviceSymbol1700:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				ROM.ContainsDB(SYMBOL_1700_DB)
				)
				return true;
			break;

		case kDeviceTRGpro:
			if (ROM.GetCardManufacturer() == TRGPRO_MANUF)
				return true;
			break;

		case kDeviceVisor:
			if (ROM.GetCardManufacturer() == HANDSPRING_MANUF)
				return true;
			break;

		case kDevicem100:
			if (ROM.ContainsDB(PALM_m100_DB))
				return true;
			break;

		case kDevicePalmVIIx:
			if ((ROM.GetCardManufacturer() == PALM_MANUF) &&
				is7 &&
				ROM.ContainsDB(PALM_VIIX_DB) &&
				!ROM.ContainsDB(PALM_VII_EZ_NEW_DB) &&
				!isColor
				)
				return true;
			break;

		default:
			break;

		case kDeviceLast:
			assert (false);
			break;
	}
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::CreateRegs
 *
 * DESCRIPTION: Create the register object that manages the Dragonball
 *				(or derivative) hardware registers.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Pointer to the created object.  Caller is responsible
 *				for destroying the object when it's done with it.
 *
 ***********************************************************************/

void EmDevice::CreateRegs (void) const
{
	switch (fDeviceType)
	{
		case kDeviceUnspecified:
			break;

		case kDevicePilot1000:
			EmBankRegs::AddSubBank (new EmRegs328Pilot);
			break;

		case kDevicePilot5000:
			EmBankRegs::AddSubBank (new EmRegs328Pilot);
			break;

		case kDevicePalmPilotPersonal:
			EmBankRegs::AddSubBank (new EmRegs328PalmPilot);
			break;

		case kDevicePalmPilotProfessional:
			EmBankRegs::AddSubBank (new EmRegs328PalmPilot);
			break;

		case kDevicePalmPilotWithUpgrade:
			break;

		case kDevicePalmIII:
			EmBankRegs::AddSubBank (new EmRegs328PalmIII);
			break;

		case kDevicePalmVII:
			EmBankRegs::AddSubBank (new EmRegs328PalmVII);
			break;

		case kDeviceUnused1:
			break;

		case kDeviceQualComm:
			break;

		case kDevicePalmV:
			EmBankRegs::AddSubBank (new EmRegsEZPalmV);
			break;

		case kDevicePalmIIIx:
			EmBankRegs::AddSubBank (new EmRegsEZPalmIIIx);
			break;

		case kDeviceUnused2:
			break;

		case kDevicePalmIIIc:
			EmBankRegs::AddSubBank (new EmRegsEZPalmIIIc);
			EmBankRegs::AddSubBank (new EmRegsSED1375 ((emuptr) sed1375RegsAddr, sed1375BaseAddress));
			EmBankRegs::AddSubBank (new EmRegsFrameBuffer (sed1375BaseAddress, sed1375VideoMemSize));
			break;

		case kDevicePalmVIIEZ:
			EmBankRegs::AddSubBank (new EmRegsEZPalmVII);
			EmBankRegs::AddSubBank (new EmRegsPLDPalmVIIEZ);
			break;

		case kDevicePalmIIIe:
			EmBankRegs::AddSubBank (new EmRegsEZPalmIIIe);
			break;

		case kDevicePalmVx:
			EmBankRegs::AddSubBank (new EmRegsEZPalmVx);
			break;

		case kDeviceSymbol1700:
			EmBankRegs::AddSubBank (new EmRegs328Symbol1700);
			EmBankRegs::AddSubBank (new EmRegsASICSymbol1700);
			break;

		case kDeviceTRGpro:
			OEMCreateTRGRegObjs(hwrTRGproID);
			break;

		case kDeviceVisor:
			EmBankRegs::AddSubBank (new EmRegsEZVisor);
			EmBankRegs::AddSubBank (new EmRegsUSBVisor);
			break;

		case kDevicem100:
			EmBankRegs::AddSubBank (new EmRegsEZm100);
			break;

		case kDevicePalmVIIx:
			EmBankRegs::AddSubBank (new EmRegsEZPalmVIIx);
			EmBankRegs::AddSubBank (new EmRegsPLDPalmVIIEZ);
			break;

 		case kDeviceReserved1:
			break;

		case kDeviceReserved2:
			break;

		case kDeviceReserved3:
			break;

		case kDeviceReserved4:
			break;

		case kDeviceReserved5:
			break;

		case kDeviceVZGeneric:
			EmBankRegs::AddSubBank (new EmRegsVZGenericColor);
			EmBankRegs::AddSubBank (new EmRegsSED1376 (sed1376RegsAddr, sed1376VideoMemStart));
			EmBankRegs::AddSubBank (new EmRegsFrameBuffer (sed1376VideoMemStart, sed1376VideoMemSize));
			EmBankRegs::AddSubBank (new EmRegsUSBPhilipsPDIUSBD12 (0x10400000));
			break;

		case kDeviceReserved7:
			break;

		case kDeviceReserved8:
			break;

		case kDeviceLast:
			assert (false);
			break;
	}

	EmHAL::EnsureCoverage ();
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::HasFlash
 *
 * DESCRIPTION: Returns whether or not this device uses Flash RAM.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    True if so.
 *
 ***********************************************************************/

Bool EmDevice::HasFlash (void) const
{
	return (kDeviceInfo[fDeviceType].fFlags & kHasFlash) != 0;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::HardwareID
 *
 * DESCRIPTION: Returns the hardware ID of this device.  This ID is the
 *				one that eventually gets put into GHwrMiscGFlags by
 *				HwrIdentifyFeatures.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    The device's hardware ID.
 *
 ***********************************************************************/

long EmDevice::HardwareID (void) const
{
	assert (kDeviceInfo[fDeviceType].fHardwareID != hwrMiscFlagIDNone);

	return kDeviceInfo[fDeviceType].fHardwareID;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::HardwareSubID
 *
 * DESCRIPTION: Returns the hardware ID of this device.  This ID is the
 *				one that eventually gets put into GHwrMiscGFlags by
 *				HwrIdentifyFeatures.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    The device's hardware ID.
 *
 ***********************************************************************/

long EmDevice::HardwareSubID (void) const
{
	assert (kDeviceInfo[fDeviceType].fHardwareSubID != hwrMiscFlagExtSubIDNone);

	return kDeviceInfo[fDeviceType].fHardwareSubID;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::GetMenuItemList
 *
 * DESCRIPTION: Return the list of items that should appear in the
 *				various listboxes and menus that allow the user to
 *				select a device.  The list is returned as a collection
 *				of pairs.  The first element of the pair is the
 *				DeviceType, and the second element of the pair is
 *				the name of the device in text form, suitable for
 *				displaying to the user.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Collection containing the results
 *
 ***********************************************************************/

EmDeviceMenuItemList EmDevice::GetMenuItemList (void)
{
	EmDeviceMenuItemList	result;

	result.push_back (make_pair (kDevicePilot1000,			string ("Pilot")));
	result.push_back (make_pair (kDevicePalmPilotPersonal,	string ("PalmPilot")));
	result.push_back (make_pair (kDevicePalmIII,			string ("Palm III")));
	result.push_back (make_pair (kDevicePalmIIIc,			string ("Palm IIIc")));
	result.push_back (make_pair (kDevicePalmIIIe,			string ("Palm IIIe")));
	result.push_back (make_pair (kDevicePalmIIIx,			string ("Palm IIIx")));
	result.push_back (make_pair (kDevicePalmV,				string ("Palm V")));
	result.push_back (make_pair (kDevicePalmVx,				string ("Palm Vx")));
	result.push_back (make_pair (kDevicePalmVII,			string ("Palm VII")));
	result.push_back (make_pair (kDevicePalmVIIEZ,			string ("Palm VII (EZ)")));
	result.push_back (make_pair (kDevicePalmVIIx,			string ("Palm VIIx")));

	result.push_back (make_pair (kDeviceSymbol1700,			string ("Symbol 1700")));
	result.push_back (make_pair (kDeviceTRGpro,				string ("TRGpro")));
	result.push_back (make_pair (kDeviceVisor,				string ("Visor")));
	result.push_back (make_pair (kDevicem100,				string ("m100")));

	return result;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::GetStartupItemList
 *
 * DESCRIPTION: Return a list of items used to help process the -device
 *				option from command-line options.  Each item in the
 *				list is a DeviceType along with a string that the user
 *				can type in to select that device.
 *
 *				This list is also used to save the device ID to the
 *				session file.  When loading, this list is used to turn
 *				the externalized string into a numeric ID.  When saving,
 *				the first item in the list for a particular ID is used
 *				to externalize the ID.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Collection containing the results
 *
 ***********************************************************************/

EmDeviceStartupItemList EmDevice::GetStartupItemList (void)
{
	EmDeviceStartupItemList	result;

	// Add text entries for all the devices.  Do it in this goofy way
	// so that we can take advantages of compilers that report when
	// not all enumerated types are accounted for in the case statements.

	for (DeviceType device = kDeviceUnspecified; device < kDeviceLast; ++device)
	{
		const char* name = NULL;

		switch (device)
		{
			case kDeviceUnspecified:
				break;

			case kDevicePilot1000:
				result.push_back (make_pair (device, string ("Pilot")));
				result.push_back (make_pair (device, string ("Pilot1000")));
				result.push_back (make_pair (device, string ("Pilot5000")));
				break;

			case kDevicePilot5000:
				break;

			case kDevicePalmPilotPersonal:
				result.push_back (make_pair (device, string ("PalmPilot")));
				result.push_back (make_pair (device, string ("PalmPilotPersonal")));
				result.push_back (make_pair (device, string ("PalmPilotProfessional")));
				break;

			case kDevicePalmPilotProfessional:
				break;

			case kDevicePalmPilotWithUpgrade:
				break;

			case kDevicePalmIII:
				result.push_back (make_pair (device, string ("PalmIII")));
				break;

			case kDevicePalmVII:
				result.push_back (make_pair (device, string ("PalmVII")));
				break;

			case kDeviceUnused1:
				break;

			case kDeviceQualComm:
				break;

			case kDevicePalmV:
				result.push_back (make_pair (device, string ("PalmV")));
				break;

			case kDevicePalmIIIx:
				result.push_back (make_pair (device, string ("PalmIIIx")));
				break;

			case kDeviceUnused2:
				break;

			case kDevicePalmIIIc:
				result.push_back (make_pair (device, string ("PalmIIIc")));
				break;

			case kDevicePalmVIIEZ:
				result.push_back (make_pair (device, string ("PalmVIIEZ")));
				break;

			case kDevicePalmIIIe:
				result.push_back (make_pair (device, string ("PalmIIIe")));
				break;

			case kDevicePalmVx:
				result.push_back (make_pair (device, string ("PalmVx")));
				break;

			case kDeviceSymbol1700:
				result.push_back (make_pair (device, string ("Symbol1700")));
				break;

			case kDeviceTRGpro:
				result.push_back (make_pair (device, string ("TRGpro")));
				break;

			case kDeviceVisor:
				result.push_back (make_pair (device, string ("Visor")));
				break;

			case kDevicem100:
				result.push_back (make_pair (device, string ("m100")));
				break;

			case kDevicePalmVIIx:
				result.push_back (make_pair (device, string ("PalmVIIx")));
				break;

			case kDeviceReserved1:
				break;

			case kDeviceReserved2:
				break;

			case kDeviceReserved3:
				break;

			case kDeviceReserved4:
				break;

			case kDeviceReserved5:
				break;

			case kDeviceVZGeneric:
				result.push_back (make_pair (device, string ("VZDevice")));
				break;

			case kDeviceReserved7:
				break;

			case kDeviceReserved8:
				break;

			case kDeviceLast:
				assert (false);
				break;
		}

		if (name)
			result.push_back (make_pair (device, string (name)));
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:    EmDevice::GetPrefItemList
 *
 * DESCRIPTION: Return a list of items used to generate keys in the
 *				preferences file.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Collection containing the results
 *
 ***********************************************************************/

EmDevicePrefItemList EmDevice::GetPrefItemList (void)
{
	EmDevicePrefItemList	result;

	// Add text entries for all the devices.  Do it in this goofy way
	// so that we can take advantages of compilers that report when
	// not all enumerated types are accounted for in the case statements.

	for (DeviceType device = kDeviceUnspecified; device < kDeviceLast; ++device)
	{
		const char* name = NULL;

		switch (device)
		{
			case kDeviceUnspecified:			name = "Unspecified";			break;
			case kDevicePilot1000:				name = "Pilot1000";				break;
			case kDevicePilot5000:				name = "Pilot5000";				break;
			case kDevicePalmPilotPersonal:		name = "PalmPilotPersonal";		break;
			case kDevicePalmPilotProfessional:	name = "PalmPilotProfessional";	break;
			case kDevicePalmPilotWithUpgrade:	name = "PalmPilotWithUpgrade";	break;
			case kDevicePalmIII:				name = "PalmIII";				break;
			case kDevicePalmVII:				name = "PalmVII";				break;
			case kDeviceUnused1:												break;
			case kDeviceQualComm:				name = "QualComm";				break;
			case kDevicePalmV:					name = "PalmV";					break;
			case kDevicePalmIIIx:				name = "PalmIIIx";				break;
			case kDeviceUnused2:												break;
			case kDevicePalmIIIc: 				name = "PalmIIIc";				break;
			case kDevicePalmVIIEZ: 				name = "PalmVIIEZ";				break;
			case kDevicePalmIIIe:				name = "PalmIIIe";				break;
			case kDevicePalmVx:					name = "PalmVx";				break;
			case kDeviceSymbol1700:				name = "Symbol1700";			break;
			case kDeviceTRGpro:					name = "TRGpro";				break;
			case kDeviceVisor:					name = "Visor";					break;
			case kDevicem100:					name = "m100";					break;
			case kDevicePalmVIIx:				name = "PalmVIIx";				break;
			case kDeviceReserved1:												break;
			case kDeviceReserved2:												break;
			case kDeviceReserved3:												break;
			case kDeviceReserved4:												break;
			case kDeviceReserved5:												break;
			case kDeviceVZGeneric:				name = "VZDevice";				break;
			case kDeviceReserved7:												break;
			case kDeviceReserved8:												break;
			case kDeviceLast:					assert (false);					break;
		}

		if (name)
			result.push_back (make_pair (device, string (name)));
	}

	return result;
}
