/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDlg.h"

#include "EmFileImport.h"		// EmFileImport
#include "EmROMTransfer.h"		// EmROMTransfer
#include "EmROMReader.h"		// EmROMReader
#include "EmStreamFile.h"		// EmStreamFile
#include "EmTransportSerial.h"	// PortNameList, GetSerialPortNameList, etc.
#include "Hordes.h"				// Hordes::New
#include "Logging.h"			// FOR_EACH_LOG_PREF
#include "Miscellaneous.h"		// MemoryTextList, GetMemoryTextList
#include "Platform.h"			// Platform
#include "PreferenceMgr.h"		// Preference, gEmuPrefs
#include "ROMStubs.h"			// FSCustomControl, SysLibFind
#include "Strings.r.h"			// kStr_AppName
#include "Skins.h"				// SkinNameList

#include "algorithm"			// find, remove_if, unary_function<>


static void PrvConvertBaudListToStrings (const EmTransportSerial::BaudList& baudList,
									StringList& baudStrList);

/***********************************************************************
 *
 * FUNCTION:	DoGetFile
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID EmDlg::DoGetFile (	EmFileRef& result,
								const string& prompt,
								const EmDirRef& defaultPath,
								const EmFileTypeList& filterList)
{
	return EmDlg::HostRunGetFile (result, prompt, defaultPath, filterList);
}


/***********************************************************************
 *
 * FUNCTION:	DoGetFileList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID EmDlg::DoGetFileList (	EmFileRefList& results,
									const string& prompt,
									const EmDirRef& defaultPath,
									const EmFileTypeList& filterList)
{
	return EmDlg::HostRunGetFileList (results, prompt, defaultPath, filterList);
}


/***********************************************************************
 *
 * FUNCTION:	DoPutFile
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID EmDlg::DoPutFile (	EmFileRef& result,
								const string& prompt,
								const EmDirRef& defaultPath,
								const EmFileTypeList& filterList,
								const string& defaultName)
{
	return EmDlg::HostRunPutFile (result, prompt, defaultPath, filterList, defaultName);
}


/***********************************************************************
 *
 * FUNCTION:	DoGetDirectory
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID EmDlg::DoGetDirectory (	EmDirRef& result,
									const string& prompt,
									const EmDirRef& defaultPath)
{
	return EmDlg::HostRunGetDirectory (result, prompt, defaultPath);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoAboutBox
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgFnResult EmDlg::PrvAboutBox (EmDlgContext& context)
{
	EmDlgRef	dlg = context.fDlg;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			if (context.fPanelID == kDlgPanelAbtMain)
			{
				Errors::SetParameter ("%application", Platform::GetString (kStr_AppName));
				Errors::SetParameter ("%version", Platform::GetShortVersionString ());

				string	appAndVersion = Errors::ReplaceParameters (kStr_AppAndVers);

				EmDlg::SetItemText (dlg, kDlgItemAbtAppName, appAndVersion);
			}
			
			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}
				
				case kDlgItemAbtURLPalmWeb:
				case kDlgItemAbtURLPalmMail:
				case kDlgItemAbtURLWindowsWeb:
				case kDlgItemAbtURLWindowsMail:
				case kDlgItemAbtURLMacWeb:
				case kDlgItemAbtURLMacMail:
				case kDlgItemAbtURLUAEWeb:
				case kDlgItemAbtURLUAEMail:
				case kDlgItemAbtURLQNXWeb:
				case kDlgItemAbtURLQNXMail:
				{
					string	URL = EmDlg::GetItemText (dlg, context.fItemID);
#if 0
					::LaunchURL (URL);
#endif
					break;
				}
				
				default:
					assert (false);
			}

			break;
		}

		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoAboutBox (void)
{
	return EmDlg::HostRunDialog (EmDlg::PrvAboutBox, NULL, kDlgAboutBox);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoSessionNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

struct SessionNewData
{
	Configuration*		cfg;
	Configuration		fWorkingCfg;
	SkinNameList		fSkinChoices;
	EmDeviceMenuItemList	fDeviceChoices;
};

EmDlgFnResult EmDlg::PrvSessionNew (EmDlgContext& context)
{
	SessionNewData*	data = (SessionNewData*) context.fUserData;
	EmDlgRef		dlg = context.fDlg;
	Configuration&	cfg = data->fWorkingCfg;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			// Make a copy of the configuration that we update as the
			// user makes dialog selections.

			cfg = *(data->cfg);

			// Get the list of skins that the user wants to use.

			Preference<SkinNameList>	prefSkins (kPrefKeySkins);
			data->fSkinChoices = *prefSkins;

			// If a file is specified, make sure it is in the MRU list
			if (cfg.fROMFile.IsSpecified())
				gEmuPrefs->UpdateROMMRU (cfg.fROMFile);

			// Build our menus.

			EmDlg::PrvBuildROMMenu (context);
			EmDlg::PrvBuildDeviceMenu (context);
			EmDlg::PrvBuildSkinMenu (context);
			EmDlg::PrvBuildMemorySizeMenu (context);
			EmDlg::PrvSetROMName (context);
			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					*(data->cfg) = cfg;
					
					Preference<SkinNameList>	prefSkins (kPrefKeySkins);
					prefSkins = data->fSkinChoices;
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}
				
				case kDlgItemNewDevice:
				{
					EmDeviceMenuItemList::size_type	index	= EmDlg::GetItemValue (dlg, kDlgItemNewDevice);

					cfg.fDeviceType = data->fDeviceChoices [index].first;

					EmDlg::PrvBuildSkinMenu (context);
					break;
				}

				case kDlgItemNewSkin:
				{
					SkinNameList	skins;
					::SkinGetSkinNames (cfg.fDeviceType, skins);

					DeviceType				device	= cfg.fDeviceType;
					SkinNameList::size_type	index	= EmDlg::GetItemValue (dlg, kDlgItemNewSkin);

					data->fSkinChoices [device] = skins [index];
					break;
				}

				case kDlgItemNewMemorySize:
				{
					MemoryTextList	sizes;
					::GetMemoryTextList (sizes);

					MemoryTextList::size_type	index = EmDlg::GetItemValue (dlg, kDlgItemNewMemorySize);

					cfg.fRAMSize = sizes [index].first;
					break;
				}

				case kDlgItemNewROMBrowse:
					break;

				case kDlgItemNewROM:
				{
					EmFileTypeList	typeList;
					typeList.push_back (kFileTypeROM);

					EmFileRef	romFileRef;
					EmDirRef	romDirRef (cfg.fROMFile.GetParent ());

					int		index = EmDlg::GetItemValue (dlg, kDlgItemNewROM);
					
					if (index >= 2) 
					{
						EmFileRef       fileRef = gEmuPrefs->GetIndROMMRU (index - 2);

						cfg.fROMFile = fileRef;
					}
					else if (EmDlg::DoGetFile (romFileRef, "Choose ROM file:",
										romDirRef, typeList) == kDlgItemOK)
					{
						EmDeviceMenuItemList devices = EmDevice::GetMenuItemList ();
						EmDeviceMenuItemList::iterator	devices_end;
						
						devices_end = PrvFilterDeviceMenuItemList(romFileRef, devices);

						if (devices_end == devices.begin()) {
							// No devices matched
							
							Errors::DoDialog(	kStr_UnknownDevice,
												Errors::kCautionAlert |
												Errors::kOK);
						}
					
						cfg.fROMFile = romFileRef;
					}
					
					if (cfg.fROMFile.IsSpecified())
						gEmuPrefs->UpdateROMMRU (cfg.fROMFile);

					EmDlg::PrvSetROMName (context);
						
					EmDlg::PrvBuildROMMenu (context);
					EmDlg::PrvBuildDeviceMenu (context);
					EmDlg::PrvBuildSkinMenu (context);

					break;
				}
				
				default:
					assert (false);
			}	// switch (context.fItemID)

			break;
		}	// case kDlgCmdItemSelected

		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}	// switch (command)

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoSessionNew (Configuration& cfg)
{
	SessionNewData	data;
	data.cfg = &cfg;

	return EmDlg::HostRunDialog (EmDlg::PrvSessionNew, &data, kDlgSessionNew);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoSessionSave
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID	EmDlg::DoSessionSave (const string& docName, Bool quitting)
{
	string	appName = Platform::GetString (kStr_AppName);
	string	docName2 (docName);

	if (docName2.empty ())
		docName2 = Platform::GetString (kStr_Untitled);

	return EmDlg::HostRunSessionSave (appName, docName2, quitting);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoHordeNew
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

struct HordeNewData
{
	HordeInfo		info;
	EmDatabaseList	appList;
};

EmDlgFnResult EmDlg::PrvHordeNew (EmDlgContext& context)
{
	HordeNewData*	data = (HordeNewData*) context.fUserData;
	EmDlgRef		dlg = context.fDlg;
	HordeInfo&		info = data->info;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			// Get Horde preference settings.

			Preference<HordeInfo>	pref (kPrefKeyHordeInfo);
			info = *pref;

			// Install the settings into the dialog box items.

			EmDlg::SetItemValue (dlg, kDlgItemHrdStart,			info.fStartNumber);
			EmDlg::SetItemValue (dlg, kDlgItemHrdStop,			info.fStopNumber);
			EmDlg::SetItemValue (dlg, kDlgItemHrdDepthSwitch,	info.fDepthSwitch);
			EmDlg::SetItemValue (dlg, kDlgItemHrdDepthSave,		info.fDepthSave);
			EmDlg::SetItemValue (dlg, kDlgItemHrdDepthStop,		info.fDepthStop);
			EmDlg::SetItemValue (dlg, kDlgItemHrdCheckSwitch,	info.fCanSwitch);
			EmDlg::SetItemValue (dlg, kDlgItemHrdCheckSave,		info.fCanSave);
			EmDlg::SetItemValue (dlg, kDlgItemHrdCheckStop,		info.fCanStop);

			// Get the list of applications.

			::GetDatabases (data->appList, kApplicationsOnly);

			// Insert the names of the applications into the list.

			Bool	selected = false;
			EmDatabaseList::iterator	iter = data->appList.begin();
			while (iter != data->appList.end())
			{
				DatabaseInfo&	thisDB = *iter;

				// Add the current item.

				EmDlg::AppendToList (dlg, kDlgItemHrdAppList, thisDB.name);

				// If the item we just added is in our "selected" list,
				// then select it.

				EmDatabaseList::iterator	begin = info.fAppList.begin ();
				EmDatabaseList::iterator	end = info.fAppList.end ();

				if (find (begin, end, thisDB) != end)
				{
					selected = true;
					int index = iter - data->appList.begin();
					EmDlg::SelectListItem (dlg, kDlgItemHrdAppList, index);
				}

				++iter;
			}

			// If nothing's selected, then select something (the first item).

			if (!selected)
			{
				EmDlg::SelectListItem (dlg, kDlgItemHrdAppList, 0);
			}

			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					info.fStartNumber	= EmDlg::GetItemValue (dlg, kDlgItemHrdStart);
					info.fStopNumber	= EmDlg::GetItemValue (dlg, kDlgItemHrdStop);
					info.fDepthSwitch	= EmDlg::GetItemValue (dlg, kDlgItemHrdDepthSwitch);
					info.fDepthSave		= EmDlg::GetItemValue (dlg, kDlgItemHrdDepthSave);
					info.fDepthStop		= EmDlg::GetItemValue (dlg, kDlgItemHrdDepthStop);
					info.fCanSwitch		= EmDlg::GetItemValue (dlg, kDlgItemHrdCheckSwitch);
					info.fCanSave		= EmDlg::GetItemValue (dlg, kDlgItemHrdCheckSave);
					info.fCanStop		= EmDlg::GetItemValue (dlg, kDlgItemHrdCheckStop);

					// Get the indexes for the selected items.

					EmDlgListIndexList	items;
					EmDlg::GetSelectedItems (dlg, kDlgItemHrdAppList, items);

					// Use the indexes to get the actual items, and add those
					// items to the "result" container.

					info.fAppList.clear ();
					EmDlgListIndexList::iterator	iter = items.begin();
					while (iter != items.end())
					{
						DatabaseInfo	dbInfo = data->appList[*iter];
						info.fAppList.push_back (dbInfo);
						++iter;
					}

					// Transfer the new fields to the old fields for Hordes to use.

					info.NewToOld ();

					// Save the preferences

					Preference<HordeInfo>	pref (kPrefKeyHordeInfo);
					pref = info;

					// Startup up the gremlin sub-system.

					Hordes::New (info);

					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				case kDlgItemHrdAppList:
				case kDlgItemHrdStart:
				case kDlgItemHrdStop:
				case kDlgItemHrdCheckSwitch:
				case kDlgItemHrdCheckSave:
				case kDlgItemHrdCheckStop:
				case kDlgItemHrdDepthSwitch:
				case kDlgItemHrdDepthSave:
				case kDlgItemHrdDepthStop:
					break;

				case kDlgItemHrdLogging:
					EmDlg::DoEditLoggingOptions (kGremlinLogging);
					break;

				default:
					assert (false);
			}

			break;
		}

		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoHordeNew (void)
{
	HordeNewData	data;

	return EmDlg::HostRunDialog (EmDlg::PrvHordeNew, &data, kDlgHordeNew);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoDatabaseImport
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

struct DatabaseImportData
{
	EmFileRefList	fFiles;
	int				fMethod;
	EmStreamFile*	fStream;
	EmFileImport*	fImporter;
	long			fProgressCurrent;	// Progress into the current file
	long			fProgressBase;		// Progress accumulated by previous files
	long			fProgressMax;		// Progress max value.
	long			fCurrentFile;		// Current file we're installing.
	long			fDoneStart;
};

EmDlgFnResult EmDlg::PrvDatabaseImport (EmDlgContext& context)
{
	DatabaseImportData*	data = (DatabaseImportData*) context.fUserData;
	EmDlgRef			dlg = context.fDlg;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			data->fStream		= NULL;
			data->fImporter		= NULL;

			try
			{
				data->fCurrentFile		= 0;
				data->fStream			= new EmStreamFile (data->fFiles[data->fCurrentFile], kOpenExistingForRead);
				data->fImporter			= new EmFileImport (*data->fStream, data->fMethod);
				data->fProgressCurrent	= -1;
				data->fProgressBase		= 0;
				data->fProgressMax		= EmFileImport::CalculateProgressMax (data->fFiles, data->fMethod);
				data->fDoneStart		= 0;

				long	remainingFiles	= data->fFiles.size () - data->fCurrentFile;
				string	curAppName		= data->fFiles[data->fCurrentFile].GetName ();

				EmDlg::SetItemValue	(dlg, kDlgItemImpNumFiles,		remainingFiles);
				EmDlg::SetItemText	(dlg, kDlgItemImpCurAppName,	curAppName);
			}
			catch (...)
			{
				delete data->fImporter;
				delete data->fStream;

				data->fStream		= NULL;
				data->fImporter		= NULL;
			}

			context.fNeedsIdle = true;
			break;
		}

		case kDlgCmdIdle:
		{
			if (data->fImporter)
			{
				// Prevent against recursion (can occur if we need to display an error dialog).

				static Bool	inIdle;
				if (inIdle)
					break;

				inIdle = true;

				// Install a little bit of the file.

				ErrCode	err = data->fImporter->Continue ();

				// If an error occured, display a dialog.

				if (err)
				{
					string	curAppName		= data->fFiles[data->fCurrentFile].GetName ();
					Errors::SetParameter ("%filename", curAppName);
					Errors::ReportIfError (kStr_CmdInstall, err, 0, false);
				}

				// Everything went OK.  Get the current progress amount.
				// If it changed, update the dialog.
				else
				{
					long	progressCurrent	= data->fImporter->GetProgress ();
					long	progressMax		= data->fProgressMax;

					if (data->fProgressCurrent != progressCurrent)
					{
						data->fProgressCurrent = progressCurrent;

						progressCurrent += data->fProgressBase;

						// Scale progressMax to < 32K (progress control on Windows
						// prefers it that way).

						long	divider = progressMax / (32 * 1024L) + 1;

						progressMax /= divider;
						progressCurrent /= divider;

						long	remainingFiles	= data->fFiles.size () - data->fCurrentFile;
						string	curAppName		= data->fFiles[data->fCurrentFile].GetName ();

						EmDlg::SetItemValue	(dlg, kDlgItemImpNumFiles,		remainingFiles);
						EmDlg::SetItemText	(dlg, kDlgItemImpCurAppName,	curAppName);
						EmDlg::SetItemMax	(dlg, kDlgItemImpProgress,		progressMax);
						EmDlg::SetItemValue	(dlg, kDlgItemImpProgress,		progressCurrent);
					}
				}

				// If we're done with this file, clean up, and prepare for any
				// subsequent files.

				if (data->fImporter->Done ())
				{
					gEmuPrefs->UpdatePRCMRU (data->fFiles[data->fCurrentFile]);

					data->fProgressCurrent = -1;
					data->fProgressBase += data->fImporter->GetProgress ();

					delete data->fImporter;
					data->fImporter = NULL;

					delete data->fStream;
					data->fStream = NULL;

					data->fCurrentFile++;
					if (data->fCurrentFile < (long) data->fFiles.size ())
					{
						try
						{
							data->fStream	= new EmStreamFile (data->fFiles[data->fCurrentFile], kOpenExistingForRead);
							data->fImporter	= new EmFileImport (*data->fStream, data->fMethod);
						}
						catch (...)
						{
							delete data->fImporter;
							delete data->fStream;

							data->fStream		= NULL;
							data->fImporter		= NULL;
						}
					}
					else
					{
						// There are no more files to install.  Initialize our
						// counter used to delay the closing of the dialog.

						data->fDoneStart = Platform::GetMilliseconds ();
					}
				}

				inIdle = false;
			}

			// After we're done installing all files, wait a little bit in order to
			// see the completed progress dialog.

			else if (Platform::GetMilliseconds () - data->fDoneStart > 500)
			{
				return kDlgResultClose;
			}

			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					if (context.fItemID == kDlgItemCancel)
						data->fImporter->Cancel ();

					delete data->fImporter;
					data->fImporter = NULL;

					return kDlgResultClose;
				}

				case kDlgItemImpNumFiles:
				case kDlgItemImpCurAppName:
				case kDlgItemImpProgress:
					break;

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}


EmDlgItemID EmDlg::DoDatabaseImport (const EmFileRefList& files, int method)
{
	// Validate the list of files.

	EmFileRefList					newFileList;
	EmFileRefList::const_iterator	iter = files.begin ();

	while (iter != files.end ())
	{
		if (iter->IsSpecified () && iter->Exists ())
		{
			newFileList.push_back (*iter);
		}
		else
		{
			// !!! Report that the file doesn't exist.
		}

		++iter;
	}

	if (newFileList.size () <= 0)
	{
		return kDlgItemOK;
	}

	DatabaseImportData	data;

	data.fFiles		= newFileList;
	data.fMethod	= method;

	return EmDlg::HostRunDialog (EmDlg::PrvDatabaseImport, &data, kDlgDatabaseImport);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoDatabaseExport
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

#if 0
struct ExportDatbaseData
{
	EmDatabaseList*	dbList;
	EmDatabaseList	allDatabases;
};
#endif

EmDlgFnResult EmDlg::PrvDatabaseExport (EmDlgContext&)
{
#if 0
	ExportDatbaseData*	data = (ExportDatbaseData*) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			// Clear out the results container.

			data->selectedDatabases.clear();

			// Get the list of installed databases.

			::GetDatabases (data->allDatabases, kAllDatabases);

			// Add then names of the databases to the list.

			EmDatabaseList::iterator	iter = data->allDatabases.begin();
			while (iter != data->allDatabases.end())
			{
				EmDlg::AppendToList (dlg, kDlgItemExpDbList, iter->name);
				++iter;
			}
			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK;
				{
					// Get the indexes for the selected items.

					EmDlgSelectedItemsList	items;
					EmDlg::GetSelectedItems(dlg, kDlgItemExpDbList, items);

					// Use the indexes to get the actual items, and add those
					// items to the "result" container.

					EmDlgSelectedItemsList::iterator	iter = items.begin();
					while (iter != items.end())
					{
						data->dbList->push_back (data->allDatabases[*iter]);
						++iter;
					}

					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				default:
					assert (false);
			}

			break;
		}
	}
#endif

	return kDlgResultContinue;
}

#if 0
EmDlgItemID EmDlg::DoDatabaseExport (EmDatabaseList& dbList)
{
	ExportDatbaseData	data;
	data.dbList = &dbList;

	return EmDlg::HostRunDialog (EmDlg::PrvDatabaseExport, &data, kDlgDatabaseExport);
}
#endif


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoROMTransferQuery
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgFnResult EmDlg::PrvROMTransferQuery (EmDlgContext& context)
{
	EmDlgRef				dlg = context.fDlg;
	EmTransport::Config**	config = (EmTransport::Config**) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			EmTransportSerial::PortNameList	portList;
			EmTransportSerial::GetSerialPortNameList (portList);

			EmTransportSerial::BaudList		baudList;
			EmTransportSerial::GetSerialBaudList (baudList);

			StringList		baudStrList;
			::PrvConvertBaudListToStrings (baudList, baudStrList);

			// On Unix, the port is specified in an edit text item.
			// On other platforms, we present a list of ports in a menu.

			// !!! TBD: restore port setting from a preference.

#if PLATFORM_UNIX
			EmDlg::SetItemText (dlg, kDlgItemDlqPortList, portList[0]);
#else
			EmDlg::AppendToMenu (dlg, kDlgItemDlqPortList, portList);
#endif
			EmDlg::AppendToMenu (dlg, kDlgItemDlqBaudList, baudStrList);

			EmDlg::SetItemValue (dlg, kDlgItemDlqPortList, 0);		// First port
			EmDlg::SetItemValue (dlg, kDlgItemDlqBaudList, 0);		// 115K baud

			// Load the instructions string.  Load it in 9 parts,
			// as the entire string is greater than:
			//
			//	Rez can handle on the Mac.
			//	MSDev's string editor on Windows.
			//	The ANSI-allowed string length (for Unix).

			string	ins;
			for (int ii = 0; ii < 9; ++ii)
			{
				string	temp = Platform::GetString (kStr_ROMTransferInstructions + ii);
				ins += temp;
			}

			EmDlg::SetItemText (dlg, kDlgItemDlqInstructions, ins);

			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// Get the selected settings.

#if PLATFORM_UNIX
					string	portName = EmDlg::GetItemText (dlg, kDlgItemDlqPortList);
#else
					long	portNum = EmDlg::GetItemValue (dlg, kDlgItemDlqPortList);

					EmTransportSerial::PortNameList	portList;
					EmTransportSerial::GetSerialPortNameList (portList);

					string	portName = portList [portNum];
#endif
					long	baudNum = EmDlg::GetItemValue (dlg, kDlgItemDlqBaudList);

					EmTransportSerial::BaudList		baudList;
					EmTransportSerial::GetSerialBaudList (baudList);

					EmTransportSerial::ConfigSerial*	serConfig = new EmTransportSerial::ConfigSerial;

					serConfig->fPort			= portName;
					serConfig->fBaud			= baudList [baudNum];
					serConfig->fParity			= EmTransportSerial::kNoParity;
					serConfig->fStopBits		= 1;
					serConfig->fDataBits		= 8;
					serConfig->fHwrHandshake	= true;

					*config = serConfig;

					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				case kDlgItemDlqInstructions:
				case kDlgItemDlqPortList:
				case kDlgItemDlqBaudList:
					break;

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoROMTransferQuery (EmTransport::Config*& config)
{
	config = NULL;
	return EmDlg::HostRunDialog (EmDlg::PrvROMTransferQuery, &config, kDlgROMTransferQuery);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoROMTransferProgress
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgFnResult EmDlg::PrvROMTransferProgress (EmDlgContext& context)
{
	EmDlgRef		dlg = context.fDlg;
	EmROMTransfer*	transferer = (EmROMTransfer*) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			context.fNeedsIdle = true;
			break;
		}

		case kDlgCmdIdle:
		{
			if (!transferer->Continue (dlg))
				return kDlgResultClose;

			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					transferer->Abort (dlg);
					return kDlgResultClose;
				}

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoROMTransferProgress (EmROMTransfer& transferer)
{
	return EmDlg::HostRunDialog (EmDlg::PrvROMTransferProgress, &transferer, kDlgROMTransferProgress);
}


/***********************************************************************
 *
 * FUNCTION:	GremlinControlOpen
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:	GremlinControlUpdate
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:	GremlinControlClose
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:	DoEditPreferences
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoEditLoggingOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

struct EditLoggingOptionsData
{
	LoggingType		fActiveSet;

	// Define a "uint8" variable for each of the logging options we
	// support.  The name of each variable is of the form "f<name>",
	// where "<name>" is what appear in the FOR_EACH_LOG_PREF macro.

#define DEFINE_STORAGE(name)	uint8	f##name;
	FOR_EACH_LOG_PREF (DEFINE_STORAGE)
};

static void PrvFetchPrefs (EmDlgContext& context)
{
	// Transfer all of the logging values from the preferences
	// to our local storage.

	EditLoggingOptionsData&	data = *(EditLoggingOptionsData*) context.fUserData;

	#define GET_PREF(name)	\
		{	\
			Preference<uint8> pref(kPrefKey##name);	\
			data.f##name = *pref;	\
		}

	FOR_EACH_LOG_PREF (GET_PREF)
}

static void PrvInstallPrefs (EmDlgContext& context)
{
	// Transfer all of the logging values from our local
	// storages to the preferences system.

	EditLoggingOptionsData&	data = *(EditLoggingOptionsData*) context.fUserData;

	#define SET_PREF(name)	\
		{	\
			Preference<uint8> pref(kPrefKey##name);	\
			pref = data.f##name;	\
		}

	FOR_EACH_LOG_PREF (SET_PREF)
}

static void PrvPrefsToButtons (EmDlgContext& context)
{
	// Set the buttons in in the current panel.

	EditLoggingOptionsData&	data = *(EditLoggingOptionsData*) context.fUserData;
	int	bitmask = data.fActiveSet;

	#define TO_BUTTON(name)	\
		EmDlg::SetItemValue (context.fDlg, kDlgItemLog##name, (data.f##name & bitmask) != 0);

	FOR_EACH_LOG_PREF (TO_BUTTON)
}

static void PrvPrefsFromButtons (EmDlgContext& context)
{
	// Update our local preference values from the buttons in the current panel.

	EditLoggingOptionsData&	data = *(EditLoggingOptionsData*) context.fUserData;
	int	bitmask = data.fActiveSet;

	#define FROM_BUTTON(name)	\
		if (EmDlg::GetItemValue (context.fDlg, kDlgItemLog##name) != 0)	\
			data.f##name |= bitmask;	\
		else	\
			data.f##name &= ~bitmask;

	FOR_EACH_LOG_PREF (FROM_BUTTON)
}

EmDlgFnResult EmDlg::PrvEditLoggingOptions (EmDlgContext& context)
{
	EmDlgRef				dlg = context.fDlg;
	EditLoggingOptionsData&	data = *(EditLoggingOptionsData*) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			::PrvFetchPrefs (context);
			::PrvPrefsToButtons (context);

			if (data.fActiveSet == kNormalLogging)
				EmDlg::SetItemValue (dlg, kDlgItemLogNormal, 1);
			else
				EmDlg::SetItemValue (dlg, kDlgItemLogGremlins, 1);

			// Disable unsupported options.

			EmDlg::DisableItem (dlg, kDlgItemLogLogErrorMessages);
			EmDlg::DisableItem (dlg, kDlgItemLogLogWarningMessages);
			EmDlg::DisableItem (dlg, kDlgItemLogLogCPUOpcodes);
			EmDlg::DisableItem (dlg, kDlgItemLogLogApplicationCalls);

			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					::PrvPrefsFromButtons (context);
					::PrvInstallPrefs (context);

					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				case kDlgItemLogNormal:
				case kDlgItemLogGremlins:
					::PrvPrefsFromButtons (context);
					if (context.fItemID == kDlgItemLogNormal)
						data.fActiveSet = kNormalLogging;
					else
						data.fActiveSet = kGremlinLogging;
					::PrvPrefsToButtons (context);
					break;

				#define DUMMY_CASE(name)	case kDlgItemLog##name:
				FOR_EACH_LOG_PREF (DUMMY_CASE)
					break;

				default:
					assert (false);
			}

			break;
		}

		case kDlgCmdPanelEnter:
			break;

		case kDlgCmdPanelExit:
			break;

		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}


EmDlgItemID EmDlg::DoEditLoggingOptions (LoggingType initialSet)
{
	EditLoggingOptionsData	data;

	data.fActiveSet	= initialSet;

	return EmDlg::HostRunDialog (EmDlg::PrvEditLoggingOptions, &data, kDlgEditLogging);
}


/***********************************************************************
 *
 * FUNCTION:	DoEditDebuggingOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoEditSkins
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgFnResult EmDlg::PrvEditSkins (EmDlgContext& context)
{
	EmDlgRef	dlg = context.fDlg;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			// Get the current configuration so that we know what
			// device is being emulated.

			Preference<Configuration>	prefConfiguration (kPrefKeyLastConfiguration);
			Configuration				cfg = *prefConfiguration;

			// Get the list of selected skins so that we know which
			// one to select in the listbox.

			Preference<SkinNameList>	prefSkins (kPrefKeySkins);
			SkinNameList				skinChoices = *prefSkins;

			// Get the list of skins for the selected device.

			SkinNameList	skins;
			::SkinGetSkinNames (cfg.fDeviceType, skins);

			// Add each skin in the list to the UI element.

			Bool					selected = false;
			int						index = 0;
			SkinNameList::iterator	skinIter = skins.begin();

			while (skinIter != skins.end())
			{
				string&	skinName = *skinIter;
				EmDlg::AppendToList (dlg, kDlgItemSknSkinList, skinName);

				// Select the currently-used skin.

				if (skinIter->c_str() == skinChoices[cfg.fDeviceType])
				{
					EmDlg::SelectListItem (dlg, kDlgItemSknSkinList, index);
					selected = true;
				}

				++index;
				++skinIter;
			}

			// Ensure that *something* was selected.

			if (!selected)
			{
				EmDlg::SelectListItem (dlg, kDlgItemSknSkinList, 0);
			}

			// Set up the checkboxes

			Preference<ScaleType>	prefScale (kPrefKeyScale);
			EmDlg::SetItemValue (dlg, kDlgItemSknDoubleScale, (*prefScale == 2) ? 1 : 0);

			Preference<RGBType>		prefBackgroundColor (kPrefKeyBackgroundColor);
			EmDlg::SetItemValue (dlg, kDlgItemSknWhiteBackground, prefBackgroundColor.Loaded () ? 1 : 0);

			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// Update the skin choices.
					
					// Get the current configuration so that we know what
					// device is being emulated.

					Preference<Configuration>	prefConfiguration (kPrefKeyLastConfiguration);
					Configuration				cfg = *prefConfiguration;

					// Get the selected skin.

					EmDlgListIndex	index = EmDlg::GetSelectedItem (dlg, kDlgItemSknSkinList);
					assert (index != kDlgItemListNone);

					// Get the skins for this device.

					SkinNameList	deviceSkins;
					::SkinGetSkinNames (cfg.fDeviceType, deviceSkins);

					// Get the skins preferences.

					Preference<SkinNameList>	prefSkins (kPrefKeySkins);

					// Update the preferences.

					SkinNameList	skins;
					skins = *prefSkins;
					skins[cfg.fDeviceType] = deviceSkins[index];
					prefSkins = skins;

					// Update the scale.

					Preference<ScaleType>	prefScale (kPrefKeyScale);
					prefScale = (GetItemValue (dlg, kDlgItemSknDoubleScale) != 0) ? 2 : 1;

					// Update the background color.

					if (EmDlg::GetItemValue (dlg, kDlgItemSknWhiteBackground) != 0)
					{
						RGBType				color (0xFF, 0xFF, 0xFF);
						Preference<RGBType>	prefBackgroundColor (kPrefKeyBackgroundColor);
						prefBackgroundColor = color;
					}
					else
					{
						gPrefs->DeletePref (kPrefKeyBackgroundColor);
					}
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				case kDlgItemSknSkinList:
				case kDlgItemSknDoubleScale:
				case kDlgItemSknWhiteBackground:
					break;

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoEditSkins (void)
{
	return EmDlg::HostRunDialog (EmDlg::PrvEditSkins, NULL, kDlgEditSkins);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoEditCardOptions
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

struct EditCardOptionsData
{
	SlotInfoList	fWorkingInfo;
};


EmDlgFnResult EmDlg::PrvEditCardOptions (EmDlgContext& context)
{
	EmDlgRef				dlg = context.fDlg;
	EditCardOptionsData&	data = *(EditCardOptionsData*) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			// Get the current preferences.

			Preference<SlotInfoList>	pref (kPrefKeySlotList);
			data.fWorkingInfo = *pref;

			// Make sure there are at least 4 entries.  If not, then
			// fill in the missing ones.

			long	curSize = data.fWorkingInfo.size ();
			while (curSize < 4)
			{
				SlotInfoType	info;
				info.fSlotNumber = curSize + 1;
				info.fSlotOccupied = false;
				data.fWorkingInfo.push_back (info);

				curSize++;
			}

			// Select one of the cards.

			EmDlg::ClearList (dlg, kDlgItemCrdList);
			EmDlg::AppendToList (dlg, kDlgItemCrdList, "1");
			EmDlg::AppendToList (dlg, kDlgItemCrdList, "2");
			EmDlg::AppendToList (dlg, kDlgItemCrdList, "3");
			EmDlg::AppendToList (dlg, kDlgItemCrdList, "4");

			EmDlg::SelectListItem (dlg, kDlgItemCrdList, 0);

			// Call ourselves to make sure that the other dialog
			// items are properly initialized in light of the
			// currently selected item.

			EmDlgContext	subContext (context);
			subContext.fCommandID = kDlgCmdItemSelected;
			subContext.fItemID = kDlgItemCrdList;
			EmDlg::PrvEditCardOptions (subContext);

			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// User pressed OK, save the changes.

					Preference<SlotInfoList>	pref (kPrefKeySlotList);
					SlotInfoList				origList = *pref;
					pref = data.fWorkingInfo;

					// For any that changed, mount/unmount the cards.

					UInt16	refNum;
					Err		err = ::SysLibFind ("HostFS Library", &refNum);

					if (err == errNone && refNum != sysInvalidRefNum)
					{
						SlotInfoList::iterator	iter1 = origList.begin ();
						while (iter1 != origList.end ())
						{
							SlotInfoList::iterator	iter2 = data.fWorkingInfo.begin ();
							while (iter2 != data.fWorkingInfo.end ())
							{
								if (iter1->fSlotNumber == iter2->fSlotNumber)
								{
									if (iter1->fSlotOccupied != iter2->fSlotOccupied)
									{
										const UInt32	kCreator	= 'pose';
										const UInt16	kMounted	= 1;
										const UInt16	kUnmounted	= 0;

										UInt16	selector = iter2->fSlotOccupied ? kMounted : kUnmounted;
										void*	cardNum = (void*) iter1->fSlotNumber;

										// Note, in order to make this call, the CPU should be stopped at
										// an idle event (kStopOnIdle).  That's because mounting and
										// unmounting can send out notification.  If the notification is
										// sent out in the background (that is, while the current Palm OS
										// task is not the UI task), then the notification manager calls
										// SysTaskWait.  This will switch to another task if it can, and
										// prime a timer to re-awake the background task if not.  However,
										// this timer is based on an interrupt going off, and while we're
										// calling into the ROM, interrupts are turned off, leading to a
										// hang.  Therefore, is is imperative that we call this function
										// while the UI task is the current task.

										::FSCustomControl (refNum, kCreator, selector, cardNum, NULL);
									}

									break;
								}

								++iter2;
							}

							++iter1;
						}
					}

					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}
				
				case kDlgItemCrdList:
				{
					// User changed the currently selected card.

					EmDlgListIndex	item = EmDlg::GetSelectedItem (dlg, kDlgItemCrdList);
					if (item >= 0)
					{
						SlotInfoType&	info = data.fWorkingInfo[item];

						// Get the path to the card's root.  If empty, display
						// a default message so that the user doesn't just see
						// a blank pane.

						string	fullPath (info.fSlotRoot.GetFullPath ());
						if (fullPath.empty ())
							fullPath = "<Not selected>";

						// Set the root path text and the button that says if
						// it's mounted or not.

						EmDlg::SetItemText (dlg, kDlgItemCrdPath, fullPath);
						EmDlg::SetItemValue (dlg, kDlgItemCrdMounted, info.fSlotOccupied);
					}
					break;
				}

				case kDlgItemCrdPath:
					break;

				case kDlgItemCrdMounted:
				{
					// User toggled the checkbox that says whether or not the
					// card is mounted.  Save the new setting.

					EmDlgListIndex	item = EmDlg::GetSelectedItem (dlg, kDlgItemCrdList);
					if (item >= 0)
					{
						SlotInfoType&	info = data.fWorkingInfo[item];

						info.fSlotOccupied = EmDlg::GetItemValue (dlg, kDlgItemCrdMounted);
					}
					break;
				}

				case kDlgItemCrdBrowse:
				{
					// User clicked on the Browse button.  Bring up a
					// "Get Directory" dialog.

					EmDlgListIndex	item = EmDlg::GetSelectedItem (dlg, kDlgItemCrdList);
					if (item >= 0)
					{
						SlotInfoType&	info = data.fWorkingInfo[item];

						EmDirRef	result;
						string		prompt ("Select a directory:");
						EmDirRef	defaultPath (info.fSlotRoot);

						if (EmDlg::DoGetDirectory (result, prompt, defaultPath) == kDlgItemOK)
						{
							info.fSlotRoot = result;
							EmDlg::SetItemText (dlg, kDlgItemCrdPath, result.GetFullPath ());
						}
					}
					break;
				}

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}


EmDlgItemID EmDlg::DoEditCardOptions (void)
{
	EditCardOptionsData	data;

	return EmDlg::HostRunDialog (EmDlg::PrvEditCardOptions, &data, kDlgEditCards);
}


/***********************************************************************
 *
 * FUNCTION:	DoDialog
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	DoSaveBound
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgFnResult EmDlg::PrvSaveBound (EmDlgContext& context)
{
	EmDlgRef	dlg = context.fDlg;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			EmDlg::DisableItem (dlg, kDlgItemOK);
			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					Bool			fullSave = EmDlg::GetItemValue (dlg, kDlgItemBndSaveRAM) != 0;
					EmFileRef		destFile;
					EmFileTypeList	filterList (1, kFileTypeApplication);

#if PLATFORM_MAC
					string			newName ("Emulator_Bound");
#elif PLATFORM_UNIX
					string			newName ("pose_bound");
#elif PLATFORM_WINDOWS
					string			newName ("Emulator_Bound.exe");
#else
#error "Unsupported platform"
#endif

					if (EmDlg::DoPutFile (destFile, "Save new emulator",
						EmDirRef::GetEmulatorDirectory (), filterList, newName) == kDlgItemOK)
					{
						Platform::BindPoser (fullSave, destFile);
					}
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				case kDlgItemBndSaveROM:
				case kDlgItemBndSaveRAM:
					EmDlg::EnableItem (dlg, kDlgItemOK);
					break;

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}

EmDlgItemID EmDlg::DoSaveBound (void)
{
	return EmDlg::HostRunDialog (EmDlg::PrvSaveBound, NULL, kDlgSaveBound);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemText (EmDlgRef dlg, EmDlgItemID item, StrCode strCode)
{
	string	str = Platform::GetString (strCode);
	EmDlg::SetItemText (dlg, item, str);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemText (EmDlgRef dlg, EmDlgItemID item, const char* str)
{
	EmDlg::SetItemText (dlg, item, string (str));
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToMenu (EmDlgRef dlg, EmDlgItemID item, const string& text)
{
	StringList	strList;
	strList.push_back (text);

	EmDlg::AppendToMenu (dlg, item, strList);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToList (EmDlgRef dlg, EmDlgItemID item, const string& text)
{
	StringList	strList;
	strList.push_back (text);

	EmDlg::AppendToList (dlg, item, strList);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SelectListItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SelectListItem (EmDlgRef dlg, EmDlgItemID item, EmDlgListIndex listItem)
{
	EmDlgListIndexList	itemList;
	itemList.push_back (listItem);

	EmDlg::SelectListItems (dlg, item, itemList);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetSelectedItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgListIndex EmDlg::GetSelectedItem (EmDlgRef dlg, EmDlgItemID item)
{
	EmDlgListIndexList	itemList;
	EmDlg::GetSelectedItems (dlg, item, itemList);
	
	if (itemList.size () > 0)
	{
		return itemList[0];
	}
	
	return kDlgItemListNone;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmDlgContext c'tor
 *
 * DESCRIPTION:	Constructor.  Initialize our data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgContext::EmDlgContext (void) :
	fFn (NULL),
	fFnResult (kDlgResultNone),
	fDlg (NULL),
	fDlgID (kDlgNone),
	fPanelID (kDlgPanelNone),
	fItemID (kDlgItemNone),
	fCommandID (kDlgCmdNone),
	fNeedsIdle (false),
	fUserData (NULL)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext c'tor
 *
 * DESCRIPTION:	Constructor.  Initialize our data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgContext::EmDlgContext (const EmDlgContext& other) :
	fFn (other.fFn),
	fFnResult (other.fFnResult),
	fDlg (other.fDlg),
	fDlgID (other.fDlgID),
	fPanelID (other.fPanelID),
	fItemID (other.fItemID),
	fCommandID (other.fCommandID),
	fNeedsIdle (other.fNeedsIdle),
	fUserData (other.fUserData)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext::Init
 *
 * DESCRIPTION:	Call the custom dialog handler to initialize the dialog.
 *				If the dialog says that it wants idle time, then start
 *				idleing it.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The result returned by the dialog handler.
 *
 ***********************************************************************/

EmDlgFnResult EmDlgContext::Init (void)
{
	fItemID		= kDlgItemNone;
	fCommandID	= kDlgCmdInit;

	fFnResult	= fFn (*this);

	if (fNeedsIdle)
		EmDlg::HostStartIdling (*this);

	return fFnResult;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext::Event
 *
 * DESCRIPTION:	Call the custom dialog handler to handle an event.
 *
 * PARAMETERS:	item - dialog item that was selected.
 *
 * RETURNED:	The result returned by the dialog handler.
 *
 ***********************************************************************/

EmDlgFnResult EmDlgContext::Event (EmDlgItemID itemID)
{	
	fItemID		= itemID;
	fCommandID	= kDlgCmdItemSelected;

	fFnResult	= fFn (*this);
	return fFnResult;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext::Idle
 *
 * DESCRIPTION:	Call the custom dialog handler to idle.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The result returned by the dialog handler.
 *
 ***********************************************************************/

EmDlgFnResult EmDlgContext::Idle (void)
{	
	fItemID		= kDlgItemNone;
	fCommandID	= kDlgCmdIdle;

	fFnResult	= fFn (*this);
	return fFnResult;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext::PanelEnter
 *
 * DESCRIPTION:	Call the custom dialog handler to idle.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The result returned by the dialog handler.
 *
 ***********************************************************************/

EmDlgFnResult EmDlgContext::PanelEnter (void)
{	
	fItemID		= kDlgItemNone;
	fCommandID	= kDlgCmdPanelEnter;

	fFnResult	= fFn (*this);
	return fFnResult;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlgContext::PanelExit
 *
 * DESCRIPTION:	Call the custom dialog handler to idle.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The result returned by the dialog handler.
 *
 ***********************************************************************/

EmDlgFnResult EmDlgContext::PanelExit (void)
{	
	fItemID		= kDlgItemNone;
	fCommandID	= kDlgCmdPanelExit;

	fFnResult	= fFn (*this);
	return fFnResult;
}


#pragma mark -

void EmDlg::PrvBuildROMMenu (const EmDlgContext& context)
{
	const EmDlgRef&	dlg		= context.fDlg;
	SessionNewData*	data	= (SessionNewData*) context.fUserData;
	Configuration&	cfg		= data->fWorkingCfg;

	EmDeviceMenuItemList			devices = EmDevice::GetMenuItemList ();
	
	int 							menuID = 0;
	Bool							selected = false;

	EmDlg::ClearMenu (dlg, kDlgItemNewROM);

	EmDlg::AppendToMenu (dlg, kDlgItemNewROM, Platform::GetString (kStr_OtherMRU));
	EmDlg::AppendToMenu (dlg, kDlgItemNewROM, std::string());
	
	menuID = 2;
	
	for (int ii = 0; ; ++ii)
	{
		EmFileRef	fileRef = gEmuPrefs->GetIndROMMRU (ii);
		if (!fileRef.IsSpecified())
			break;
		
		EmDlg::AppendToMenu (dlg, kDlgItemNewROM, fileRef.GetName());

		if (cfg.fROMFile.GetFullPath() == fileRef.GetFullPath())
		{
			EmDlg::SelectMenuItem (dlg, kDlgItemNewROM, menuID);
			selected = true;
		}

		++menuID;

	}
	
	if (menuID == 2)
	{
		EmDlg::AppendToMenu (dlg, kDlgItemNewROM, Platform::GetString (kStr_EmptyMRU));
		EmDlg::DisableMenuItem (dlg, kDlgItemNewROM, 2);
	}
	
	if (!selected) 
	{
		EmDlg::SelectMenuItem (dlg, kDlgItemNewROM, 2);
	}

}


struct	PrvSupportsROM : unary_function<EmDeviceMenuItem &, bool>
{
	PrvSupportsROM(EmROMReader<LAS> & inROM) : ROM(inROM) {}	
	bool operator()(EmDeviceMenuItem & item)
	{
		return !EmDevice(item.first).SupportsROM(ROM);
	}

private:
	const EmROMReader<LAS> & ROM;
};

EmDeviceMenuItemList::iterator EmDlg::PrvFilterDeviceMenuItemList(const EmFileRef & romFile, EmDeviceMenuItemList & devices)
{
	EmDeviceMenuItemList::iterator devices_end = devices.end();
	if (romFile.IsSpecified())
	{
		try
		{
			EmStreamFile	hROM(romFile, kOpenExistingForRead);
			StMemory    	romImage (hROM.GetLength());
        
			hROM.GetBytes (romImage.Get (), hROM.GetLength());

			EmROMReader<LAS> ROM(romImage.Get(), hROM.GetLength());
						
			if (ROM.AcquireCardHeader()) {
				ROM.AcquireROMHeap();
				ROM.AcquireDatabases();
				ROM.AcquireFeatures();
				ROM.AcquireSplashDB();
			}
		
			devices_end = remove_if(devices.begin(), devices.end(),
				PrvSupportsROM(ROM));
				
		}
		catch (ErrCode)
		{
			/* On any of our errors, don't remove any devices */
		}
	}
	
	return devices_end;
}

void EmDlg::PrvBuildDeviceMenu (const EmDlgContext& context)
{
	const EmDlgRef&	dlg		= context.fDlg;
	SessionNewData*	data	= (SessionNewData*) context.fUserData;
	Configuration&	cfg		= data->fWorkingCfg;

	EmDeviceMenuItemList			devices = EmDevice::GetMenuItemList ();
	
	EmDeviceMenuItemList::iterator	iter = devices.begin();
	EmDeviceMenuItemList::iterator	devices_end = devices.end();
	int 							menuID = 0;
	Bool							selected = false;

	devices_end = PrvFilterDeviceMenuItemList(cfg.fROMFile, devices);
	iter = devices.begin();
	
	if (iter == devices_end)
	{
		/* We filtered out too many things, so reset to displaying all */
		devices_end = devices.end();
	}

	data->fDeviceChoices = EmDeviceMenuItemList(iter, devices_end);

	EmDlg::ClearMenu (dlg, kDlgItemNewDevice);

	while (iter != devices_end)
	{
		EmDlg::AppendToMenu (dlg, kDlgItemNewDevice, iter->second);

		if (cfg.fDeviceType == iter->first)
		{
			EmDlg::SelectMenuItem (dlg, kDlgItemNewDevice, menuID);
			selected = true;
		}

		++iter;
		++menuID;
	}
	
	if (!selected)
	{
		EmDlg::SelectMenuItem (dlg, kDlgItemNewDevice, menuID-1);
		cfg.fDeviceType = data->fDeviceChoices [menuID-1].first;
		
		/* Rely on caller to (always) invoke PrvBuildSkinMenu */
	}

}


void EmDlg::PrvBuildSkinMenu (const EmDlgContext& context)
{
	const EmDlgRef&	dlg			= context.fDlg;
	SessionNewData*	data		= (SessionNewData*) context.fUserData;
	Configuration&	cfg			= data->fWorkingCfg;
	SkinNameList&	chosenSkins	= data->fSkinChoices;

	SkinNameList	skins;
	::SkinGetSkinNames (cfg.fDeviceType, skins);

	SkinNameList::iterator	iter = skins.begin ();
	int 					menuID = 0;

	EmDlg::ClearMenu (dlg, kDlgItemNewSkin);

	while (iter != skins.end())
	{
		EmDlg::AppendToMenu (dlg, kDlgItemNewSkin, *iter);

		if (chosenSkins[cfg.fDeviceType] == *iter)
		{
			EmDlg::SelectMenuItem (dlg, kDlgItemNewSkin, menuID);
		}

		++iter;
		++menuID;
	}
}


void EmDlg::PrvBuildMemorySizeMenu (const EmDlgContext& context)
{
	const EmDlgRef&	dlg		= context.fDlg;
	SessionNewData*	data	= (SessionNewData*) context.fUserData;
	Configuration&	cfg		= data->fWorkingCfg;

	MemoryTextList	sizes;
	::GetMemoryTextList (sizes);

	MemoryTextList::iterator	iter = sizes.begin();
	int							menuID = 0;

	EmDlg::ClearMenu (dlg, kDlgItemNewMemorySize);

	while (iter != sizes.end())
	{
		EmDlg::AppendToMenu (dlg, kDlgItemNewMemorySize, iter->second);

		if (cfg.fRAMSize == iter->first)
		{
			EmDlg::SelectMenuItem (dlg, kDlgItemNewMemorySize, menuID);
		}

		++iter;
		++menuID;
	}
}


void EmDlg::PrvSetROMName (const EmDlgContext& context)
{
	const EmDlgRef&	dlg		= context.fDlg;
	SessionNewData*	data	= (SessionNewData*) context.fUserData;
	Configuration&	cfg		= data->fWorkingCfg;
	EmFileRef&		rom		= cfg.fROMFile;

	if (rom.IsSpecified ())
	{
		EmDlg::EnableItem (dlg, kDlgItemOK);
	}
	else
	{
		EmDlg::DisableItem (dlg, kDlgItemOK);
	}
}


void PrvConvertBaudListToStrings (const EmTransportSerial::BaudList& baudList,
									StringList& baudStrList)
{
	EmTransportSerial::BaudList::const_iterator	iter = baudList.begin();
	while (iter != baudList.end())
	{
		char	buffer[20];
		::FormatInteger (buffer, *iter);
		strcat (buffer, " bps");
		baudStrList.push_back (buffer);
		++iter;
	}
}


#if 0
	// Template for callback function

EmDlgFnResult EmDlg::PrvCallback (EmDlgContext& context)
{
	EmDlgRef				dlg = context.fDlg;
	<type>*	data = (<type>*) context.fUserData;

	switch (context.fCommandID)
	{
		case kDlgCmdInit:
		{
			break;
		}

		case kDlgCmdIdle:
		{
			break;
		}

		case kDlgCmdItemSelected:
		{
			switch (context.fItemID)
			{
				case kDlgItemOK:
				{
					// Fall thru...
				}

				case kDlgItemCancel:
				{
					return kDlgResultClose;
				}

				default:
					assert (false);
			}

			break;
		}
		
		case kDlgCmdPanelEnter:
		case kDlgCmdPanelExit:
		case kDlgCmdNone:
			assert (false);
			break;
	}

	return kDlgResultContinue;
}
#endif
