/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmSession.h"

#include "ChunkFile.h"			// ChunkFile
#include "EmCPU.h"				// EmCPU::Execute
#include "EmCPU68K.h"			// new EmCPU68K
#include "EmMemory.h"			// Memory::ResetBankHandlers
#include "EmPalmFunction.h"		// GetTrapName
#include "EmStreamFile.h"		// EmStreamFile
#include "Hordes.h"				// Hordes::AutoSaveState, etc.
#include "Logging.h"			// LogAppendMsg
#include "Miscellaneous.h"		// EmValueChanger
#include "PreferenceMgr.h"		// Preference
#include "ROMStubs.h"			// EvtWakeup
#include "SessionFile.h"		// SessionFile
#include "Strings.r.h"			// kStr_EnterPen

#include "EmMemory.h"			// Memory::Initialize ();
#include "Platform.h"			// Platform::Initialize ();
#include "DebugMgr.h"			// Debug::Initialize ();
#include "HostControlPrv.h"		// Host::Initialize ();
#include "EmScreen.h"			// EmScreen::Initialize ();
#include "ErrorHandling.h"		// Errors::Initialize ();
#include "EmPalmOS.h"			// EmPalmOS::Initialize
#if HAS_TRACER
#include "TracerPlatform.h"		// gTracer.Initialize();
#endif

#include <strstream>			// strstream

EmSession*	gSession;

static uint32	gLastButtonEvent;
const uint32	kButtonEventThreshold = 100;


Bool	PrvCanBotherCPU	(void);
void	PrvWakeUpCPU	(long strID);


/*
	Sub-system methods:

		Startup:
			Called just once when Poser is started.

		Initialize:
			Called just once when a session is created.  Will be followed
			by one or more Reset or Load calls.

		Reset:
			Called any time the Reset menu item is selected or the SysReset
			function is called.

		Save:
			Called to save the state to a file.  May be called when the user
			selects the Save menu item, when the user closes the session,
			or as part of a Gremlin Horde auto-saving sequence.

		Load:
			Called to restore the saved state from a file.  Can assume that
			Initialize has been called first.

		Dispose:
			Called just once when a session is closed.  May be called on a
			partially constructed session, so Dispose methods should be
			prepared to handle NULL pointers, etc.

		Shutdown:
			Called just once when Poser quits.
*/


// ---------------------------------------------------------------------------
//		 EmSession::EmSession
// ---------------------------------------------------------------------------
// EmSession constructor.  Initialize data members and point the global
// "current EmSession" pointer to us.  This method does not explicitly
// throw any exceptions, and really shouldn't fail unless we've exhausted
// the free store.

EmSession::EmSession (void) :
	fConfiguration (),
	fFile (),
	fCPU (NULL),
#if HAS_OMNI_THREAD
	fThread (NULL),
	fSharedLock (),
	fSharedCondition (&fSharedLock),
	fSleepLock (),
	fSleepCondition (&fSleepLock),
	fStop (false),
#endif
	fSuspendState (),
	fState (kStopped),
	fTitle (),
	fMessage (),
	fButtonOptions (0),
	fReadParameters (false),
	fDialogResult (0),
	fBreakOnSysCall (false),
	fBreakOnIdle (false),
	fNestLevel (0),
	fReset (false),
	fResetBanks (false),
	fHordeAutoSaveState (false),
	fHordeSaveRootState (false),
	fHordeSaveSuspendState (false),
	fHordeLoadRootState (false),
	fHordeNextGremlinFromRootState (false),
	fHordeNextGremlinFromSuspendState (false),
	fHardwareReset (false),
	fButtonQueue (),
	fKeyQueue (),
	fPenQueue (),
	fLastPenEvent (EmPoint (-1, -1), false)
{
	fSuspendState.fAllCounters = 0;

	assert (gSession == NULL);
	gSession = this;
}


// ---------------------------------------------------------------------------
//		 EmSession::~EmSession
// ---------------------------------------------------------------------------
// EmSession destructor.  Stop the CPU thread and release all resources.  
// Clear the global "current EmSession" pointer.  This method really shouldn't
// throw any exceptions or in any other way fail.

EmSession::~EmSession (void)
{
	this->DestroyThread ();
	this->Dispose ();

	// Delete the CPU object here instead of in Dispose.  When reloading a
	// saved state as part of a Gremlin Horde procedure,  EmSession::Load
	// is called.  That method calls Dispose, Initialize, and then the
	// various Load methods of all the subsystems.  This process is
	// performed while the current session and cpu objects are active (that
	// is, functions belonging to them are on the stack).  Since Dispose is
	// called, it would be unwise for Dispose to delete the active cpu object.

	delete fCPU;
	fCPU = NULL;

	assert (gSession == this);
	gSession = NULL;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::CreateNew
// ---------------------------------------------------------------------------
// Create a new session based on the given configuration.  This method can
// throw an exception if the creation attemp fails (for instance, the
// configuration is invalid, or if the free store is exhausted).

void EmSession::CreateNew (const Configuration& cfg)
{
	this->Initialize (cfg);
	this->Reset (true);
}


// ---------------------------------------------------------------------------
//		 EmSession::CreateOld
// ---------------------------------------------------------------------------
// Create a new session based on information stored in the given disk file.
// If possible, the entire previously-saved state is restored.  However, if
// any critical information is missing from the file, the emulated state may
// be reset so that it can be used (as if the user had used a pin to reset
// an actual device).  If the file is corrupted or missing even the basic
// saved-state information, and exception is thrown.

void EmSession::CreateOld (const EmFileRef& ref)
{
	EmStreamFile	stream (ref, kOpenExistingForRead);
	ChunkFile		chunkFile (stream);
	SessionFile		sessionFile (chunkFile);

	// Load enough information so that we can initialize the system.

	Configuration	cfg;
	if (!sessionFile.ReadConfiguration (cfg))
	{
		Errors::Throw (kError_InvalidSessionFile);
	}

	this->Initialize (cfg);

	// Now load the saved state.

	this->Load (sessionFile);

	// Remember who we are.

	fFile = ref;
}


// ---------------------------------------------------------------------------
//		 EmSession::CreateBound
// ---------------------------------------------------------------------------
// Create a new session based on information stored in resources or other data
// attached to the emulator executable.

void EmSession::CreateBound (void)
{
	Chunk	psf;
	Platform::GetPSFResource (psf);

	EmStreamBlock	stream (psf.GetPointer (), psf.GetLength ());
	ChunkFile		chunkFile (stream);
	SessionFile 	sessionFile (chunkFile);


	// Load enough information so that we can initialize the system.

	Configuration	cfg;
	if (!sessionFile.ReadConfiguration (cfg))
	{
		Errors::Throw (kError_InvalidSessionFile);
	}

	this->Initialize (cfg);

	// Now load the saved state.

	this->Load (sessionFile);
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::Initialize
// ---------------------------------------------------------------------------
// Called by the various CreateFoo methods to initialize our state according
// to the given configuration.  Also calls the Reset method.

void EmSession::Initialize (const Configuration& cfg)
{
	// Set the hardware device here.  We need to have it set up now
	// for things like EmBankSRAM::Initialize and Memory::Initialize,
	// which need to know what mode they're in.

	fConfiguration = cfg;

	// Ideally, we can initialize sub-systems in any order.

	assert (!fCPU);
	fCPU = new EmCPU68K (this);

	// If ROM is an embedded resouce, use it.

	if (::IsBound ())
	{
		Chunk	rom;
		Bool	resourceLoaded = Platform::GetROMResource (rom);
		assert (resourceLoaded);

		EmStreamChunk	stream (rom);
		Memory::Initialize (stream, cfg.fRAMSize);
	}

	// If ROM is not embedded, use the filespec.

	else
	{
		EmStreamFile	stream (cfg.fROMFile, kOpenExistingForRead);
		Memory::Initialize (stream, cfg.fRAMSize);
	}

	Platform::Initialize ();
	Debug::Initialize ();
	Host::Initialize ();
	EmScreen::Initialize ();
	Errors::Initialize ();
#if HAS_TRACER
	gTracer.Initialize();
#endif

	EmPalmOS::Initialize ();
}


// ---------------------------------------------------------------------------
//		 EmSession::Dispose
// ---------------------------------------------------------------------------

void EmSession::Dispose (void)
{
	// Ideally, we can dispose sub-systems in any order.  However,
	// it's probably a good idea to dispose them in the reverse
	// order in which they were initialized.
	//
	// Note also that this is called from the destructor, which could
	// be called on an object that was not completely initialized.
	// Therefore, each Dispose method should be prepared to handle
	// NULL pointers, etc.

	EmPalmOS::Dispose ();

#if HAS_TRACER
	gTracer.Dispose();
#endif
	Errors::Dispose ();
	EmScreen::Dispose ();
	Host::Dispose ();
	Debug::Dispose ();
	Platform::Dispose ();

	Memory::Dispose ();

	fInstructionBreakFuncs.clear ();
	fDataBreakFuncs.clear ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::Reset
// ---------------------------------------------------------------------------

void EmSession::Reset (Bool hardwareReset)
{
	assert (fNestLevel == 0);

	// !!! Need to re-establish any pressed buttons.

	// Ideally, we can reset sub-systems in any order.	However,
	// it's probably a good idea to reset them in the same order
	// in which they were initialized.

	// Reset Memory *before* CPU, as the CPU looks at memory location
	// zero to find it's reset address.  Memory can't be accessed until
	// the first call to Memory::ResetBankHandlers, which occurs at the
	// end of Memory::Reset.

	Memory::Reset (hardwareReset);

	assert (fCPU);
	fCPU->Reset (hardwareReset);

	Platform::Reset ();
	Debug::Reset ();
	Host::Reset ();
	EmScreen::Reset ();
	Errors::Reset ();
#if HAS_TRACER
	gTracer.Reset();
#endif

	EmPalmOS::Reset ();


#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	// Now reset self.

//	fSuspendState.fCounters.fSuspendByUIThread			= 0;
	fSuspendState.fCounters.fSuspendByDebugger			= 0;
	fSuspendState.fCounters.fSuspendByExternal			= 0;
//	fSuspendState.fCounters.fSuspendByTimeout			= 0;
	fSuspendState.fCounters.fSuspendBySysCall			= 0;
	fSuspendState.fCounters.fSuspendByIdle				= 0;
	fSuspendState.fCounters.fSuspendBySubroutineReturn	= 0;

	fBreakOnSysCall = false;
	fBreakOnIdle = false;
	fNestLevel = 0;

	fReset = false;
	fResetBanks = false;
	fHordeAutoSaveState = false;
	fHordeSaveRootState = false;
	fHordeSaveSuspendState = false;
	fHordeLoadRootState = false;
	fHordeNextGremlinFromRootState = false;
	fHordeNextGremlinFromSuspendState = false;

	fHardwareReset = false;

	fButtonQueue.Clear ();
	fKeyQueue.Clear ();
	fPenQueue.Clear ();

	fLastPenEvent = EmPenEvent (EmPoint (-1, -1), false);

	// All of meta-memory gets wiped out on reset; re-establish these.

	this->InstallInstructionBreaks ();
	this->InstallDataBreaks ();
}


// ---------------------------------------------------------------------------
//		 EmSession::Save
// ---------------------------------------------------------------------------

void EmSession::Save (SessionFile& f)
{
	// Write out the device type.

	assert (fConfiguration.fDeviceType != kDeviceUnspecified);
	f.WriteDeviceType (fConfiguration.fDeviceType);

	// Ideally, we can save sub-systems in any order.  However,
	// it's probably a good idea to save them in the same order
	// in which they were initialized.

	assert (fCPU);
	fCPU->Save (f);
	Memory::Save (f);

	Platform::Save (f);
	Debug::Save (f);
	Host::Save (f);
	EmScreen::Save (f);
	Errors::Save (f);
#if HAS_TRACER
	gTracer.Save (f);
#endif 

	EmPalmOS::Save (f);
}


// ---------------------------------------------------------------------------
//		 EmSession::Load
// ---------------------------------------------------------------------------

void EmSession::Load (SessionFile& f)
{
	// Load the saved state from the session file.	First, set the flag
	// that says whether or not we can successfully restart from the
	// information in this file.  As parts are loaded, the various
	// sub-systems will have a chance to veto this optimistic assumption.

	f.SetCanReload (true);

	// Ideally, we can load sub-systems in any order.  However,
	// it's probably a good idea to load them in the same order
	// in which they were initialized.

	// Load Memory before EmCPU.  That way, the memory system will be
	// initialized when we restore the CPU registers (which will include
	// a sanity check of the PC).

	Memory::Load (f);

	assert (fCPU);
	fCPU->Load (f);

	Platform::Load (f);
	Debug::Load (f);
	Host::Load (f);
	EmScreen::Load (f);
	Errors::Load (f);
#if HAS_TRACER
	gTracer.Load (f);
#endif 

	EmPalmOS::Load (f);

	// If we weren't able to get all the pieces from the file we needed,
	// force a reset.

	if (!f.GetCanReload ())
	{
		this->Reset (true);
		fNeedPostLoad = false;
	}

	// Otherwise, set a flag saying that "post load" activities need to
	// take place.  These activities normally take place while the device
	// is first booted up.  However, any side-effect of those activities
	// may no longer be valid (for instance, we may need to install or
	// remove a 'gdbS' Feature).
	//
	// We set a flag to schedule these activities rather than performing
	// them now, as the emulated state may not be in any shape to respond
	// to them.  For instance, the OS may not be able to handle a call to
	// FtrSet at the moment.  Therefore, we check for this flag at the next
	// convenient moment and take action at that time.

	else
	{
		fNeedPostLoad = true;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::Save
// ---------------------------------------------------------------------------

void EmSession::Save (const EmFileRef& ref, Bool updateFileRef)
{
	EmStreamFile	stream (ref, kCreateOrEraseForUpdate,
						kFileCreatorEmulator, kFileTypeSession);
	ChunkFile		chunkFile (stream);
	SessionFile		sessionFile (chunkFile);

	this->Save (sessionFile);

	if (updateFileRef)
	{
		fFile = ref;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::Load
// ---------------------------------------------------------------------------

void EmSession::Load (const EmFileRef& ref)
{
	EmStreamFile	stream (ref, kOpenExistingForRead);
	ChunkFile		chunkFile (stream);
	SessionFile		sessionFile (chunkFile);

	this->Load (sessionFile);
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::CreateThread
// ---------------------------------------------------------------------------

void EmSession::CreateThread (Bool suspended)
{
#if HAS_OMNI_THREAD
	if (fThread)
		return;
#endif

	// Initialize some variables that control the thread state.

#if HAS_OMNI_THREAD
	fStop				= false;
#endif

	fSuspendState.fAllCounters	= 0;
	fSuspendState.fCounters.fSuspendByUIThread	= suspended ? 1 : 0;
	fState						= kSuspended;

//	LogAppendMsg ("EmSession::CreateThread: fState = %ld", (long) fState);

#if HAS_OMNI_THREAD
	// Create the thread and start it running.

	fThread = new omni_thread (&EmSession::RunStatic, this);
	fThread->start ();
#endif
}


// ---------------------------------------------------------------------------
//		 EmSession::DestroyThread
// ---------------------------------------------------------------------------

void EmSession::DestroyThread (void)
{
#if HAS_OMNI_THREAD
	if (!fThread)
		return;

	omni_mutex_lock	lock (fSharedLock);

	fStop = true;
	fSuspendState.fCounters.fSuspendByUIThread++;
	fSharedCondition.broadcast ();

	while (fState != kStopped)
		fSharedCondition.wait ();

	// fThread thread will quit and destroy itself.

	fThread = NULL;
#endif
}


// ---------------------------------------------------------------------------
//		 EmSession::SuspendThread
// ---------------------------------------------------------------------------
// Suspend the thread in the requested fashion.  Returns true if the attempt
// succeeded and the thread needs to be resumed with a call to ResumeThread.
// Returns false if the thread could not or was not suspended, and
// ResumeThread should not be called.

Bool EmSession::SuspendThread (EmStopMethod how)
{
	if (how == kStopNone)
		return false;

	// Stop on a system call first, so that we can call EvtWakeup.  Stopping
	// on idle will hang if the device is asleep and not idling at all (e.g.,
	// you're in Calculator).
	//
	// Creating an EmSessionStopper recurses into this function, so make sure
	// we do this *before* creating the mutex lock (so that we don't acquire
	// it twice).

	if (how == kStopOnIdle)
	{
		EmSessionStopper	stopper (this, kStopOnSysCall);

		if (!stopper.Stopped ())
			return false;

		::EvtWakeup ();
	}

#if HAS_OMNI_THREAD

	assert (fThread);

	omni_mutex_lock	lock (fSharedLock);

	// Set a flag for the CPU thread to find, telling it how to stop.
	//
	// !!! What to do when fSuspendByUIThread, fSuspendByDebugger or
	// fSuspendByExternal are set, especially if how == kStopOnSysCall or
	// kStopOnIdle?

	Bool	desiredBreakOnSysCall = false;
	Bool	desiredBreakOnIdle = false;

//	LogAppendMsg ("EmSession::SuspendThread (enter): fState = %ld", (long) fState);

	switch (how)
	{
		case kStopNone:
			assert (false);
			break;

		case kStopNow:
			fSuspendState.fCounters.fSuspendByUIThread++;
			break;

		case kStopOnCycle:
			fSuspendState.fCounters.fSuspendByUIThread++;
			break;

		case kStopOnSysCall:
			desiredBreakOnSysCall = true;
			break;

		case kStopOnIdle:
			desiredBreakOnIdle = true;
			break;
	}

	// Get it to a suspended or blocked state, if not there already.
	// Run the CPU loop until we are suspended or blocked on the UI.
	//
	// Note: if we're doing a kStopOnSysCall or kStopOnIdle, and the
	// the thread's already been suspended by, say, kStopNow, then
	// we might want to start the thread up again if it's OK.

	if (fState == kRunning)
	{
		// Wake up the thread if it's sleeping.

		fSleepLock.lock ();
		fSleepCondition.broadcast ();
		fSleepLock.unlock ();

		// Wait for it to stop.
		// !!! Do a timed wait in case we never reach the desired stop point?

		while (fState == kRunning)
		{
			// Establish these inside the while loop.  It's possible for them
			// to get cleared in Reset, and we don't want that.
			//
			// More specifically, we're trying to prevent a condition that can
			// occur when ending a CodeWarrior debug session.  After clicking
			// on the X button in CodeWarrior, we get an RPC packet for
			// SysReset.  In handling that packet, Poser merely sets the
			// PC to the SysReset function address.  When the CPU is
			// restarted, it starts executing that code.  When SysReset
			// gets to the ultimate JSR to the reset vector, Poser notices
			// that and call gSession->Reset(false).  EmSession::Reset
			// clears the fBreakOnSysCall and fBreakOnIdle flags.
			//
			// In the meantime, CodeWarrior has broken the socket connection
			// to us.  Poser responds to that by trying to unregister the
			// 'gdbS' feature.  In order to do that, it must halt the CPU
			// on a system call.  On my PC, the disconnect and call to
			// FtrUnregister can happen *before* we get to this code here.
			//
			// So, here is the UI thread, waiting for a system call to
			// occur.  Before that can happen, the Reset code executes,
			// clearing the fBreakOnSysCall flag!  Thus, the UI thread keeps
			// waiting for a system call to occur, while the CPU thread no
			// longer thinks it needs to halt on that condition.

			fBreakOnSysCall	= desiredBreakOnSysCall;
			fBreakOnIdle	= desiredBreakOnIdle;
			fSharedCondition.broadcast ();

			fSharedCondition.wait ();
//			LogAppendMsg ("EmSession::SuspendThread (waking): fState = %ld", (long) fState);

#ifndef NDEBUG
			if (how == kStopNow || how == kStopOnCycle)
			{
				if (!this->IsNested ())
				{
					assert (fSuspendState.fCounters.fSuspendByUIThread != 0);
				}
			}
#endif
		}
	}

#else

	// Set a flag for the CPU thread to find, telling it how to stop.
	//
	// !!! What to do when fSuspendByUIThread, fSuspendByDebugger or
	// fSuspendByExternal are set, especially if how == kStopOnSysCall or
	// kStopOnIdle?

	switch (how)
	{
		case kStopNone:
			assert (false);
			break;

		case kStopNow:
			fSuspendState.fCounters.fSuspendByUIThread++;
			break;

		case kStopOnCycle:
			fSuspendState.fCounters.fSuspendByUIThread++;
			break;

		case kStopOnSysCall:
			fBreakOnSysCall = true;
			break;

		case kStopOnIdle:
			fBreakOnIdle = true;
			break;
	}

	// Run the CPU loop until we break at the right point.

	fSuspendState.fCounters.fSuspendByTimeout = 0;

	while (fState == kSuspended && fSuspendState.fAllCounters == 0)
	{
		this->ExecuteIncremental ();

		// Ignore this suspend condition.

		fSuspendState.fCounters.fSuspendByTimeout = 0;
	}

#endif

	// Resulting state should be kSuspended or kBlockedOnUI.  It should never
	// be kStopped, because we didn't set the fStop flag.

	assert (fState == kSuspended || fState == kBlockedOnUI);


	// Make sure we stopped the way we wanted to.
	//
	// If appropriate, clear the flag that got us here and say that the UI
	// thread stopped us.  If any other flags remain set, the client needs
	// to deal with that.

	Bool	result = true;	// assume we stopped OK

	switch (how)
	{
		case kStopNone:
			assert (false);
			break;

		case kStopNow:
			// Stopped on either "suspended" or "blocked" is OK.
			result = true;
			break;

		case kStopOnCycle:
			// Only "stopped on suspended" is OK
			result = (fState == kSuspended);
			break;

		case kStopOnSysCall:
			// Must be "stopped on suspended" and on a system call.
			result = (fState == kSuspended) && fSuspendState.fCounters.fSuspendBySysCall;
			if (result)
			{
				fSuspendState.fCounters.fSuspendByUIThread++;
			}
			break;

		case kStopOnIdle:
			// Must be "stopped on suspended" and on an idle call.
			result = (fState == kSuspended) && fSuspendState.fCounters.fSuspendByIdle;
			if (result)
			{
				fSuspendState.fCounters.fSuspendByUIThread++;
			}
			break;
	}

	fBreakOnSysCall = false;
	fBreakOnIdle = false;

	if (result)
	{
		assert (fSuspendState.fCounters.fSuspendByUIThread > 0);
		assert (fSuspendState.fCounters.fSuspendBySubroutineReturn == 0);
		assert (fNestLevel == 0 || fState == kBlockedOnUI);	// (If blocked on UI, fNestLevel may be > 0).
	}

//	LogAppendMsg ("EmSession::SuspendThread (exit): fState = %ld", (long) fState);

	return result;
}


// ---------------------------------------------------------------------------
//		 EmSession::ResumeThread
// ---------------------------------------------------------------------------

void EmSession::ResumeThread (void)
{
#if HAS_OMNI_THREAD
	assert (fThread);

	omni_mutex_lock	lock (fSharedLock);
#endif

//	LogAppendMsg ("EmSession::ResumeThread (enter): fState = %ld", (long) fState);

	if (fSuspendState.fCounters.fSuspendByUIThread > 0)
	{
		--fSuspendState.fCounters.fSuspendByUIThread;

		if (fSuspendState.fCounters.fSuspendByUIThread == 0 &&
			fSuspendState.fCounters.fSuspendByExternal == 0)
		{
			fSuspendState.fCounters.fSuspendBySysCall = 0;
			fSuspendState.fCounters.fSuspendByIdle = 0;
		}

#if HAS_OMNI_THREAD
		if (fSuspendState.fAllCounters == 0)
		{
			// Don't change the state if it's kBlockedOnUI.
			if (fState == kSuspended)
			{
				fState = kRunning;
			}
		}

		fSharedCondition.broadcast ();
#endif
	}

//	LogAppendMsg ("EmSession::ResumeThread (exit): fState = %ld", (long) fState);
}


// ---------------------------------------------------------------------------
//		 EmSession::Sleep
// ---------------------------------------------------------------------------

#if HAS_OMNI_THREAD
void EmSession::Sleep (unsigned long msecs)
{
	const unsigned long	kMillisecondsPerSecond = 1000;
	const unsigned long	kNanosecondsPerMillisecond = 1000000;

	unsigned long	secs = msecs / kMillisecondsPerSecond;
	unsigned long	nsecs = (msecs % kMillisecondsPerSecond) * kNanosecondsPerMillisecond;

	fThread->get_time (&secs, &nsecs, secs, nsecs);

	fSleepLock.lock ();
	fSleepCondition.timedwait (secs, nsecs);
	fSleepLock.unlock ();
}
#endif


// ---------------------------------------------------------------------------
//		 EmSession::InCPUThread
// ---------------------------------------------------------------------------

#if HAS_OMNI_THREAD
Bool EmSession::InCPUThread (void) const
{
	return (omni_thread::self () == fThread);
}
#endif


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::GetSessionState
// ---------------------------------------------------------------------------

EmSessionState EmSession::GetSessionState (void) const
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	return fState;
}


// ---------------------------------------------------------------------------
//		 EmSession::GetSuspendState
// ---------------------------------------------------------------------------

EmSuspendState EmSession::GetSuspendState (void) const
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);

	if (this->InCPUThread ())
		assert (fState == kRunning);
	else
		assert ((fNestLevel == 0 && fState != kRunning) || (fNestLevel > 0 && fState == kRunning));
#endif

	return fSuspendState;
}


// ---------------------------------------------------------------------------
//		 EmSession::SetSuspendState
// ---------------------------------------------------------------------------

void EmSession::SetSuspendState (const EmSuspendState& s)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);

	if (this->InCPUThread ())
		assert (fState == kRunning);
	else
		assert ((fNestLevel == 0 && fState != kRunning) || (fNestLevel > 0 && fState == kRunning));
#endif

	fSuspendState = s;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::ExecuteIncremental
// ---------------------------------------------------------------------------

void EmSession::ExecuteIncremental (void)
{
//	LogAppendMsg ("EmSession::ExecuteIncremental (enter): fState = %ld", (long) fState);

	fSuspendState.fCounters.fSuspendByTimeout = 0;

	assert (fState == kSuspended || fState == kBlockedOnUI);
	assert (fSuspendState.fCounters.fSuspendByTimeout == 0);
	assert (fNestLevel == 0);

	// Enter the CPU loop only if it's not suspended.  The thread could
	// be suspended if fSuspendByUIThread, fSuspendByDebugger or
	// fSuspendByExternal are true.

	if (fState == kBlockedOnUI)
		return;

	if (fSuspendState.fAllCounters == 0)
	{
		try
		{
			this->CallCPU ();
		}
		catch (EmExceptionReset& e)
		{
			e.Display ();
			this->Reset (true);
		}
		catch (...)
		{
			assert (false);
		}

		/*
			Check the reason for EmCPU::Execute returning:

			fSuspendByUIThread
				Should not happen.  That flag is set if the UI thread asks
				the CPU thread to suspend.  ExecuteIncremental is called
				on the Mac with no multiple threads.

			fSuspendByDebugger
				Could happen.  Let it make this function exit.  The debugger
				should already have been notified.

			fSuspendByExternal
				Could happen.  Let it make this function exit.

			fSuspendByTimeout
				Could happen.  Let it make this function exit.

			fSuspendBySysCall
				Could happen.  Let it make this function exit.

			fSuspendByIdle
				Could happen.  Let it make this function exit.

			fSuspendBySubroutineReturn
				Should not happen.  Should occur only on calls to
				ExecuteSubroutine.
		*/

		assert (fSuspendState.fCounters.fSuspendByUIThread == 0);
	}

	assert (fState == kSuspended);
	assert (fSuspendState.fCounters.fSuspendBySubroutineReturn == 0);
	assert (fNestLevel == 0);

//	LogAppendMsg ("EmSession::ExecuteIncremental (exit): fState = %ld", (long) fState);
}


// ---------------------------------------------------------------------------
//		 EmSession::ExecuteSubroutine
// ---------------------------------------------------------------------------

void EmSession::ExecuteSubroutine (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	assert (fNestLevel >= 0);

//	LogAppendMsg ("EmSession::ExecuteSubroutine (enter): fState = %ld", (long) fState);

#if HAS_OMNI_THREAD
	if (this->InCPUThread ())
		assert (fState == kRunning);
	else
		assert ((fNestLevel == 0 && fState != kRunning) || (fNestLevel > 0 && fState == kRunning));
#endif

#ifndef NDEBUG
	EmSuspendCounters	origState = fSuspendState.fCounters;
	Bool				uiFlagChanged = false;
#endif

	EmSuspendCounters	oldState = fSuspendState.fCounters;
	fSuspendState.fAllCounters = 0;

	while (fSuspendState.fAllCounters == 0)
	{
		// Enter new scope so that the omni_mutex_unlock will re-lock
		// the mutex before we look at fSuspendState, etc.
		{
			EmValueChanger<int>	oldNestLevel (fNestLevel, fNestLevel + 1);

#if HAS_OMNI_THREAD
			fSharedCondition.broadcast ();

			omni_mutex_unlock	unlock (fSharedLock);
#endif

			this->CallCPU ();
		}

		/*
			Check the reason for EmCPU::Execute returning:

			fSuspendByUIThread
			fSuspendByExternal
			fSuspendByTimeout
			fSuspendBySysCall
			fSuspendByIdle
				Could happen.  Remember that it occured, clear it, and
				re-enter the CPU loop.  On the way out of this function
				re-establish the request.

			fSuspendByDebugger
				Could happen.  Let it make this function exit.

			fSuspendBySubroutineReturn
				Could happen.  Let it make this function exit.
		*/

#ifndef NDEBUG
		if (fSuspendState.fCounters.fSuspendByUIThread)
		{
			uiFlagChanged = true;
		}
#endif

		oldState.fSuspendByUIThread += fSuspendState.fCounters.fSuspendByUIThread;
		fSuspendState.fCounters.fSuspendByUIThread = 0;

		oldState.fSuspendByDebugger += fSuspendState.fCounters.fSuspendByDebugger;

		oldState.fSuspendByExternal += fSuspendState.fCounters.fSuspendByExternal;
		fSuspendState.fCounters.fSuspendByExternal = 0;

		fSuspendState.fCounters.fSuspendBySysCall = 0;
		fSuspendState.fCounters.fSuspendByIdle = 0;

		oldState.fSuspendByTimeout |= fSuspendState.fCounters.fSuspendByTimeout;
		fSuspendState.fCounters.fSuspendByTimeout = 0;
	}

	fSuspendState.fCounters = oldState;

#ifndef NDEBUG
	if (uiFlagChanged)
	{
		assert (fSuspendState.fCounters.fSuspendByUIThread != origState.fSuspendByUIThread);
		assert (fSuspendState.fCounters.fSuspendByUIThread > 0);
	}
#endif

	// This could have gone negative..._HostSignalWait will decrement the
	// counter as a courtesy.

	if (fSuspendState.fCounters.fSuspendByExternal < 0)
	{
		fSuspendState.fCounters.fSuspendByExternal = 0;
	}

	assert (fNestLevel >= 0);

#if HAS_OMNI_THREAD
	if (this->InCPUThread ())
		assert (fState == kRunning);
	else
		assert ((fNestLevel == 0 && fState != kRunning) || (fNestLevel > 0 && fState == kRunning));

	fSharedCondition.broadcast ();
#endif

#ifndef NDEBUG
	if (uiFlagChanged)
	{
		assert (fSuspendState.fCounters.fSuspendByUIThread != origState.fSuspendByUIThread);
		assert (fSuspendState.fCounters.fSuspendByUIThread > 0);
	}
#endif

//	LogAppendMsg ("EmSession::ExecuteSubroutine (exit): fState = %ld", (long) fState);
}


// ---------------------------------------------------------------------------
//		 EmSession::ExecuteSpecial
// ---------------------------------------------------------------------------

Bool EmSession::ExecuteSpecial (Bool checkForResetOnly)
{
	if (fReset)
	{
		fReset = false;
		fResetBanks = false;

		this->Reset (fHardwareReset);
	}

	if (checkForResetOnly)
		return false;

	if (fResetBanks)
	{
		fResetBanks = false;

		Memory::ResetBankHandlers ();
	}

	if (fHordeAutoSaveState)
	{
		fHordeAutoSaveState = false;

		Hordes::AutoSaveState ();
	}

	if (fHordeSaveRootState)
	{
		assert (!fHordeSaveSuspendState);

		assert (!fHordeLoadRootState);
		assert (!fHordeNextGremlinFromRootState);
		assert (!fHordeNextGremlinFromSuspendState);

		fHordeSaveRootState = false;

		Hordes::SaveRootState ();
	}

	if (fHordeSaveSuspendState)
	{
		assert (!fHordeSaveRootState);

		assert (!fHordeLoadRootState);
		assert (!fHordeNextGremlinFromRootState);
		assert (!fHordeNextGremlinFromSuspendState);

		fHordeSaveSuspendState = false;

		Hordes::SaveSuspendedState ();
	}

	if (fHordeLoadRootState)
	{
		assert (!fHordeSaveSuspendState);
		assert (!fHordeSaveRootState);

		assert (!fHordeNextGremlinFromRootState);
		assert (!fHordeNextGremlinFromSuspendState);

		fHordeLoadRootState = false;

		Hordes::LoadRootState ();
	}

	if (fHordeNextGremlinFromRootState)
	{
		assert (!fHordeSaveSuspendState);
		assert (!fHordeSaveRootState);

		assert (!fHordeLoadRootState);
		assert (!fHordeNextGremlinFromSuspendState);

		fHordeNextGremlinFromRootState = false;

		if (Hordes::LoadRootState () == errNone)
		{
			Hordes::StartGremlinFromLoadedRootState ();
		}
		else
		{
			Hordes::TurnOn (false);
		}
	}

	if (fHordeNextGremlinFromSuspendState)
	{
		assert (!fHordeSaveSuspendState);
		assert (!fHordeSaveRootState);

		assert (!fHordeLoadRootState);
		assert (!fHordeNextGremlinFromRootState);

		fHordeNextGremlinFromSuspendState = false;

		if (Hordes::LoadSuspendedState () == errNone)
		{
			Hordes::StartGremlinFromLoadedSuspendedState ();
		}
		else
		{
			Hordes::TurnOn (false);
		}
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 EmSession::CheckForBreak
// ---------------------------------------------------------------------------
// Check to see if the conditions tell us to break from the CPU Execute loop.

Bool EmSession::CheckForBreak (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	if (fSuspendState.fAllCounters == 0)
		return false;

	// If nested, ignore fSuspendByExternal.  If we're nested and it's
	// non-zero, then that's because someone just made a HostControl call to
	// change it.  We need to let that call finish.  Because we're in a
	// nested call, fSuspendByExternal starts out at zero.  Therefore, a
	// call to suspend the CPU thread will increment it to 1, and a call to
	// resume the thread will decrement it to -1.  We need to preserve those
	// values so that they can be integrated into the the state that was
	// saved in EmSession::ExecuteSubroutine.

	if (this->IsNested ())
	{
		int	old = fSuspendState.fCounters.fSuspendByExternal;
		fSuspendState.fCounters.fSuspendByExternal = 0;

		Bool result = fSuspendState.fAllCounters != 0;

		fSuspendState.fCounters.fSuspendByExternal = old;

		return result;
	}

	return fSuspendState.fAllCounters != 0;
}


// ---------------------------------------------------------------------------
//		 EmSession::CallCPU
// ---------------------------------------------------------------------------
// Wrapper for EmCPU::Execute.  Called by EmSession::ExecuteIncremental,
// EmSession::ExecuteSubroutine, and EmSession::Run.  This wrapper ensures
// that the state is set to "running", calls EmCPU::Execute, and restores the
// state to what it was before this function was called when it exits.

void EmSession::CallCPU (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock					lock (fSharedLock);
#endif

	EmValueChanger<EmSessionState>	oldState (fState, kRunning);

#if HAS_OMNI_THREAD
	omni_mutex_unlock				unlock (fSharedLock);
#endif

	assert (fCPU);
	fCPU->Execute ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::DisplayDialog
// ---------------------------------------------------------------------------
// Schedule an error dialog to be displayed.  If running on a multi-threaded
// implementation and we're not in the UI thread, save the necessary strings
// and other data, and signal the UI thread to use them to display a dialog
// box.  Otherwise, display the dialog box directly.
//
// The result of this function is the button used to dismiss the dialog.  If
// the UI thread is more interested in aborting the CPU thread instead of
// displaying a dialog box, it will cause this function to return -1.

int EmSession::DisplayDialog (	const string& title,
								const string& message,
								int buttonOptions)
{
	fTitle			= title;
	fMessage		= message;
	fButtonOptions	= buttonOptions;
	fReadParameters	= false;
	fDialogResult	= 0;

#if HAS_OMNI_THREAD
	if (this->InCPUThread ())
	{
		omni_mutex_lock	lock (fSharedLock);

//		LogAppendMsg ("EmSession::DisplayDialog (enter): fState = %ld", (long) fState);

		{
		EmValueChanger<EmSessionState>	oldState (fState, kBlockedOnUI);

		// Broadcast the change in fState.

		fSharedCondition.broadcast ();

		while (fDialogResult == 0 && !fStop)
		{
//			LogAppendMsg ("EmSession::DisplayDialog (middle): fState = %ld", (long) fState);
			assert (fState == kBlockedOnUI);
			fSharedCondition.wait ();
		}
		}

		// Broadcast the change in fState.

		fSharedCondition.broadcast ();

//		LogAppendMsg ("EmSession::DisplayDialog (exit): fState = %ld", (long) fState);
	}
	else
#endif

	{
//		LogAppendMsg ("EmSession::DisplayDialog (enter): fState = %ld", (long) fState);
		{
		EmValueChanger<EmSessionState>	oldState (fState, kBlockedOnUI);

		fDialogResult = Platform::CommonDialog (
							fTitle.c_str (),
							fMessage.c_str (),
							fButtonOptions);
		}
//		LogAppendMsg ("EmSession::DisplayDialog (exit): fState = %ld", (long) fState);
	}

	// !!! Throw an exception if fDialogResult == -1.

	return fDialogResult;
}


// ---------------------------------------------------------------------------
//		 EmSession::BeginDialog
// ---------------------------------------------------------------------------

Bool EmSession::BeginDialog (	string& title,
								string& message,
								int& buttonOptions)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	if (fState == kBlockedOnUI && !fReadParameters)
	{
		title			= fTitle;
		message			= fMessage;
		buttonOptions	= fButtonOptions;

		fReadParameters = true;

		return true;
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 EmSession::EndDialog
// ---------------------------------------------------------------------------

void EmSession::EndDialog (int dialogResult)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
	fSharedCondition.broadcast ();
#endif

	fDialogResult = dialogResult;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::PostButtonEvent
//		 EmSession::HasButtonEvent
//		 EmSession::PeekButtonEvent
//		 EmSession::GetButtonEvent
// ---------------------------------------------------------------------------

void EmSession::PostButtonEvent (const EmButtonEvent& event)
{
	if (!::PrvCanBotherCPU())
		return;

	fButtonQueue.Put (event);
}


Bool EmSession::HasButtonEvent (void)
{
	// Don't feed hardware events out too quickly.  Otherwise, the OS
	// may not have time to react to the register changes.

	uint32	now = Platform::GetMilliseconds ();

	if (now - gLastButtonEvent < kButtonEventThreshold)
	{
		return false;
	}

	return fButtonQueue.GetUsed () > 0;
}


EmButtonEvent EmSession::PeekButtonEvent (void)
{
	return fButtonQueue.Peek ();
}


EmButtonEvent EmSession::GetButtonEvent (void)
{
	// Don't feed hardware events out too quickly.  Otherwise, the OS
	// may not have time to react to the register changes.

	gLastButtonEvent = Platform::GetMilliseconds ();

	return fButtonQueue.Get ();
}


// ---------------------------------------------------------------------------
//		 EmSession::PostKeyEvent
//		 EmSession::HasKeyEvent
//		 EmSession::PeekKeyEvent
//		 EmSession::GetKeyEvent
// ---------------------------------------------------------------------------

void EmSession::PostKeyEvent (const EmKeyEvent& event)
{
	if (!::PrvCanBotherCPU())
		return;

	fKeyQueue.Put (event);

	// Wake up the CPU in case it's sleeping so that it can
	// quickly handle the event.

	::PrvWakeUpCPU (kStr_EnterPen);
}


Bool EmSession::HasKeyEvent (void)
{
	return fKeyQueue.GetUsed () > 0;
}


EmKeyEvent EmSession::PeekKeyEvent (void)
{
	return fKeyQueue.Peek ();
}


EmKeyEvent EmSession::GetKeyEvent (void)
{
	return fKeyQueue.Get ();
}


// ---------------------------------------------------------------------------
//		 EmSession::PostPenEvent
//		 EmSession::HasPenEvent
//		 EmSession::PeekPenEvent
//		 EmSession::GetPenEvent
// ---------------------------------------------------------------------------

void EmSession::PostPenEvent (const EmPenEvent& event)
{
	if (!::PrvCanBotherCPU())
		return;

	// If this pen-down event is the same as the last pen-down
	// event, do nothing.

	if (event.fPenIsDown && event == fLastPenEvent)
	{
		return;
	}

	// Add the event to our queue.

	fPenQueue.Put (event);

	// Remember this event for the next time.

	fLastPenEvent = event;

	// Wake up the CPU in case it's sleeping so that it can
	// quickly handle the event.

	::PrvWakeUpCPU (kStr_EnterPen);
}


Bool EmSession::HasPenEvent (void)
{
	return fPenQueue.GetUsed () > 0;
}


EmPenEvent EmSession::PeekPenEvent (void)
{
	return fPenQueue.Peek ();
}


EmPenEvent EmSession::GetPenEvent (void)
{
	return fPenQueue.Get ();
}


// ---------------------------------------------------------------------------
//		 PrvCanBotherCPU
// ---------------------------------------------------------------------------

Bool PrvCanBotherCPU (void)
{
	if (Hordes::IsOn ())
		return false;

	if (!Patches::UIInitialized ())
		return false;

	// !!! Need a check for if the device is sleeping.

	return true;
}


// ---------------------------------------------------------------------------
//		 PrvWakeUpCPU
// ---------------------------------------------------------------------------

void PrvWakeUpCPU (long strID)
{
	// Make sure the app's awake.  Normally, we post events on a patch to
	// SysEvGroupWait.	However, if the Palm device is already waiting,
	// then that trap will never get called.  By calling EvtWakeup now,
	// we'll wake up the Palm device from its nap.

	EmSessionStopper	stopper (gSession, kStopOnSysCall);

	if (stopper.Stopped ())
	{
		Errors::ReportIfPalmError (strID, ::EvtWakeup ());
	}
}

#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::GetConfiguration
// ---------------------------------------------------------------------------

Configuration EmSession::GetConfiguration (void)
{
	return fConfiguration;
}


// ---------------------------------------------------------------------------
//		 EmSession::GetFile
// ---------------------------------------------------------------------------

EmFileRef EmSession::GetFile (void)
{
	return fFile;
}


// ---------------------------------------------------------------------------
//		 EmSession::GetDevice
// ---------------------------------------------------------------------------

EmDevice EmSession::GetDevice (void)
{
	return fConfiguration.fDeviceType;
}


// ---------------------------------------------------------------------------
//		 EmSession::GetBreakOnSysCall
// ---------------------------------------------------------------------------

Bool EmSession::GetBreakOnSysCall (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	return fBreakOnSysCall;
}


// ---------------------------------------------------------------------------
//		 EmSession::GetBreakOnIdle
// ---------------------------------------------------------------------------

Bool EmSession::GetBreakOnIdle (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	return fBreakOnIdle;
}


// ---------------------------------------------------------------------------
//		 EmSession::IsNested
// ---------------------------------------------------------------------------

#if 0	// Inlined
Bool EmSession::IsNested (void)
{
	return fNestLevel > 0;
}
#endif


// ---------------------------------------------------------------------------
//		 EmSession::GetNeedPostLoad
// ---------------------------------------------------------------------------

Bool EmSession::GetNeedPostLoad (void)
{
	return fNeedPostLoad;
}


// ---------------------------------------------------------------------------
//		 EmSession::SetNeedPostLoad
// ---------------------------------------------------------------------------

void EmSession::SetNeedPostLoad (Bool newValue)
{
	fNeedPostLoad = newValue;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::ScheduleSuspendException
//		 EmSession::ScheduleSuspendError
//		 EmSession::ScheduleSuspendExternal
//		 EmSession::ScheduleSuspendTimeout
//		 EmSession::ScheduleSuspendSysCall
//		 EmSession::ScheduleSuspendIdle
//		 EmSession::ScheduleSuspendSubroutineReturn
// ---------------------------------------------------------------------------

void EmSession::ScheduleSuspendException (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendByDebugger++;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendError (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendByDebugger++;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendExternal (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendByExternal++;
	fSuspendState.fCounters.fSuspendBySysCall = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendTimeout (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendByTimeout = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendSysCall (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendBySysCall = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendIdle (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendByIdle = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSuspendSubroutineReturn (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	fSuspendState.fCounters.fSuspendBySubroutineReturn = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleResumeExternal (void)
{
#if HAS_OMNI_THREAD
	omni_mutex_lock	lock (fSharedLock);
#endif

	// Let it go negative.  It needs to do this if we're resuming.
	// See comments in EmCPU68K::CheckForBreak.

//	if (fSuspendState.fCounters.fSuspendByExternal)
	{
		fSuspendState.fCounters.fSuspendByExternal--;
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::ScheduleReset
//		 EmSession::ScheduleResetBanks
//		 EmSession::ScheduleAutoSaveState
//		 EmSession::ScheduleSaveRootState
//		 EmSession::ScheduleSaveSuspendedState
//		 EmSession::ScheduleLoadRootState
//		 EmSession::ScheduleNextGremlinFromRootState
//		 EmSession::ScheduleNextGremlinFromSuspendedState
// ---------------------------------------------------------------------------

void EmSession::ScheduleReset (Bool hardwareReset)
{
	fReset = 1;
	fHardwareReset = hardwareReset;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleResetBanks (void)
{
	fResetBanks = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleAutoSaveState (void)
{
	fHordeAutoSaveState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSaveRootState (void)
{
	fHordeSaveRootState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleSaveSuspendedState (void)
{
	fHordeSaveSuspendState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleLoadRootState (void)
{
	fHordeLoadRootState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleNextGremlinFromRootState (void)
{
	fHordeNextGremlinFromRootState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


void EmSession::ScheduleNextGremlinFromSuspendedState (void)
{
	fHordeNextGremlinFromSuspendState = 1;

	assert (fCPU);
	fCPU->CheckAfterCycle ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::AddInstructionBreakHandlers
// ---------------------------------------------------------------------------

void EmSession::AddInstructionBreakHandlers (	InstructionBreakInstaller f1,
												InstructionBreakRemover f2,
												InstructionBreakReacher f3)
{
	InstructionBreakFuncs	funcs;

	funcs.fInstaller = f1;
	funcs.fRemover = f2;
	funcs.fReacher = f3;

	fInstructionBreakFuncs.push_back (funcs);
}


// ---------------------------------------------------------------------------
//		 EmSession::AddDataBreakHandlers
// ---------------------------------------------------------------------------

void EmSession::AddDataBreakHandlers (	DataBreakInstaller f1,
										DataBreakRemover f2,
										DataBreakReacher f3)
{
	DataBreakFuncs	funcs;

	funcs.fInstaller = f1;
	funcs.fRemover = f2;
	funcs.fReacher = f3;

	fDataBreakFuncs.push_back (funcs);
}


// ---------------------------------------------------------------------------
//		 EmSession::InstallInstructionBreaks
// ---------------------------------------------------------------------------

void EmSession::InstallInstructionBreaks (void)
{
	InstructionBreakFuncList::iterator	iter = fInstructionBreakFuncs.begin ();
	
	while (iter != fInstructionBreakFuncs.end ())
	{
		iter->fInstaller ();

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::RemoveInstructionBreaks
// ---------------------------------------------------------------------------

void EmSession::RemoveInstructionBreaks (void)
{
	InstructionBreakFuncList::iterator	iter = fInstructionBreakFuncs.begin ();
	
	while (iter != fInstructionBreakFuncs.end ())
	{
		iter->fRemover ();

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::HandleInstructionBreak
// ---------------------------------------------------------------------------

void EmSession::HandleInstructionBreak (void)
{
	InstructionBreakFuncList::iterator	iter = fInstructionBreakFuncs.begin ();
	
	while (iter != fInstructionBreakFuncs.end ())
	{
		iter->fReacher ();

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::InstallDataBreaks
// ---------------------------------------------------------------------------

void EmSession::InstallDataBreaks (void)
{
	DataBreakFuncList::iterator	iter = fDataBreakFuncs.begin ();
	
	while (iter != fDataBreakFuncs.end ())
	{
		iter->fInstaller ();

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::RemoveDataBreaks
// ---------------------------------------------------------------------------

void EmSession::RemoveDataBreaks (void)
{
	DataBreakFuncList::iterator	iter = fDataBreakFuncs.begin ();
	
	while (iter != fDataBreakFuncs.end ())
	{
		iter->fRemover ();

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmSession::HandleDataBreak
// ---------------------------------------------------------------------------

void EmSession::HandleDataBreak (emuptr address, int size, Bool forRead)
{
	DataBreakFuncList::iterator	iter = fDataBreakFuncs.begin ();
	
	while (iter != fDataBreakFuncs.end ())
	{
		iter->fReacher (address, size, forRead);

		++iter;
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSession::RunStatic
// ---------------------------------------------------------------------------

#if HAS_OMNI_THREAD

void EmSession::RunStatic (void* arg)
{
	assert (arg);
	((EmSession*) arg)->Run ();
}

#endif


// ---------------------------------------------------------------------------
//		 EmSession::Run
// ---------------------------------------------------------------------------

#if HAS_OMNI_THREAD

void EmSession::Run ()
{
	assert (fCPU);

	// Acquire the lock to the shared variables so that we can check fState.

	omni_mutex_lock	lock (fSharedLock);

	// Loop until we're asked to stop.

//	LogAppendMsg ("EmSession::Run (enter): fState = %ld", (long) fState);

	try
	{
		while (!fStop)
		{
//			LogAppendMsg ("EmSession::Run (top outer loop): fState = %ld", (long) fState);

			// Block while we're suspended.  We set fState to kSuspended so
			// that clients know our state, and unlock fSharedLock so that
			// fSuspendState or fStop can be changed externally.

			if (fSuspendState.fAllCounters)
			{
				while (this->IsNested () || (fSuspendState.fAllCounters && !fStop))
				{
//					LogAppendMsg ("EmSession::Run (top inner loop): fState = %ld", (long) fState);

					// If we were asked to start, our state will have been set to kRunning.
					// If we were asked to stop before we could actually start running,
					// we need to set it to kSuspended again.

					if (!this->IsNested ())
					{
						fState = kSuspended;
						fSharedCondition.broadcast ();
					}

					fSharedCondition.wait ();
				}

//				LogAppendMsg ("EmSession::Run (after inner loop): fState = %ld", (long) fState);

				if (fStop)
				{
					continue;
				}
			}

//			LogAppendMsg ("EmSession::Run (after if): fState = %ld", (long) fState);

			assert (fSuspendState.fAllCounters == 0);
			assert (fNestLevel == 0);
			assert (fState == kRunning);

			// We're no longer suspended.  Release our shared globals.

			fSharedLock.unlock ();

			// Execute the "fetch an opcode and emulate it" loop.  This
			// function returns only if requested or an error occurs.

			try
			{
				this->CallCPU ();
			}
			catch (EmExceptionReset& e)
			{
				e.Display ();
				this->Reset (true);
			}
			catch (...)
			{
				assert (false);
			}

			fSharedLock.lock ();

//			LogAppendMsg ("EmSession::Run (after CallCPU): fState = %ld", (long) fState);

			assert (fState == kRunning);
			assert (fNestLevel == 0);
		}
	}
	catch (...)
	{
		// We don't actually know if fSharedLock is acquired or
		// not on this exception.  Hopefully, its being unlocked
		// if it's already unlocked is OK.
	}

	// fStop is set to true
	// fSharedLock is locked

	fState = kStopped;
	fSharedCondition.broadcast ();
}

#endif


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmSessionStopper::EmSessionStopper
// ---------------------------------------------------------------------------

EmSessionStopper::EmSessionStopper (EmSession* cpu, EmStopMethod how) :
	fSession (cpu),
	fHow (how),
	fStopped (false)
{
	if (fSession)
	{
		fStopped = fSession->SuspendThread (how);
	}
}


// ---------------------------------------------------------------------------
//		 EmSessionStopper::~EmSessionStopper
// ---------------------------------------------------------------------------

EmSessionStopper::~EmSessionStopper (void)
{
	if (fSession && fStopped)
	{
		fSession->ResumeThread ();
	}
}


// ---------------------------------------------------------------------------
//		 EmSessionStopper::Stopped
// ---------------------------------------------------------------------------

Bool EmSessionStopper::Stopped (void)
{
	return fStopped;
}


// ---------------------------------------------------------------------------
//		 EmSessionStopper::CanCall
// ---------------------------------------------------------------------------

Bool EmSessionStopper::CanCall (void)
{
	return	(fSession != NULL) &&
			(fHow == kStopOnSysCall || fHow == kStopOnIdle) &&
			(fStopped /*== true*/);
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmExceptionReset::EmExceptionReset
// ---------------------------------------------------------------------------

EmExceptionReset::EmExceptionReset (void) :
	EmException (),
	fWhat (),
	fMessage (),
	fException (0),
	fTrapWord (0)
{
}

	
// ---------------------------------------------------------------------------
//		 EmExceptionReset::~EmExceptionReset
// ---------------------------------------------------------------------------

EmExceptionReset::~EmExceptionReset (void)
{
}


// ---------------------------------------------------------------------------
//		 EmExceptionReset::what
// ---------------------------------------------------------------------------

const char* EmExceptionReset::what (void) const
{
	if (fWhat.empty ())
	{
		strstream	buffer;

		string	trapName (::GetTrapName (fTrapWord));

		if (fException)
		{
			buffer
				<< "Hardware exception #"
				<< fException
				<< " occured while the emulator was calling the Palm OS function \""
				<< trapName
				<< "\".";
		}
		else if (!fMessage.empty ())
		{
			buffer
				<< fMessage
				<< "\n\n"
				<< "The error occured while the emulator was calling the Palm OS function \""
				<< trapName
				<< "\".";
		}
		else
		{
			buffer
				<< "Something funky happened.";
		}

		buffer << "  The emulator is now in an unstable state and will reset.";

		fWhat = string (buffer.str (), buffer.pcount ());
	}

	return fWhat.c_str ();
}


// ---------------------------------------------------------------------------
//		 EmExceptionReset::Display
// ---------------------------------------------------------------------------

void EmExceptionReset::Display (void) const
{
	Errors::DoDialog (NULL, this->what (), Errors::kReset);
}
