/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#ifndef _EMULATORTYPES_H_
#define _EMULATORTYPES_H_

#include "EmFileRef.h"			// EmFileRef
#include "EmDevice.h"			// DeviceType

#include <deque>
#include <exception>
#include <map>
#include <string>
#include <vector>


// ======================================
// ========== ENUMERATED TYPES ==========
// ======================================

// ---------- <Less of a type, more of a constant...> ----------

enum
{
	MRU_COUNT = 9
};


// ---------- CloseActionType ----------

enum CloseActionType
{
	kSaveAlways,
	kSaveAsk,
	kSaveNever
};
typedef vector<CloseActionType>	CloseActionList;


// ---------- LoggingType ----------

enum LoggingType
{
	kNormalLogging = 0x01,
	kGremlinLogging = 0x02,
	kAllLogging = kNormalLogging | kGremlinLogging
};
typedef vector<LoggingType>	LoggingList;


// =====================================
// ========== SIMPLE TYPEDEFS ==========
// =====================================

typedef uint8					ScaleType;
typedef int32					RAMSizeType;
typedef int32					EmCoord;

typedef vector<ScaleType>		ScaleList;
typedef vector<RAMSizeType>		RAMSizeList;

typedef vector<unsigned char>	ByteList;
typedef vector<string>			StringList;

typedef map<string, string>		StringStringMap;



// ================================
// ========== STRUCTURES ==========
// ================================

// ---------- RGBType ----------

struct RGBType
{
	RGBType (void) {}
	RGBType (uint8 red, uint8 green, uint8 blue) :
		fRed (red),
		fGreen (green),
		fBlue (blue)
	{}

	bool operator== (const RGBType& rhs) const
	{
		return fRed == rhs.fRed && fGreen == rhs.fGreen && fBlue == rhs.fBlue;
	}

	bool operator!= (const RGBType& rhs) const
	{
		return !(*this == rhs);
	}

	uint8		fRed;
	uint8		fGreen;
	uint8		fBlue;
	uint8		fFiller;	// Pad this out to 4 bytes to make array lookups more efficient.
};
typedef vector<RGBType>	RGBList;


// ---------- SystemCallContext ----------

struct SystemCallContext
{
	emuptr		fPC;			// PC at which the system call was made.
	emuptr		fNextPC;		// PC of instruction after system call.
	emuptr		fDestPC;		// Address of system function.
	uint16		fTrapWord;		// Trapword, even for SYSTRAP_FAST calls.
	uint16		fTrapIndex;		// Trapword with the high 4 bits removed.
	emuptr		fExtra;			// RefNum for library calls, D2 for "dispatch" calls.
	Bool		fViaTrap;		// True if called via TRAP $F.
	Bool		fViaJsrA1;		// True if called via SYSTRAP_FASTER.

	long		fError;			// If an error occurred getting the context, error is here
	long		fLibIndex;		// Resource base number for function name
	long		fMaxRefNum;		// If the refNum was too big, this is the max value it could be.
};
typedef vector<SystemCallContext>	SystemCallContextList;


// ---------- DatabaseInfo ----------

struct DatabaseInfo
{
	UInt32		creator;
	UInt32		type;
	UInt16		version;
	LocalID		dbID;
	UInt16 		cardNo;
	UInt32		modDate;
	UInt16		dbAttrs;

	char		name[dmDBNameLength];
	char		dbName[dmDBNameLength];

	Bool operator < (const DatabaseInfo& other) const
		{ if (creator < other.creator) return true;
		  if (creator > other.creator) return false;
		  if (type < other.type) return true;
		  return false; }

	Bool operator == (const DatabaseInfo& other) const
		{ return (creator == other.creator) &&
				 (type == other.type); }
};
typedef vector<DatabaseInfo>	DatabaseInfoList;


// ---------- SlotInfoType ----------

struct SlotInfoType
{
	long		fSlotNumber;
	Bool		fSlotOccupied;
	EmDirRef	fSlotRoot;
};
typedef vector<SlotInfoType>	SlotInfoList;


// ---------- Configuration ----------

struct Configuration
{
	Configuration () :
		fDeviceType (kDeviceUnspecified),
		fRAMSize (1024),
		fROMFile ()
	{
	}

	Configuration (DeviceType dt, long size, const EmFileRef& rom) :
		fDeviceType (dt),
		fRAMSize (size),
		fROMFile (rom)
	{
	}

	DeviceType				fDeviceType;
	RAMSizeType				fRAMSize;
	EmFileRef				fROMFile;
};
typedef vector<Configuration>	ConfigurationList;


// ---------- GremlinInfo ----------

struct GremlinInfo
{
	GremlinInfo () :
		fNumber (0),
		fSteps (-1),
		fSaveFrequency (10000),
		fAppList ()
	{
	}

	long				fNumber;
	long				fSteps;
	long				fSaveFrequency;
	DatabaseInfoList	fAppList;
};
typedef vector<GremlinInfo>	GremlinInfoList;


// ---------- HordeInfo ----------

struct HordeInfo
{
	HordeInfo () :
		fStartNumber (0),
		fStopNumber (0),
		fDepthSwitch (30000),
		fDepthSave (10000),
		fDepthStop (1000000),
		fCanSwitch (false),
		fCanSave (false),
		fCanStop (false),
		fAppList ()

	{
		NewToOld ();
	}

	long				fStartNumber;
	long				fStopNumber;

	long				fDepthSwitch;
	long				fDepthSave;
	long				fDepthStop;

	Bool				fCanSwitch;
	Bool				fCanSave;
	Bool				fCanStop;

	DatabaseInfoList	fAppList;

	// Old fields that I want to get rid of, but
	// I need to update Gremlins and Hordes first.

	long				fSaveFrequency;
	long				fSwitchDepth;
	long				fMaxDepth;

	void NewToOld (void)
	{
		fSwitchDepth	= fCanSwitch ? fDepthSwitch : -1;
		fSaveFrequency	= fCanSave ? fDepthSave : 0;
		fMaxDepth		= fCanStop ? fDepthStop : -1;
	}

	void OldToNew (void)
	{
		fCanSwitch = fSwitchDepth > 0;
		fDepthSwitch = fCanSwitch ? fSwitchDepth : 30000;

		fCanSave = fSaveFrequency > 0;
		fDepthSave = fCanSave ? fSaveFrequency : 10000;

		fCanStop = fMaxDepth > 0;
		fDepthStop = fCanStop ? fMaxDepth : 1000000;
	}
};
typedef vector<HordeInfo>	HordeInfoList;


// ---------- SysLibTblEntryTypeV10 ----------

typedef struct SysLibTblEntryTypeV10 {
	MemPtr*			dispatchTblP;					// pointer to library dispatch table
	void*			globalsP;						// Library globals
	} SysLibTblEntryTypeV10;
typedef SysLibTblEntryTypeV10*	SysLibTblEntryV10Ptr;



// ---------- EmException ----------

class EmException : public exception
{
	public:
								EmException () : exception () {}
		virtual					~EmException () {}
};


// ---------- EmExceptionReset ----------

class EmExceptionReset : public EmException
{
	public:
								EmExceptionReset	(void);
		virtual					~EmExceptionReset	(void);

		void					SetException		(uint16 e)		{ fException = e; }
		void					SetMessage			(const char* m)	{ fMessage = m; }
		void					SetTrapWord			(uint16 tw)		{ fTrapWord = tw; }

		virtual const char*		what				(void) const;

		void					Display				(void) const;

	private:
		mutable string			fWhat;
		string					fMessage;
		uint16					fException;
		uint16					fTrapWord;
};


// ============================================
// ========== Windows socket mapping ==========
// ============================================

#if !PLATFORM_WINDOWS
	#include <sys/types.h>
	#include <sys/socket.h>			// socklen_t

	#if PLATFORM_UNIX && !defined(HAVE_TYPE_SOCKLEN_T)
		typedef int			socklen_t;	// not defined in <sys/socket.h> on solaris or FreeBSD
	#endif

	typedef int				SOCKET;

	#define WSAEINVAL		EINVAL
	#define INVALID_SOCKET	((SOCKET) (~0))
	#define SOCKET_ERROR	(-1)
	#define closesocket		close
	#define ioctlsocket		ioctl

	#ifndef INADDR_LOOPBACK
	#define INADDR_LOOPBACK	0x7f000001
	#endif
#else
	typedef int				socklen_t;
#endif


#endif	// _EMULATORTYPES_H_
