/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

// ---------------------------------------------------------------------------
#pragma mark ===== Includes
// ---------------------------------------------------------------------------

#include "EmCommon.h"
#include "EmBankDRAM.h"

#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "EmBankROM.h"			// IsPCInRAM (implicitly, through META_CHECK)
#include "EmBankSRAM.h"			// gRAMBank_Size, gRAM_Memory, gMemoryAccess
#include "EmCPU68K.h"			// ProcessException
#include "EmHAL.h"				// EmHAL
#include "EmMemory.h"			// Memory::InitializeBanks
#include "EmScreen.h"			// EmScreen::MarkDirty
#include "MetaMemory.h"			// MetaMemory
#include "TrapPatches.h"		// META_CHECK calls Patches::IsPCInMemMgr


// ---------------------------------------------------------------------------
#pragma mark ===== Types
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
#pragma mark ===== Functions
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
#pragma mark ===== Constants
// ---------------------------------------------------------------------------

const uint32	kMemoryStart = 0x00000000;
const emuptr	kGlobalStart = offsetof (LowMemHdrType, globals);


// ---------------------------------------------------------------------------
#pragma mark ===== Variables
// ---------------------------------------------------------------------------

// ----- Saved variables -----------------------------------------------------

static uint32	gDynamicHeapSize;


// ----- UnSaved variables ---------------------------------------------------

static AddressBank	gAddressBank =
{
	EmBankDRAM::GetLong,
	EmBankDRAM::GetWord,
	EmBankDRAM::GetByte,
	EmBankDRAM::SetLong,
	EmBankDRAM::SetWord,
	EmBankDRAM::SetByte,
	EmBankDRAM::GetRealAddress,
	EmBankDRAM::ValidAddress,
	EmBankDRAM::GetMetaAddress,
	EmBankDRAM::AddOpcodeCycles
};


// ---------------------------------------------------------------------------
#pragma mark ===== Inlines
// ---------------------------------------------------------------------------

static inline int InlineValidAddress (emuptr address, uint32 size)
{
	int	result = (address + size) <= gRAMBank_Size;

	return result;
}


static inline uint8* InlineGetRealAddress (emuptr address)
{
	return (uint8*) &(gRAM_Memory[address]);
}


static inline uint8* InlineGetMetaAddress (emuptr address)
{
	return (uint8*) &(gRAM_MetaMemory[address]);
}


static inline void ScreenCheck (uint8* metaAddress, emuptr address, uint32 size)
{
#if defined (macintosh)

	if (size == 1 && !MetaMemory::IsScreenBuffer8 (metaAddress))
		return;

	if (size == 2 && !MetaMemory::IsScreenBuffer16 (metaAddress))
		return;

	if (size == 4 && !MetaMemory::IsScreenBuffer32 (metaAddress))
		return;

#else

	if (MetaMemory::IsScreenBuffer (metaAddress, size))

#endif

	{
		EmScreen::MarkDirty (address, size);
	}
}


#pragma mark -

// ===========================================================================
//		 DRAM Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's random
// access memory.

/***********************************************************************
 *
 * FUNCTION:	EmBankDRAM::Initialize
 *
 * DESCRIPTION: Standard initialization function.  Responsible for
 *				initializing this sub-system when a new session is
 *				created.  Will be followed by at least one call to
 *				Reset or Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmBankDRAM::Initialize (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmBankDRAM::Reset
 *
 * DESCRIPTION:	Standard reset function.  Sets the sub-system to a
 *				default state.  This occurs not only on a Reset (as
 *				from the menu item), but also when the sub-system
 *				is first initialized (Reset is called after Initialize)
 *				as well as when the system is re-loaded from an
 *				insufficient session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmBankDRAM::Reset (Bool /*hardwareReset*/)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmBankDRAM::Save
 *
 * DESCRIPTION:	Standard save function.  Saves any sub-system state to
 *				the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmBankDRAM::Save (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmBankDRAM::Load
 *
 * DESCRIPTION:	Standard load function.  Loads any sub-system state
 *				from the given session file.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmBankDRAM::Load (SessionFile&)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmBankDRAM::Dispose
 *
 * DESCRIPTION:	Standard dispose function.  Completely release any
 *				resources acquired or allocated in Initialize and/or
 *				Load.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmBankDRAM::Dispose (void)
{
}


/***********************************************************************
 *
 * FUNCTION:    EmBankDRAM::SetBankHandlers
 *
 * DESCRIPTION: Set the bank handlers UAE uses to dispatch memory
 *				access operations.
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

void EmBankDRAM::SetBankHandlers (void)
{
	// First few memory banks are managed by the functions in EmBankDRAM.

	gDynamicHeapSize = EmHAL::GetDynamicHeapSize ();
	
	if (gDynamicHeapSize > gRAMBank_Size)
		gDynamicHeapSize = gRAMBank_Size;

	uint32	sixtyFourK	= 64 * 1024L;
	uint32	numBanks	= (gDynamicHeapSize + sixtyFourK - 1) / sixtyFourK;

	Memory::InitializeBanks (	gAddressBank,
								bankindex (kMemoryStart),
								numBanks);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::GetLong
// ---------------------------------------------------------------------------

uint32 EmBankDRAM::GetLong (emuptr address)
{
	if (address > gDynamicHeapSize)
		return EmBankSRAM::GetLong (address);

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMLongRead]++;
	if (address & 2)
		gMemoryAccess[kDRAMLongRead2]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (address & 1) != 0)
	{
		AddressError (address, sizeof (uae_u32), true);
	}

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, GetLong, uae_u32, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!InlineValidAddress (address, sizeof (uae_u32)))
	{
		InvalidAccess (address, sizeof (uae_u32), true);
	}

#if (HAS_PROFILING)
	CYCLE_GETLONG (WAITSTATES_DRAM);
#endif

	return do_get_mem_long (gRAM_Memory + address);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::GetWord
// ---------------------------------------------------------------------------

uint32 EmBankDRAM::GetWord (emuptr address)
{
	if (address > gDynamicHeapSize)
		return EmBankSRAM::GetWord (address);

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMWordRead]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (address & 1) != 0)
	{
		AddressError (address, sizeof (uae_u16), true);
	}

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, GetWord, uae_u16, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!InlineValidAddress (address, sizeof (uae_u16)))
	{
		InvalidAccess (address, sizeof (uae_u16), true);
	}

#if (HAS_PROFILING)
	CYCLE_GETWORD (WAITSTATES_DRAM);
#endif

	return do_get_mem_word (gRAM_Memory + address);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::GetByte
// ---------------------------------------------------------------------------

uint32 EmBankDRAM::GetByte (emuptr address)
{
	if (address > gDynamicHeapSize)
		return EmBankSRAM::GetByte (address);

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMByteRead]++;
#endif

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, GetByte, uae_u8, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!InlineValidAddress (address, sizeof (uae_u8)))
	{
		InvalidAccess (address, sizeof (uae_u8), true);
	}

#if (HAS_PROFILING)
	CYCLE_GETBYTE (WAITSTATES_DRAM);
#endif

	return do_get_mem_byte (gRAM_Memory + address);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::SetLong
// ---------------------------------------------------------------------------

void EmBankDRAM::SetLong (emuptr address, uint32 value)
{
	if (address > gDynamicHeapSize)
	{
		EmBankSRAM::SetLong (address, value);
		return;
	}

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMLongWrite]++;
	if (address & 2)
		gMemoryAccess[kDRAMLongWrite2]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (address & 1) != 0)
	{
		AddressError (address, sizeof (uae_u32), false);
	}

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, SetLong, uae_u32, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!InlineValidAddress (address, sizeof (uae_u32)))
	{
		InvalidAccess (address, sizeof (uae_u32), false);
	}

	ScreenCheck (metaAddress, address, sizeof (uae_u32));

#if (HAS_PROFILING)
	CYCLE_PUTLONG (WAITSTATES_DRAM);
#endif

	do_put_mem_long (gRAM_Memory + address, value);

#if FOR_LATER
	// Mark that this memory location can now be read from.

	MetaMemory::MarkLongInitialized (address);
#endif

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, sizeof (uae_u32));
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::SetWord
// ---------------------------------------------------------------------------

void EmBankDRAM::SetWord (emuptr address, uint32 value)
{
	if (address > gDynamicHeapSize)
	{
		EmBankSRAM::SetWord (address, value);
		return;
	}

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMWordWrite]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (address & 1) != 0)
	{
		AddressError (address, sizeof (uae_u16), false);
	}

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, SetWord, uae_u16, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!InlineValidAddress (address, sizeof (uae_u16)))
	{
		InvalidAccess (address, sizeof (uae_u16), false);
	}

	ScreenCheck (metaAddress, address, sizeof (uae_u16));

#if (HAS_PROFILING)
	CYCLE_PUTWORD (WAITSTATES_DRAM);
#endif

	do_put_mem_word (gRAM_Memory + address, value);

#if FOR_LATER
	// Mark that this memory location can now be read from.

	MetaMemory::MarkWordInitialized (address);
#endif

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, sizeof (uae_u16));
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::SetByte
// ---------------------------------------------------------------------------

void EmBankDRAM::SetByte (emuptr address, uint32 value)
{
	if (address > gDynamicHeapSize)
	{
		EmBankSRAM::SetByte (address, value);
		return;
	}

#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMByteWrite]++;
#endif

	register uint8*	metaAddress = InlineGetMetaAddress (address);
	META_CHECK (metaAddress, address, SetByte, uae_u8, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!InlineValidAddress (address, sizeof (uae_u8)))
	{
		InvalidAccess (address, sizeof (uae_u8), false);
	}

	ScreenCheck (metaAddress, address, sizeof (uae_u8));

#if (HAS_PROFILING)
	CYCLE_PUTBYTE (WAITSTATES_DRAM);
#endif

	do_put_mem_byte (gRAM_Memory + address, value);

#if FOR_LATER
	// Mark that this memory location can now be read from.

	MetaMemory::MarkByteInitialized (address);
#endif

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (address, sizeof (uae_u8));
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::ValidAddress
// ---------------------------------------------------------------------------

int EmBankDRAM::ValidAddress (emuptr address, uint32 size)
{
	int	result = InlineValidAddress (address, size);

	return result;
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::GetRealAddress
// ---------------------------------------------------------------------------

uint8* EmBankDRAM::GetRealAddress (emuptr address)
{
	return InlineGetRealAddress (address);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::GetMetaAddress
// ---------------------------------------------------------------------------

uint8* EmBankDRAM::GetMetaAddress (emuptr address)
{
	return InlineGetMetaAddress (address);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::AddOpcodeCycles
// ---------------------------------------------------------------------------

void EmBankDRAM::AddOpcodeCycles (void)
{
#if (HAS_PROFILING)
	CYCLE_GETWORD (WAITSTATES_DRAM);
#endif
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::AddressError
// ---------------------------------------------------------------------------

void EmBankDRAM::AddressError (emuptr address, long size, Bool forRead)
{
	assert (gCPU68K);
	gCPU68K->AddressError (address, size, forRead);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::InvalidAccess
// ---------------------------------------------------------------------------

void EmBankDRAM::InvalidAccess (emuptr address, long size, Bool forRead)
{
	assert (gCPU68K);
	gCPU68K->BusError (address, size, forRead);
}


// ---------------------------------------------------------------------------
//		 EmBankDRAM::ProbableCause
// ---------------------------------------------------------------------------

void EmBankDRAM::ProbableCause (emuptr address, long size, Bool forRead)
{
	Errors::EAccessType	whatHappened = MetaMemory::GetWhatHappened (address, size, forRead);

	if (whatHappened != Errors::kOKAccess)
	{
		assert (whatHappened != Errors::kUnknownAccess);

		Memory::PreventedAccess (address, size, forRead, whatHappened);
	}
}
