/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmCPU68K.h"

#include "EmBankROM.h"			// EmBankROM::GetMemoryStart
#include "EmHAL.h"				// EmHAL::GetInterruptLevel
#include "EmMemory.h"			// CEnableFullAccess
#include "EmSession.h"			// HandleInstructionBreak
#include "Logging.h"			// LogAppendMsg
#include "MetaMemory.h"			// IsCPUBreak
#include "Platform.h"			// GetMilliseconds
#include "SessionFile.h"		// WriteDBallRegs, etc.
#include "StringData.h"			// kExceptionNames

#include <algorithm>			// find

#if __profile__
#include <Profiler.h>
#endif

#if HAS_OMNI_THREAD
const uint32	kCheckFrequencyMask	= 0x000000FF;
#endif

#if defined (macintosh) && defined (_DEBUG) && 0
	#define HAVE_DEAD_MANS_SWITCH	1
#else
	#define HAVE_DEAD_MANS_SWITCH	0
#endif

// !!! Why is this here?  Shouldn't it be in Profiling.h?
#if HAS_PROFILING
perfRec perftbl[65536];
#endif

// Define our own flags for regs.spcflag.  Please do not let these
// overlap with UAE-defined flags (should not fall below 0x0002000)
// and avoid using the high bit just for safety.

#define SPCFLAG_END_OF_CYCLE	(0x40000000)


// Data needed by UAE.

int	areg_byteinc[] = { 1,1,1,1,1,1,1,2 };	// (normally in newcpu.c)
int	imm8_table[] = { 8,1,2,3,4,5,6,7 }; 	// (normally in newcpu.c)

int	movem_index1[256];						// (normally in newcpu.c)
int	movem_index2[256];						// (normally in newcpu.c)
int	movem_next[256];						// (normally in newcpu.c)

cpuop_func*	cpufunctbl[65536];				// (normally in newcpu.c)

uint16	last_op_for_exception_3;			/* Opcode of faulting instruction */
emuptr	last_addr_for_exception_3;			/* PC at fault time */
emuptr	last_fault_for_exception_3; 		/* Address that generated the exception */

struct regstruct	regs;					// (normally in newcpu.c)
struct flag_struct	regflags;				// (normally in support.c)


// These variables should strictly be in a sub-system that implements
// the stack overflow checking, etc.  However, for performance reasons,
// we need to expose them to UAE (see the CHECK_STACK_POINTER_ASSIGNMENT,
// et al macros), so define them here.
//
// Similar comments for the CheckKernelStack function.

uae_u32	gStackHigh;
uae_u32	gStackLowWarn;
uae_u32	gStackLow;
uae_u32	gKernelStackOverflowed;


// Definitions of the stack frames used in EmCPU68K::ProcessException.

#include "PalmPack.h"

struct ExceptionStackFrame1
{
	uint16 statusRegister;
	uint32 programCounter;
};

struct ExceptionStackFrame2
{
	uint16 functionCode;
	uint32 accessAddress;
	uint16 instructionRegister;
	uint16 statusRegister;
	uint32 programCounter;
};

#include "PalmPackPop.h"


EmCPU68K*	gCPU68K;


// ---------------------------------------------------------------------------
//		 EmCPU68K::Cycle
// ---------------------------------------------------------------------------

inline void EmCPU68K::Cycle (Bool sleeping)
{
	// Don't do anything if we're in the middle of an ATrap call.  We don't
	// need interrupts firing or tmr counters incrementing.

	assert (fSession);
	if (fSession->IsNested ())
	{
		return;
	}

	// Perform CPU-specific idling.

	EmHAL::Cycle (sleeping);

	// Note, the way we do this is very critical!  Previously, I'd updated
	// the screen and checked for events separately by calling two different
	// functions.  Each of these functions would check to see if an
	// appropropriate amount of time had elapsed, and perform their function
	// if it had.
	//
	// Making 4 function calls (one to the "update screen" function, one to
	// the "handle events" function, and two to TickCount()) was extremely
	// expensive.  Changing this to a single call to TickCount (in the inline
	// Platform::TimeForCycleSlowly function) speeds up the main CPU loop by 20%!

	if (Platform::TimeForCycleSlowly (sleeping))
	{
		EmHAL::CycleSlowly (sleeping);

		// Do some platform-specific stuff.

		Platform::CycleSlowly ();
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::EmCPU68K
// ---------------------------------------------------------------------------

EmCPU68K::EmCPU68K (EmSession* session) :
	EmCPU (session),
	fInstructionStart (UAE_NULL),
	fLastTraceAddress (UAE_NULL),
	fCycleCount (0),
//	fExceptionHandlers (),
	fHookJSR (),
	fHookJSR_Ind (),
	fHookLINK (),
	fHookRTE (),
	fHookRTS (),
	fHookNewPC (),
	fHookNewSP ()
#if REGISTER_HISTORY
	, fRegHistoryIndex (0)
//	, fRegHistory ()
#endif
#if EXCEPTION_HISTORY
	, fExceptionHistoryIndex (0)
//	, fExceptionHistory ()
#endif
{
	this->InitializeUAETables ();

	assert (gCPU68K == NULL);
	gCPU68K = this;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::~EmCPU68K
// ---------------------------------------------------------------------------

EmCPU68K::~EmCPU68K (void)
{
	assert (gCPU68K == this);
	gCPU68K = NULL;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::Reset
// ---------------------------------------------------------------------------

void EmCPU68K::Reset (Bool hardwareReset)
{
	fInstructionStart		= UAE_NULL;
	fLastTraceAddress		= UAE_NULL;
	fCycleCount				= 0;

#if REGISTER_HISTORY
	fRegHistoryIndex		= 0;
#endif

#if EXCEPTION_HISTORY
	fExceptionHistoryIndex	= 0;
#endif

	gStackHigh				= 0xFFFFFFFF;
	gStackLowWarn			= UAE_NULL;
	gStackLow				= UAE_NULL;
	gKernelStackOverflowed	= false;

	if (hardwareReset)
	{
		// (taken from m68k_reset in newcpu.c)

		// !!! I think that we really need to emulate ROM appearing at NULL.
		// That would break our dependency on EmBankROM.

		emuptr romStart = EmBankROM::GetMemoryStart ();
		m68k_areg (regs, 7) = get_long (romStart);
		m68k_setpc (get_long (romStart + 4));
		fInstructionStart = m68k_getpc ();

		// Note, on the 68K, the t0 and m flags must always be zero.

		regs.prefetch	= 0x0000;
		regs.kick_mask	= 0xF80000;	// (not a 68K register)
		regs.s			= 1;		// supervisor/user state
		regs.m			= 0;		// master/interrupt state
		regs.stopped	= 0;		// (not a 68K register)
		regs.t1 		= 0;		// 1 = trace on any instruction
		regs.t0 		= 0;		// 1 = trace on change of flow
		SET_ZFLG (0);
		SET_XFLG (0);
		SET_CFLG (0);
		SET_VFLG (0);
		SET_NFLG (0);
		regs.spcflags	= 0;		// (not a 68K register)
		regs.intmask	= 7;		// disable all interrupts
		regs.vbr = regs.sfc = regs.dfc = 0;
		regs.fpcr = regs.fpsr = regs.fpiar = 0;
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::Save
// ---------------------------------------------------------------------------

void EmCPU68K::Save (SessionFile& f)
{
	// Write out the CPU Registers

	regstruct	tempRegs;
	this->GetRegisters (tempRegs);

	Canonical (tempRegs);
	f.WriteDBallRegs (tempRegs);
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::Load
// ---------------------------------------------------------------------------

void EmCPU68K::Load (SessionFile& f)
{
	// Read in the CPU Registers.

	regstruct	tempRegs;
	f.ReadDBallRegs (tempRegs);

	Canonical (tempRegs);
	this->SetRegisters (tempRegs);
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::Execute
// ---------------------------------------------------------------------------

void EmCPU68K::Execute (void)
{
	// -----------------------------------------------------------------------
	// Declare these here (instead of where they're actually used), so that
	// VC++ doesn't give us any lip about jumping over them with our goto
	// statement.
	// -----------------------------------------------------------------------

	int		drift;
	uint32	counter				= 0;

	// Getting these into registers is a big win on Intel.  Getting cpufunctbl
	// into a register is a big win on PPC.

#define INDIRECT_ACCESS 1
#if INDIRECT_ACCESS
	register emuptr*			instructionStart	= &fInstructionStart;
	register unsigned long*		cycleCount			= &fCycleCount;
	register cpuop_func**		functable			= cpufunctbl;

	#define fInstructionStart	*instructionStart
	#define fCycleCount			*cycleCount
	#define cpufunctbl			functable
#endif

#if HAS_PROFILING_DEBUG
	UInt64	readCyclesSaved		= 0;
	UInt64	writeCyclesSaved	= 0;
	UInt64	clockCyclesSaved	= 0;
#endif

#if HAVE_DEAD_MANS_SWITCH
	// -----------------------------------------------------------------------
	// Put in a little dead-man's switch. If this function doesn't exit for a
	// long time, let us get into the debugger.
	// -----------------------------------------------------------------------

	uint32	deadManNow;
	uint32	deadManStart = Platform::GetMilliseconds ();
#endif

	// -----------------------------------------------------------------------
	// Check for the stopped flag before entering the "execute an opcode"
	// section.  It could be that we last exited the loop while still in stop
	// mode, and we need to wind our way back down to that spot.
	// -----------------------------------------------------------------------

	if ((regs.spcflags & SPCFLAG_STOP) != 0)
		goto StoppedLoop;

	while (1)
	{
#if REGISTER_HISTORY
		// -----------------------------------------------------------------------
		// Save the registers for the post-mortem, but don't record the
		// instructions we generate when calling the ROM as a subroutine.  We want
		// those to be as transparent as possible.	In particular, we don't want
		// any functions that we call as part of figuring out why a problem
		// occured to knock the problem-causing registers off of our array.
		// -----------------------------------------------------------------------

//		if (!fSession->IsNested ())
		{
			++fRegHistoryIndex;
			fRegHistory[fRegHistoryIndex & (kRegHistorySize - 1)] = regs;
		}
#endif

#if HAVE_DEAD_MANS_SWITCH
		// -----------------------------------------------------------------------
		// Put in a little dead-man's switch. If this function doesn't exit for a
		// long time, let us get into the debugger.
		// -----------------------------------------------------------------------

		deadManNow = Platform::GetMilliseconds ();
		if ((deadManNow - deadManStart) > 5000)
		{
			Platform::Debugger ();
		}
#endif

		// -----------------------------------------------------------------------
		// In the regs struct, pc_p contains the pointer to the actual memory
		// location from which to fetch the next opcode.  By contrast, pc contains
		// the PC in "virtual" address space -- that is, the address that the
		// emulated environment thinks it should be.  However, as an optimization,
		// UAE does not increment pc at the same time it increments pc_p.  That
		// would slow things down a touch.  Instead, any time a radical change is
		// made to the PC (for instance, a branch, jump, or return is performed),
		// pc is set to the virtual PC, and pc_p and pc_oldp are set to the
		// physical PC.  After that, as opcodes are fetched, only pc_p is
		// incremented.  However, the virtual PC can be rederived at any time by
		// getting the delta between pc_p and pc_oldp and adding that back to pc.
		//
		// Most of this is handled behind the scenes by the m68k_getpc,
		// m68k_setpc, and m68k_incpc, and we don't have to worry about it.
		// However, in this function, every cycle counts.  We will have to make
		// use of this delta value (below I call it "drift") twice.  Rather than
		// rely on the compiler to perform the CSE, I do it manually.
		// -----------------------------------------------------------------------

		drift = regs.pc_p - regs.pc_oldp;

		// -----------------------------------------------------------------------
		// See if we need to halt CPU execution at this location.  We could need
		// to do this for several reasons, including hitting soft breakpoints or
		// needing to execute tailpatches.
		// -----------------------------------------------------------------------

		if (MetaMemory::IsCPUBreak (regs.pc_meta_oldp + drift))
		{
			assert (fSession);
			fSession->HandleInstructionBreak ();
		}

#if HAS_PROFILING
		if (gProfilingEnabled)
		{
			// If detailed, log instruction here

			if (gProfilingDetailed)
			{
				ProfileInstructionEnter (fInstructionStart);
			}

#if HAS_PROFILING_DEBUG
			readCyclesSaved		= gReadCycles;
			writeCyclesSaved	= gWriteCycles;
			clockCyclesSaved	= gClockCycles;
#endif

			// Turn gProfilingCounted on here so the opcode fetch below is counted.

			gProfilingCounted = true;
		}
#endif

		// -----------------------------------------------------------------------
		// Remember the starting point of this opcode.	Later, if an exception
		// occurs, we may want to reset the PC to this point.
		//
		// Set fInstructionStart *after* calling HandleInstructionBreak.  That
		// function may result in Palm OS subroutine calls, which will call
		// Execute recursively, changing fInstructionStart.
		// -----------------------------------------------------------------------

		fInstructionStart = regs.pc + drift;

		// -----------------------------------------------------------------------
		// Execute the instruction.  Catch any exceptions thrown because of
		// address/bus errors.
		//
		// In CodeWarrior, the overhead of this try/catch block is minimal: the SP
		// is saved, and there's a branch around the catch block.  According to
		// Greg, the overhead in VC++ is similar.	Timing tests show no
		// appreciable difference with this try/catch block in place.
		// -----------------------------------------------------------------------

		try
		{
		// =======================================================================
		// Execute the opcode.
		// -----------------------------------------------------------------------
			uint32	opcode = get_iword (0);
			fCycleCount += (cpufunctbl[opcode]) (opcode);
		// =======================================================================

#if HAS_PROFILING
			if (gProfilingEnabled)
			{
				// Add in the extra time taken to execute the instruction.

				ProfileIncrementClock (perftbl[opcode].extraCycles);

				// Detail (instruction level) profiling support

				if (gProfilingDetailed)
				{
					ProfileInstructionExit (fInstructionStart);
				}

#if HAS_PROFILING_DEBUG
				// Validity check on EmMemory stuff.

				Boolean tryAgain = false;

				if (perftbl[opcode].readCycles != 0xFF &&
					gReadCycles - readCyclesSaved != perftbl[opcode].readCycles)
				{
					gReadMismatch += gReadCycles - readCyclesSaved - perftbl[opcode].readCycles;
				}

				if (perftbl[opcode].writeCycles != 0xFF &&
					gWriteCycles - writeCyclesSaved != perftbl[opcode].writeCycles)
				{
					gWriteMismatch += gWriteCycles - writeCyclesSaved - perftbl[opcode].writeCycles;
				}

				if (tryAgain)
				{
					(cpufunctbl[opcode]) (opcode);
				}
#endif
			}
#endif
		}
		catch (ErrCode errCode)
		{
			// We usually get here because Errors::HandleDlgButton
			// called Errors::Scram after resetting.

			assert (fSession);
			assert (!fSession->IsNested ());
			assert (errCode == kError_NoError);
		}

		// Update fInstructionStart again so that in-between-cycle
		// operations can get the correct, updated value.

		drift = regs.pc_p - regs.pc_oldp;
		fInstructionStart = regs.pc + drift;

		// Perform periodic tasks.

		this->Cycle (false);

		// -----------------------------------------------------------------------
		// Check to see if it's time to quit.  We check only every once and a
		// while for performance reasons.  Checking every time through the loop
		// slows down emulation by about 30% on Windows.
		// -----------------------------------------------------------------------

#if HAS_OMNI_THREAD
		if ((++counter & kCheckFrequencyMask) == 0)
		{
			assert (fSession);
			omni_mutex_lock	lock (fSession->fSharedLock);

			if (fSession->fSuspendState.fAllCounters)
			{
				this->CheckAfterCycle ();
			}
		}
#endif

StoppedLoop:

		// -----------------------------------------------------------------------
		// Handle special conditions.  NB: the code reached by calling
		// EmCPU68K::ExecuteSpecial used to be inline in this function.  Moving it
		// out (thus simplifying both EmCPU68K::Execute and EmCPU68K::ExecuteSpecial)
		// sped up the CPU loop by 9%!
		// -----------------------------------------------------------------------

		if (regs.spcflags)
		{
			if (this->ExecuteSpecial ())
				break;
		}

#if HAS_PROFILING
		gProfilingCounted = false;
#endif
	}	// while (1)

#if INDIRECT_ACCESS
	#undef fInstructionStart
	#undef fCycleCount
	#undef cpufunctbl
#endif
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ExecuteSimple
// ---------------------------------------------------------------------------

/*
	This is the same loop with the following removed:
	
	* Comments
	* Profiling code
	* Profiling debugging code
	* Dead mans's check on Mac
	* "Time Startup" code
	* Register history recording

	It's essentially what we get in Release/Non-profile builds.
*/

#if 0

void EmCPU68K::ExecuteSimple (void)
{
	int		drift;
	uint32	counter = 0;

	if ((regs.spcflags & SPCFLAG_STOP) != 0)
		goto StoppedLoop;

	while (1)
	{
		drift = regs.pc_p - regs.pc_oldp;
		fInstructionStart = regs.pc + drift;

		if (MetaMemory::IsCPUBreak (regs.pc_meta_oldp + drift))
		{
			this->HandleInstructionBreak ();
		}

		try
		{
			uint16	opcode = get_iword (0);
			cpufunctbl[opcode] (opcode);
		}
		catch (ErrCode errCode)
		{
		}

		this->Cycle (false);

#if HAS_OMNI_THREAD
		if ((++counter & kCheckFrequencyMask) == 0)
		{
			assert (fSession);
			omni_mutex_lock	lock (fSession->fSharedLock);

			if (fSession->fSuspendState.fAllCounters)
			{
				this->CheckAfterCycle ();
			}
		}
#endif

StoppedLoop:
		if (regs.spcflags)
		{
			if (this->ExecuteSpecial ())
				break;
		}
	}
}

#endif


// ---------------------------------------------------------------------------
//		 EmCPU68K::ExecuteSpecial
// ---------------------------------------------------------------------------

Bool EmCPU68K::ExecuteSpecial (void)
{
	// If we're making subroutine calls, then all we process are requests
	// to break from the CPU loop.  We don't want interrupts, tracing, etc.
	// getting in the way.

	assert (fSession);
	if (fSession->IsNested () != 0)
	{
		return this->CheckForBreak ();
	}

	// Check for Reset first.  If this is set, don't do anything else.

	if ((regs.spcflags & SPCFLAG_END_OF_CYCLE))
	{		
		if (fSession->ExecuteSpecial (true))
			return true;
	}

	// Execute UAE spcflags-handling code (from do_specialties in newcpu.c).

	// If we're tracing, execute the trace vector.
	//
	// The check for SPCFLAG_STOP was added in Poser.  It's needed
	// if we're re-entering ExecuteStopped loop on the Mac after
	// exiting in order to handle events.

	if ((regs.spcflags & SPCFLAG_DOTRACE) && !(regs.spcflags & SPCFLAG_STOP))
	{
		this->ProcessException (kException_Trace);
	}

	if (regs.spcflags & SPCFLAG_STOP)
	{
		if (this->ExecuteStoppedLoop ())
		{
			regs.spcflags &= ~SPCFLAG_BRK;
			return true;
		}
	}

	// Do trace-mode stuff (do_trace from newcpu.c does more,
	// but it's only needed for CPU_LEVEL > 0)

	if (regs.spcflags & SPCFLAG_TRACE)
	{
		if (regs.t1)
		{
			fLastTraceAddress = m68k_getpc ();
			regs.spcflags &= ~SPCFLAG_TRACE;
			regs.spcflags |= SPCFLAG_DOTRACE;
		}
	}

	if ((regs.spcflags & SPCFLAG_DOINT) && !gKernelStackOverflowed)
	{
		int32 interruptLevel = EmHAL::GetInterruptLevel ();
		regs.spcflags &= ~SPCFLAG_DOINT;	// was ~(SPCFLAG_INT | SPCFLAG_DOINT) in Greg and Craig, but the latest UAE has this
		if ((interruptLevel != -1) && (interruptLevel > regs.intmask))
		{
			this->ProcessInterrupt (interruptLevel);
			regs.stopped = 0;
		}
	}

	if (regs.spcflags & SPCFLAG_INT)
	{
		regs.spcflags &= ~SPCFLAG_INT;
		regs.spcflags |= SPCFLAG_DOINT;
	}

	// Check for Poser end-of-cycle operations.  This is inserted
	// before the standard UAE check of the SPCFLAG_BRK flag.

	if ((regs.spcflags & SPCFLAG_END_OF_CYCLE))
	{
		regs.spcflags &= ~SPCFLAG_END_OF_CYCLE;
		
		if (fSession->ExecuteSpecial (false))
			return true;
	}

	return this->CheckForBreak ();
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ExecuteStoppedLoop
// ---------------------------------------------------------------------------

Bool EmCPU68K::ExecuteStoppedLoop (void)
{
	assert (fSession);
	assert (!fSession->IsNested ());
	assert (regs.intmask < 7);

	// While the CPU is stopped (because a STOP instruction was
	// executed) do some idle tasks.

#if HAVE_DEAD_MANS_SWITCH
	// -----------------------------------------------------------------------
	// Put in a little dead-man's switch. If this function doesn't
	// exit for a long time, let us get into the debugger.
	// -----------------------------------------------------------------------

	uint32 deadManStart = Platform::GetMilliseconds ();
#endif

	do {
#if HAVE_DEAD_MANS_SWITCH
		// -----------------------------------------------------------------------
		// Put in a little dead-man's switch. If this function doesn't
		// exit for a long time, let us get into the debugger.
		// -----------------------------------------------------------------------

		uint32 deadManNow = Platform::GetMilliseconds ();
		if ((deadManNow - deadManStart) > 5000)
		{
			Platform::Debugger ();
		}
#endif

		// -----------------------------------------------------------------------
		// Slow down processing so that the timer used
		// to increment the tickcount doesn't run too quickly.
		// -----------------------------------------------------------------------

#if __profile__
	short	oldStatus = ProfilerGetStatus ();
	ProfilerSetStatus (false);
#endif

		Platform::Delay ();

#if __profile__
	ProfilerSetStatus (oldStatus);
#endif

		// Perform periodic tasks.

		this->Cycle (true);

		// -----------------------------------------------------------------------
		// Check to see if it's time to quit.
		// -----------------------------------------------------------------------

#if HAS_OMNI_THREAD
		{
			assert (fSession);
			omni_mutex_lock	lock (fSession->fSharedLock);

			if (fSession->fSuspendState.fAllCounters)
			{
				this->CheckAfterCycle ();
			}
		}
#endif

		// Process an interrupt (see if it's time to wake up).

		if (regs.spcflags & (SPCFLAG_INT | SPCFLAG_DOINT))
		{
			int32 interruptLevel = EmHAL::GetInterruptLevel ();

			regs.spcflags &= ~(SPCFLAG_INT | SPCFLAG_DOINT);

			if ((interruptLevel != -1) && (interruptLevel > regs.intmask))
			{
				this->ProcessInterrupt (interruptLevel);
				regs.stopped = 0;
				regs.spcflags &= ~SPCFLAG_STOP;
			}
		}

		if (this->CheckForBreak ())
		{
			return true;
		}
	} while (regs.spcflags & SPCFLAG_STOP);

	return false;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::Cycle
// ---------------------------------------------------------------------------

#if 0	// inline

void EmCPU68K::Cycle (Bool sleeping)
{
	// Don't do anything if we're in the middle of an ATrap call.  We don't
	// need interrupts firing or tmr counters incrementing.

	assert (fSession);
	if (fSession->IsNested ())
	{
		return;
	}

	// Perform CPU-specific idling.
		
	EmHAL::Cycle (sleeping);

	// Note, the way we do this is very critical!  Previously, I'd updated
	// the screen and checked for events separately by calling two different
	// functions.  Each of these functions would check to see if an
	// appropropriate amount of time had elapsed, and perform their function
	// if it had.
	//
	// Making 4 function calls (one to the "update screen" function, one to
	// the "handle events" function, and two to TickCount()) was extremely
	// expensive.  Changing this to a single call to TickCount (in the inline
	// Platform_TimeForCycle function) speeds up the main CPU loop by 20%!

	if (Platform::TimeForCycleSlowly (sleeping))
	{
		EmHAL::CycleSlowly (sleeping);

		// Do some platform-specific stuff.

		Platform::Cycle ();
	}
}

#endif

// ---------------------------------------------------------------------------
//		 EmCPU68K::CheckAfterCycle
// ---------------------------------------------------------------------------

void EmCPU68K::CheckAfterCycle (void)
{
	regs.spcflags |= SPCFLAG_END_OF_CYCLE;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::CheckForBreak
// ---------------------------------------------------------------------------
// Check to see if the conditions tell us to break from the CPU Execute loop.

Bool EmCPU68K::CheckForBreak (void)
{
	if ((regs.spcflags & SPCFLAG_BRK) != 0)
	{
		regs.spcflags &= ~SPCFLAG_BRK;
		return true;
	}

	assert (fSession);
	return fSession->CheckForBreak ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessInterrupt
// ---------------------------------------------------------------------------

void EmCPU68K::ProcessInterrupt (int32 interrupt)
{
	this->ProcessException ((ExceptionNumber) (EmHAL::GetInterruptBase () + interrupt));

	regs.intmask = interrupt;
	regs.spcflags |= SPCFLAG_INT;	// Check for higher-level interrupts
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessException
// ---------------------------------------------------------------------------

void EmCPU68K::ProcessException (ExceptionNumber exception)
{
	// Make sure the Status Register is up-to-date.

	this->UpdateSRFromRegisters ();

#if EXCEPTION_HISTORY
	// Save the exception for the post-mortem

	fExceptionHistoryIndex++;

	long	index = fExceptionHistoryIndex & (kExceptionHistorySize - 1);
	fExceptionHistory[index].name = kExceptionNames[exception];

	fExceptionHistory[index].pc = fInstructionStart;
	fExceptionHistory[index].sp = m68k_areg (regs, 7);
#endif

#if HAS_PROFILING
	// Don't count cycles spent in exception handlers against functions.

	if (gProfilingEnabled)
	{
		ProfileInterruptEnter (exception, m68k_getpc ());
	}
#endif

	// Let any custom exception handler have a go at it.  If it returns
	// true, it handled it completely, and we don't have anything else to do.

	Bool							handled = false;
	Hook68KExceptionList			fns = fExceptionHandlers[exception];
	Hook68KExceptionList::iterator	iter = fns.begin ();
	while (iter != fns.end ())
	{
		if ((*iter) (exception))
		{
			handled = true;
		}

		++iter;
	}

	if (handled)
	{
#if HAS_PROFILING
			if (gProfilingEnabled)
			{
				ProfileInterruptExit (m68k_getpc ());
			}
#endif
			return;
	}

	// The following is vaguely modelled after Exception() from newcpu.c
	// (The call to MakeSR appears at the start of this method).

	// If not in supervisor mode, set the usp, restore A7 from the isp,
	// and transfer to supervisor mode.

	if (!regs.s)
	{
		regs.usp = m68k_areg (regs, 7);
		m68k_areg (regs, 7) = regs.isp;
		regs.s = 1;
	}

	// Set up the stack frame.
	// !!! If we're handling a trace exception, I think that fLastTraceAddress
	// comes into play here instead of m68k_getpc.
	// !!! Manage this with EmPalmStructs...

	if (exception == kException_BusErr || exception == kException_AddressErr)
	{
		COMPILE_TIME_ASSERT (sizeof (ExceptionStackFrame2) == 14);
		m68k_areg (regs, 7) -= sizeof (ExceptionStackFrame2);
		CHECK_STACK_POINTER_DECREMENT ();

		emuptr frame = m68k_areg (regs, 7);

		// Eh...Palm OS doesn't use these 3 anyway...
		put_word (frame + offsetof (ExceptionStackFrame2, functionCode), 0);
		put_long (frame + offsetof (ExceptionStackFrame2, accessAddress), 0);
		put_word (frame + offsetof (ExceptionStackFrame2, instructionRegister), 0);

		put_word (frame + offsetof (ExceptionStackFrame2, statusRegister), regs.sr);
		put_long (frame + offsetof (ExceptionStackFrame2, programCounter), m68k_getpc ());
	}
	else
	{
		COMPILE_TIME_ASSERT (sizeof (ExceptionStackFrame1) == 6);
		m68k_areg (regs, 7) -= sizeof (ExceptionStackFrame1);
		CHECK_STACK_POINTER_DECREMENT ();

		emuptr frame = m68k_areg (regs, 7);

		put_word (frame + offsetof (ExceptionStackFrame1, statusRegister), regs.sr);
		put_long (frame + offsetof (ExceptionStackFrame1, programCounter), m68k_getpc ());
	}

	emuptr newpc;
	{
		CEnableFullAccess	munge;	// Remove blocks on memory access.

		// Get the exception handler address.
		newpc = get_long (regs.vbr + 4 * exception);
	}

	// Check the exception handler address and jam it into the PC.

	this->CheckNewPC (newpc);
	m68k_setpc (newpc);

	// Turn tracing off.

	regs.t1 = regs.t0 = regs.m = 0;
	regs.spcflags &= ~(SPCFLAG_TRACE | SPCFLAG_DOTRACE);

	// Update fInstructionStart again so that in-between-cycle
	// operations can get the correct, updated value.

	fInstructionStart = m68k_getpc ();
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessIllegalInstruction
// ---------------------------------------------------------------------------

void EmCPU68K::ProcessIllegalInstruction (EmOpcode68K opcode)
{
	// This function is loosely based on op_illg in newcpu.c

	// Process an FTrap.

	if ((opcode & 0xF000) == 0xF000)
	{
		this->ProcessException (kException_FTrap);
	}

	// Process an ATrap.

	else if ((opcode & 0xF000) == 0xA000)
	{
		this->ProcessException (kException_ATrap);
	}

	// Process all other opcodes.

	else
	{
		this->ProcessException (kException_IllegalInstr);
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessJSR
// ---------------------------------------------------------------------------

int EmCPU68K::ProcessJSR (emuptr oldPC, emuptr dest)
{
	int	handled = false;

	Hook68KJSRList::iterator	iter = fHookJSR.begin ();
	while (iter != fHookJSR.end ())
	{
		if ((*iter) (oldPC, dest))
		{
			handled = true;
		}

		++iter;
	}

	return handled;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessJSR_Ind
// ---------------------------------------------------------------------------

int EmCPU68K::ProcessJSR_Ind (emuptr oldPC, emuptr dest)
{
	int	handled = false;

	Hook68KJSR_IndList::iterator	iter = fHookJSR_Ind.begin ();
	while (iter != fHookJSR_Ind.end ())
	{
		if ((*iter) (oldPC, dest))
		{
			handled = true;
		}

		++iter;
	}

	return handled;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessLINK
// ---------------------------------------------------------------------------

void EmCPU68K::ProcessLINK (int linkSize)
{
	Hook68KLINKList::iterator	iter = fHookLINK.begin ();
	while (iter != fHookLINK.end ())
	{
		(*iter) (linkSize);

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessRTE
// ---------------------------------------------------------------------------

int EmCPU68K::ProcessRTE (emuptr dest)
{
	int	handled = false;

	Hook68KRTEList::iterator	iter = fHookRTE.begin ();
	while (iter != fHookRTE.end ())
	{
		if ((*iter) (dest))
		{
			handled = true;
		}

		++iter;
	}

	return handled;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::ProcessRTS
// ---------------------------------------------------------------------------

int EmCPU68K::ProcessRTS (emuptr dest)
{
	int	handled = false;

	Hook68KRTSList::iterator	iter = fHookRTS.begin ();
	while (iter != fHookRTS.end ())
	{
		if ((*iter) (dest))
		{
			handled = true;
		}

		++iter;
	}

	return handled;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::CheckNewPC
// ---------------------------------------------------------------------------

void EmCPU68K::CheckNewPC (emuptr dest)
{
	Hook68KNewPCList::iterator	iter = fHookNewPC.begin ();
	while (iter != fHookNewPC.end ())
	{
		(*iter) (dest);

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::CheckNewSP
// ---------------------------------------------------------------------------

void EmCPU68K::CheckNewSP (EmStackChangeType type)
{
	Hook68KNewSPList::iterator	iter = fHookNewSP.begin ();
	while (iter != fHookNewSP.end ())
	{
		(*iter) (type);

		++iter;
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::InstallHookException
//		 EmCPU68K::InstallHookJSR
//		 EmCPU68K::InstallHookJSR_Ind
//		 EmCPU68K::InstallHookLINK
//		 EmCPU68K::InstallHookRTE
//		 EmCPU68K::InstallHookRTS
//		 EmCPU68K::InstallHookNewPC
//		 EmCPU68K::InstallHookNewSP
// ---------------------------------------------------------------------------

void EmCPU68K::InstallHookException (ExceptionNumber exceptionNumber,
										Hook68KException fn)
{
	fExceptionHandlers[exceptionNumber].push_back (fn);
}


void EmCPU68K::InstallHookJSR (Hook68KJSR fn)
{
	fHookJSR.push_back (fn);
}


void EmCPU68K::InstallHookJSR_Ind (Hook68KJSR_Ind fn)
{
	fHookJSR_Ind.push_back (fn);
}


void EmCPU68K::InstallHookLINK (Hook68KLINK fn)
{
	fHookLINK.push_back (fn);
}


void EmCPU68K::InstallHookRTE (Hook68KRTE fn)
{
	fHookRTE.push_back (fn);
}


void EmCPU68K::InstallHookRTS (Hook68KRTS fn)
{
	fHookRTS.push_back (fn);
}


void EmCPU68K::InstallHookNewPC (Hook68KNewPC fn)
{
	fHookNewPC.push_back (fn);
}


void EmCPU68K::InstallHookNewSP (Hook68KNewSP fn)
{
	fHookNewSP.push_back (fn);
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::RemoveHookException
//		 EmCPU68K::RemoveHookJSR
//		 EmCPU68K::RemoveHookJSR_Ind
//		 EmCPU68K::RemoveHookLINK
//		 EmCPU68K::RemoveHookRTE
//		 EmCPU68K::RemoveHookRTS
//		 EmCPU68K::RemoveHookNewPC
//		 EmCPU68K::RemoveHookNewSP
// ---------------------------------------------------------------------------

void EmCPU68K::RemoveHookException (ExceptionNumber exceptionNumber,
										Hook68KException fn)
{
	Hook68KExceptionList::iterator	iter = find (
							fExceptionHandlers[exceptionNumber].begin (),
							fExceptionHandlers[exceptionNumber].end (),
							fn);

	if (iter != fExceptionHandlers[exceptionNumber].end ())
	{
		fExceptionHandlers[exceptionNumber].erase (iter);
	}
}


void EmCPU68K::RemoveHookJSR (Hook68KJSR fn)
{
	Hook68KJSRList::iterator	iter = find (fHookJSR.begin (), fHookJSR.end (), fn);

	if (iter != fHookJSR.end ())
	{
		fHookJSR.erase (iter);
	}
}


void EmCPU68K::RemoveHookJSR_Ind (Hook68KJSR_Ind fn)
{
	Hook68KJSR_IndList::iterator	iter = find (fHookJSR_Ind.begin (), fHookJSR_Ind.end (), fn);

	if (iter != fHookJSR_Ind.end ())
	{
		fHookJSR_Ind.erase (iter);
	}
}


void EmCPU68K::RemoveHookLINK (Hook68KLINK fn)
{
	Hook68KLINKList::iterator	iter = find (fHookLINK.begin (), fHookLINK.end (), fn);

	if (iter != fHookLINK.end ())
	{
		fHookLINK.erase (iter);
	}
}


void EmCPU68K::RemoveHookRTE (Hook68KRTE fn)
{
	Hook68KRTEList::iterator	iter = find (fHookRTE.begin (), fHookRTE.end (), fn);

	if (iter != fHookRTE.end ())
	{
		fHookRTE.erase (iter);
	}
}


void EmCPU68K::RemoveHookRTS (Hook68KRTS fn)
{
	Hook68KRTSList::iterator	iter = find (fHookRTS.begin (), fHookRTS.end (), fn);

	if (iter != fHookRTS.end ())
	{
		fHookRTS.erase (iter);
	}
}


void EmCPU68K::RemoveHookNewPC (Hook68KNewPC fn)
{
	Hook68KNewPCList::iterator	iter = find (fHookNewPC.begin (), fHookNewPC.end (), fn);

	if (iter != fHookNewPC.end ())
	{
		fHookNewPC.erase (iter);
	}
}


void EmCPU68K::RemoveHookNewSP (Hook68KNewSP fn)
{
	Hook68KNewSPList::iterator	iter = find (fHookNewSP.begin (), fHookNewSP.end (), fn);

	if (iter != fHookNewSP.end ())
	{
		fHookNewSP.erase (iter);
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::GetRegisters
// ---------------------------------------------------------------------------

void EmCPU68K::GetRegisters (regstruct& registers)
{
	this->UpdateSRFromRegisters ();

	registers = regs;
	registers.pc = m68k_getpc ();
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::SetRegisters
// ---------------------------------------------------------------------------

void EmCPU68K::SetRegisters (regstruct& registers)
{
	regs = registers;
	this->UpdateRegistersFromSR ();

	m68k_setpc (registers.pc);
	
	this->CheckNewSP (kStackPointerChanged);
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::UpdateSRFromRegisters
// ---------------------------------------------------------------------------
// Create a 16-bit status register value from the broken out fields.  In
// general, we keep the various fields separate for speed of access.  However,
// there are times when we need the packed 16-bit field.
//
// This function is called any time the 16-bit representation of the SR is
// needed:
//
//		on EORSR
//		on ORSR
//		on ANDSR
//		on MVSR2
//		on MV2SR
//		on RTR
//		when processing exceptions (EmCPU68K::ProcessException)
//
//		SystemPacket::GetRegs
//		EmCPU68K::GetRegisters

void EmCPU68K::UpdateSRFromRegisters (void)
{
	// (taken from MakeSR in newcpu.c)

    regs.sr = ((regs.t1 << 15) | (regs.t0 << 14)
	       | (regs.s << 13) | (regs.m << 12) | (regs.intmask << 8)
	       | (GET_XFLG << 4) | (GET_NFLG << 3) | (GET_ZFLG << 2) | (GET_VFLG << 1)
	       | GET_CFLG);
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::UpdateRegistersFromSR
// ---------------------------------------------------------------------------
// Break out all of the fields from the 16-bit status register into their own
// separate variables.
//
// This function is called any time the SR has been update and needs to be
// re-expanded:
//
//		on EORSR
//		on ORSR
//		on ANDSR
//		on MV2SR
//		on STOP
//		on RTE
//		on RTR
//
//		SystemPacket::SetRegs
//		EmCPU68K::SetRegisters

void EmCPU68K::UpdateRegistersFromSR (void)
{
	// (taken from MakeFromSR in newcpu.c)
	
//	int oldm = regs.m;
	int olds = regs.s;
	
	regs.t1 = (regs.sr >> 15) & 1;
	regs.t0 = (regs.sr >> 14) & 1;
	regs.s = (regs.sr >> 13) & 1;
	regs.m = (regs.sr >> 12) & 1;
	regs.intmask = (regs.sr >> 8) & 7;

    SET_XFLG ((regs.sr >> 4) & 1);
    SET_NFLG ((regs.sr >> 3) & 1);
    SET_ZFLG ((regs.sr >> 2) & 1);
    SET_VFLG ((regs.sr >> 1) & 1);
    SET_CFLG (regs.sr & 1);

	if (olds != regs.s)
	{
		if (olds)
		{
			regs.isp = m68k_areg(regs, 7);
			m68k_areg(regs, 7) = regs.usp;
		}
		else
		{
			regs.usp = m68k_areg(regs, 7);
			m68k_areg(regs, 7) = regs.isp;
		}
	}
	
	regs.spcflags |= SPCFLAG_INT;

	if (regs.t1 || regs.t0)
	{
		regs.spcflags |= SPCFLAG_TRACE;
	}
	else
	{
		regs.spcflags &= ~(SPCFLAG_TRACE | SPCFLAG_DOTRACE);
	}
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::GetStartPC
// ---------------------------------------------------------------------------

emuptr EmCPU68K::GetStartPC (void)
{
	return fInstructionStart;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::GetCycleCount
// ---------------------------------------------------------------------------

uint32 EmCPU68K::GetCycleCount (void)
{
	return fCycleCount;
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::BusError
// ---------------------------------------------------------------------------

void EmCPU68K::BusError (emuptr address, long size, Bool forRead)
{
	UNUSED_PARAM (address);
	UNUSED_PARAM (size);
	UNUSED_PARAM (forRead);

	this->ProcessException (kException_BusErr);

	// Get outta here fast.  We don't try to execute the rest of the
	// instruction.

	Errors::Scram ();
}


// ---------------------------------------------------------------------------
//		 EmCPU68K::AddressError
// ---------------------------------------------------------------------------

void EmCPU68K::AddressError (emuptr address, long size, Bool forRead)
{
	UNUSED_PARAM (address);
	UNUSED_PARAM (size);
	UNUSED_PARAM (forRead);

	this->ProcessException (kException_AddressErr);

	// Get outta here fast.  We don't try to execute the rest of the
	// instruction.

	Errors::Scram ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmCPU68K::InitializeUAETables
// ---------------------------------------------------------------------------

void EmCPU68K::InitializeUAETables (void)
{
	static int	initialized;

	// All of the stuff in this function needs to be done only once;
	// it doesn't need to be executed every time we create a new CPU.

	if (initialized)
		return;

	initialized = true;

	// Initialize some CPU-related tables
	// (This initialization code is taken from init_m68k in newcpu.c)

	int 	i, j;

	for (i = 0 ; i < 256 ; i++)
	{
		for (j = 0 ; j < 8 ; j++)
		{
			if (i & (1 << j))
			{
				break;
			}
		}

		movem_index1[i] = j;
		movem_index2[i] = 7-j;
		movem_next[i] = i & (~(1 << j));
	}

	read_table68k ();
	do_merges ();

	// The rest of this code is based on build_cpufunctbl in newcpu.c.

	unsigned long	opcode;
	struct cputbl*	tbl = op_smalltbl_3;

	for (opcode = 0; opcode < 65536; opcode++)
	{
		cpufunctbl[opcode] = op_illg;
	}

	for (i = 0; tbl[i].handler != NULL; i++)
	{
		if (!tbl[i].specific)
		{
			cpufunctbl[tbl[i].opcode] = tbl[i].handler;
#if HAS_PROFILING
			perftbl[tbl[i].opcode] = tbl[i].perf;
#endif
		}
	}

	for (opcode = 0; opcode < 65536; opcode++)
	{
		cpuop_func* f;

		if (table68k[opcode].mnemo == i_ILLG || table68k[opcode].clev > 0)
		{
			continue;
		}

		if (table68k[opcode].handler != -1)
		{
			f = cpufunctbl[table68k[opcode].handler];
			if (f == op_illg)
			{
				abort ();
			}

			cpufunctbl[opcode] = f;
#if HAS_PROFILING
			perftbl[opcode] = perftbl[table68k[opcode].handler];
#endif

		}
	}	

	for (i = 0; tbl[i].handler != NULL; i++)
	{
		if (tbl[i].specific)
		{
			cpufunctbl[tbl[i].opcode] = tbl[i].handler;
#if HAS_PROFILING
			perftbl[tbl[i].opcode] = tbl[i].perf;
#endif
		}
	}

	// (hey readcpu doesn't free this guy!)

	Platform::DisposeMemory (table68k);
}
