/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Document.h"

#include "DebugMgr.h"			// Debug::HandleNewPacket
#include "EmCPU.h"				// Emulator::, others
#include "EmDlg.h"				// DoHordeNew
#include "EmMemory.h"			// Memory::
#include "EmScreen.h"			// EmScreen
#include "EmSession.h"			// EmSession
#include "EmFileImport.h"		// kMethodBest
#include "EmStreamFile.h"		// EmStreamFile, kCreateOrEraseForUpdate
#include "Hordes.h"				// Hordes::New...
#include "LoadApplication.h"	// SavePalmFile
#include "PreferenceMgr.h"		// Preference (kPrefKeyGremlinInfo)
#include "SocketMessaging.h"
#include "TrapPatches.h"		// Patches::fgUIInitialized

#include <limits.h>

// #define down fl_down			// "down" defined both in WindowNew.h and Fl/Enumerations.h
#include <FL/Fl.H>
#include <FL/filename.h>		// filename_setext

#include "fltk_Dialog.h"		// showNewGremlinWindow


// -----------------------------------------------------------------------------
// constructor(s) / destructor
// -----------------------------------------------------------------------------

Document::Document (void) :
	_session (NULL),
	_currentConfiguration (),
	_currentRAMRef (),
	_currentDirRef (),
	_handlers ()
{
}

Document::~Document (void)
{
	propertyChanged (documentClosing);

	delete _session;
	_session = NULL;

	Platform::GCW_Close ();
}


// -----------------------------------------------------------------------------
// public methods
// -----------------------------------------------------------------------------

ErrCode Document::create (const Configuration& cfg)
{
	// Try initializing the emulator with all this information.
	ErrCode	result = errNone;

	try
	{
		_session = new EmSession ();
		_session->CreateNew (cfg);
		_session->CreateThread (true);

		{
			// Save the newly created configuration for next time.

			Preference<Configuration>	pref (kPrefKeyLastConfiguration);
			pref = _session->GetConfiguration ();
		}

		{
			// Zap the last-saved session file, so that we give preference
			// to a New operation at startup and not an Open operation.

			Preference<EmFileRef>	pref (kPrefKeyLastPSF);
			pref = EmFileRef();
		}
	}
	catch (ErrCode e)
	{
		delete _session;
		_session = NULL;

		result = e;
	}

	if (result == errNone)
	{
		propertyChanged (documentOpening);
	}

	return result;
}


// -----------------------------------------------------------------------------
// read the document state from the given file, then get the CPU going with it.

ErrCode Document::open (const EmFileRef& ramFileRef)
{
	ErrCode	result = errNone;

	try
	{
		_session = new EmSession ();
		_session->CreateOld (ramFileRef);
		_session->CreateThread (true);

		Hordes::PostLoad();

		// If we loaded a session file with a Gremlin running, turn
		// it off, but let the user turn it back on.

		if (Hordes::IsOn ())
		{
//			Hordes::Stop ();
			Platform::GCW_Open ();
		}

		{
			// Save the newly created configuration for next time.

			Preference<Configuration>	pref (kPrefKeyLastConfiguration);
			pref = _session->GetConfiguration ();
		}

		{
			// Save the last-opened session file for next time.

			Preference<EmFileRef>	pref (kPrefKeyLastPSF);
			pref = ramFileRef;
		}

		gEmuPrefs->UpdateRAMMRU (ramFileRef);
	}
	catch (ErrCode e)
	{
		delete _session;
		_session = NULL;

		result = e;
	}

	if (result == errNone)
	{
		_currentRAMRef = ramFileRef;
		propertyChanged (documentOpening);
	}

	return result;
}

// -----------------------------------------------------------------------------
// start the emulator running

void Document::emulatorStart (void)
{
	assert (_session);

	_session->ResumeThread ();
	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// stop the emulator

void Document::emulatorStop (void)
{
	assert (_session);

	_session->SuspendThread (kStopNow);
	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// reset the emulator

void Document::emulatorReset (void)
{
	assert (_session);

	_session->Reset (true);
	propertyChanged (cpuStatusChanged);
}

// -----------------------------------------------------------------------------
// Start hotsync in the emulator

void Document::hotsync (void)
{
	this->button (true, kElement_CradleButton);
	this->button (false, kElement_CradleButton);
}

// -----------------------------------------------------------------------------
// Pass a pen event to the emulator

void Document::pen (bool down, int x, int y)
{
	assert (_session);

	EmPenEvent event (EmPoint (x, y), down);
	_session->PostPenEvent (event);
}

// -----------------------------------------------------------------------------
// Pass a button event to the emulator

void Document::button (bool down, SkinElementType button)
{
	assert (_session);

	EmButtonEvent event (button, down);
	_session->PostButtonEvent (event);
}

// -----------------------------------------------------------------------------
// Pass a key event to the emulator

void Document::putkey (unsigned char key)
{
	assert (_session);

	EmKeyEvent event (key);
	_session->PostKeyEvent (event);
}

// -----------------------------------------------------------------------------
// load the given application file

void Document::loadApplication (const EmFileRef& appFile)
{
	EmFileRefList fileList;
	fileList.push_back (appFile);

	EmDlg::DoDatabaseImport (fileList, kMethodBest);
}

// -----------------------------------------------------------------------------
// export the application file

void Document::exportDatabase (const DatabaseInfo& db, const EmFileRef& exportFileRef)
{
	EmStreamFile		exportFile (exportFileRef, kCreateOrEraseForUpdate,
								kFileCreatorEmulator, kFileTypePalmDB);

	::SavePalmFile (exportFile, db.cardNo, db.dbName);
}

// -----------------------------------------------------------------------------
// save the memory image to current file

void Document::saveMemory (void)
{
	saveRAMImage (_currentRAMRef, false);
	propertyChanged ();
}

// -----------------------------------------------------------------------------
// save the memory image to current file

void Document::saveMemoryAs (EmFileRef& dest)
{
	saveRAMImage (dest, false);
	propertyChanged ();
}

// -----------------------------------------------------------------------------
// save the memory image to the given file ref

void Document::saveScreen (EmFileRef& dest)
{
	EmFileRef		result;
	string			prompt ("Save image as?");
	EmDirRef		defaultPath = this->getCurrentDirRef ();
	EmFileTypeList	filterList (1, kFileTypePicture);

	EmDlgItemID		item = EmDlg::DoPutFile (result, prompt, defaultPath, filterList, "");

	if (item == kDlgItemOK)
	{
		// Make sure the extension is right.

		string	fullPath = result.GetFullPath ();
		char*	fNameExt = (char*) malloc (fullPath.size () + 4);
		strcpy (fNameExt, fullPath.c_str ());
		filename_setext (fNameExt, ".ppm");

		FILE* f = fopen (fNameExt, "wb");
		if (f)
		{
			EmScreen::InvalidateAll ();

			EmScreenUpdateInfo info;
			EmScreen::GetBits (info);

			EmScreen::InvalidateAll ();

			info.fImage.ConvertToFormat (kPixMapFormat24RGB);

			// PPM format is:
			//
			// File type tag
			// Width
			// Height
			// Max component value
			// Width * Height pixels
			//
			// The first items in the file are all text and
			// seperated by whitespace.  The array of pixels
			// is in text format if the file type is P3, and
			// is in binary format if the type is P6.

			EmPoint size = info.fImage.GetSize ();

			fprintf (f, "P6 %ld %ld 255\x0D", size.fX, size.fY);

			uint8* bits = (uint8*) info.fImage.GetBits ();
			for (long yy = 0; yy < size.fY; ++yy)
			{
				long rowBytes = info.fImage.GetRowBytes ();
				uint8* basePtr = bits + yy * rowBytes;

				fwrite (basePtr, 1, rowBytes, f);
			}

			fclose (f);
		}

		this->setCurrentDirRef (result.GetParent ());
		free (fNameExt);
	}
}

// -----------------------------------------------------------------------------
// add the given callback to the list of PropertyChangedHandlers

void Document::addPropertyChangedHandler (PropertyChangedHandler callback, void* param)
{
	PropertyChangedHandlerSpec	spec;
	spec.func = callback;
	spec.param = param;

	_handlers.push_back (spec);
}

// -----------------------------------------------------------------------------
// remove the given callback from the list of PropertyChangedHandlers

void Document::removePropertyChangedHandler (PropertyChangedHandler callback)
{
	list<PropertyChangedHandlerSpec>::iterator	iter;

	for (iter = _handlers.begin (); iter != _handlers.end (); ++iter)
	{
		if (iter->func == callback)
		{
			_handlers.erase (iter);
			return;
		}
	}
}

// -----------------------------------------------------------------------------
// gremlins control functions

void Document::gremlinsNew (void)
{
	EmDlg::DoHordeNew ();

	propertyChanged (unspecifiedChanged);
}

void Document::gremlinsStep (void)
{
	Hordes::Step ();
}

void Document::gremlinsResume (void)
{
	Hordes::Resume ();

	propertyChanged (unspecifiedChanged);
}

void Document::gremlinsStop()
{
	Hordes::Stop ();

	propertyChanged (unspecifiedChanged);
}

// -----------------------------------------------------------------------------
// private methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// save the ram image to the file

void Document::saveRAMImage (const EmFileRef& ramFileRef, Bool forAutoSave)
{
	// See if there's a given name we should use.

	if (!ramFileRef.IsSpecified())
	{
		return;
	}

	try
	{
		assert (_session);
		_session->Save (ramFileRef, true);
	}
	catch (...)
	{
	}

	_currentRAMRef = ramFileRef;
}

// -----------------------------------------------------------------------------
// broadcast changed notices to all propertyChangedHandlers

void Document::propertyChanged (int property)
{
	list<PropertyChangedHandlerSpec>::iterator	iter;

	for (iter = _handlers.begin (); iter != _handlers.end (); ++iter)
	{
		(iter->func) (property, iter->param);
	}
}
