/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "fltk_LcdWindow.h"

#include "EmHAL.h"				// GetLEDState
#include "EmJPEGUnix.h"			// JPEGToBits
#include "EmPixMapUnix.h"		// ConvertPixMapToHost
#include "EmScreen.h"			// EmScreen
#include "EmSession.h"			// EmSessionStopper
#include "EmStream.h"			// EmStream
#include "Platform.h"			// CommonDialog
#include "PreferenceMgr.h"
#include "Skins.h"				// SkinElementType, SkinTestPoint
#include "SocketMessaging.h"	// CSocket::IdleAll

#include "fltk_MainWindow.h"

#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Image.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Window.H>
#include <FL/math.h>

#include "DefaultLarge.xpm"
#include "DefaultSmall.xpm"


const float REDRAW_FREQ = .1;	// = 10x / second

Bool gNeedWindowReset;
Bool gNeedWindowInvalidate;

static void			PrvGetLCDBounds	(EmRect& result);
static void			PrvGetLEDBounds (EmRect& result);
static void			PrvPaintLED (uint16 ledState);


// Add some more constants, naming the same way FLTK
// names some key constants.  These values may be
// specific to certain operating systems; I don't know.
// They were determined empirically by pressing keys
// and seeing what showed up in handle().

const int FL_F1			= 65470;
const int FL_F2			= 65471;
const int FL_F3			= 65472;
const int FL_F4			= 65473;
const int FL_F9			= 65478;
const int FL_F10		= 65479;
const int FL_PageUp		= 65365;
const int FL_PageDown	= 65366;
const int FL_LeftArrow	= 65361;
const int FL_RightArrow	= 65363;
const int FL_UpArrow	= 65362;
const int FL_DownArrow	= 65364;


// -----------------------------------------------------------------------------
// file static methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Fl_Image helper class
// -----------------------------------------------------------------------------

// An Fl_Image subclass which owns dynamicly allocated image data

class Dynamic_Image : public Fl_Image
{
public:
	Dynamic_Image(const void* data, int width, int height, int rowBytes) :
		Fl_Image ((const uchar*) data, width, height, 3, rowBytes),
		_data (data),
		_width (width),
		_height (height)
	{
	}

	virtual ~Dynamic_Image(void)
	{
		Platform::DisposeMemory (_data);
	}

	 void measure (int& w, int& h);

private:
	const void* _data;
	int _width;
	int _height;
};

void Dynamic_Image::measure (int& w, int& h)
{
	w = _width;
	h = _height;
}


// -----------------------------------------------------------------------------
// CaseImage helper class
// -----------------------------------------------------------------------------

// Another helper class.  When using external skins, we manage them with the
// Fl_Image class.  When using the internal builtin default skin, we manage
// it with the Fl_Pixmap class.  These two classes don't have a common
// ancestor, so we can't reference both of them with a single base pointer.
// Instead, create this helper class that can own one or the other and
// forward messages to the appropriate one.

class CaseImage
{
public:
	CaseImage (Dynamic_Image*);
	CaseImage (Fl_Pixmap*);
	~CaseImage (void);

	void draw (int X, int Y, int W, int H, int cx=0, int cy=0);
	void draw (int X, int Y);
	void measure (int& w, int& h);

private:
	Dynamic_Image* fImage;
	Fl_Pixmap* fPixmap;
};

CaseImage::CaseImage (Dynamic_Image* image) :
	fImage (image),
	fPixmap (NULL)
{
}

CaseImage::CaseImage (Fl_Pixmap* pixmap) :
	fImage (NULL),
	fPixmap (pixmap)
{
}

CaseImage::~CaseImage (void)
{
	delete fImage;
	delete fPixmap;
}

void CaseImage::draw (int X, int Y, int W, int H, int cx, int cy)
{
	if (fImage)
		fImage->draw (X, Y, W, H, cx, cy);

	if (fPixmap)
		fPixmap->draw (X, Y, W, H, cx, cy);
}

void CaseImage::draw (int X, int Y)
{
	if (fImage)
		fImage->draw (X, Y);

	if (fPixmap)
		fPixmap->draw (X, Y);
}

void CaseImage::measure (int& W, int& H)
{
	if (fImage)
	{
		fImage->measure (W, H);
		return;
	}

	if (fPixmap)
	{
		sscanf (fPixmap->data[0], "%d %d", &W, &H);
		return;
	}

	W = H = 0;
}

// -----------------------------------------------------------------------------
// handle periodic tasks:
// redraw the LCD every REDRAW_FREQ seconds
// check for pending data on the debug socket

static void handlePeriodic (void* data)
{
	LCD_window* lcd = (LCD_window*) data;

	if (lcd != NULL)
	{
		Document* doc = lcd->getDocument();
		if (doc != NULL)
		{
			CSocket::IdleAll ();
		}

		if (gNeedWindowReset)
		{
			gNeedWindowReset = false;
			gNeedWindowInvalidate = false;
			lcd->ResetSkin();
		}

		if (gNeedWindowInvalidate)
		{
			gNeedWindowInvalidate = false;
			lcd->damageLcdArea ();
		}

		if (gSession)
		{
			lcd->damageLcdArea ();
		}

		if (gSession)
		{
			static uint16 oldLEDState = 0;
			uint16 ledState = EmHAL::GetLEDState ();

			if (oldLEDState != ledState)
			{
				oldLEDState = ledState;

				EmRect r;
				::PrvGetLEDBounds (r);

				lcd->damage (0, 
							 r.fLeft, r.fTop,
							 r.Width (), r.Height ());
			}

			// Do the Wiggle Walk.

			static Bool	wiggled = false;
			static Bool	wasOn = false;
			const int	kOffset = 1;
			Fl_Widget* p = lcd->parent ();

			if (EmHAL::GetVibrateOn ())
			{
				wasOn = true;

				if (!wiggled)
				{
					wiggled = true;
					p->position (p->x() + kOffset, p->y());
				}
				else
				{
					wiggled = false;
					p->position (p->x() - kOffset, p->y());
				}
			}

			// If the vibrator just went off, then put the window back to where it was.

			else if (wasOn)
			{
				wasOn = false;

				if (wiggled)
				{
					wiggled = false;
					p->position (p->x() - kOffset, p->y());
				}
			}

			// Show any error messages from the CPU thread.

			{
				string	title;
				string	message;
				int		buttonOptions;

				if (gSession->BeginDialog (title, message, buttonOptions))
				{
					int	result = Platform::CommonDialog (title.c_str (), message.c_str (), buttonOptions);
					gSession->EndDialog (result);
				}
			}
		}
	}

	Fl::add_timeout (REDRAW_FREQ, handlePeriodic, lcd);
}


/***********************************************************************
 *
 * FUNCTION:    PrvPrefsChanged
 *
 * DESCRIPTION: If any preference relevent to our display has changed,
 *				set a flag remembering it.  We'll respond to that flag
 *				later in some idle-time function.  We don't respond
 *				to it now in case several preferences change all at
 *				once; we don't want to respond to each one individually.
 *
 * PARAMETERS:  key - the key of the preference that changed.
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

static void PrvPrefsChanged (PrefKeyType key, void*)
{
	if ((strcmp (key, kPrefKeyScale) == 0) || (strcmp (key, kPrefKeySkins) == 0))
	{
		gNeedWindowReset = true;
	}
	else if (strcmp (key, kPrefKeyBackgroundColor) == 0)
	{
		gNeedWindowInvalidate = true;
	}
}


// -----------------------------------------------------------------------------
// constructor / destructor
// -----------------------------------------------------------------------------

LCD_window::LCD_window (MainWindow* parent) :
	Fl_Double_Window (0, 0, 600, 600, ""),
	_case (NULL),
	_penTracking (false),
	_currentButton (kElement_None),
	_parent (parent),
	_doc (NULL),
	_lcdOn (true)
{
	_message = new Fl_Box (0, 0, 200, 40, 
						   "Right click on this window to show a menu of commands.");
	_message->box (FL_NO_BOX);
	_message->align (FL_ALIGN_CENTER | FL_ALIGN_WRAP | FL_ALIGN_INSIDE);

	this->end ();

	this->resetToDefault();
	
	Fl::add_timeout (REDRAW_FREQ, handlePeriodic, this);

	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeySkins);
	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeyScale);
	gPrefs->AddNotification (PrvPrefsChanged, kPrefKeyBackgroundColor);
}

LCD_window::~LCD_window (void)
{
	// No longer need notification.

	gPrefs->RemoveNotification (PrvPrefsChanged);

	delete _case;
	_case = NULL;
}

// -----------------------------------------------------------------------------
// public methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Fl_Double_Window overrides
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Handle the GUI event. This routine translates the FLTK GUI events into
// the corresponding emulator actions.

int LCD_window::handle (int event)
{
	PointType			where = {Fl::event_x(), Fl::event_y()};
	SkinElementType		what = ::SkinTestPoint (where);

	switch (event)
	{
		case FL_PUSH:
		case FL_DRAG:
			if (Fl::event_button () == 3)
			{
				_parent->popupMenu ();
				return 1;
			}

			if (_doc != NULL)
			{
				if (_penTracking)
				{
					PointType	whereLCD = ::SkinWindowToTouchscreen (where);
					_doc->pen (true, whereLCD.x, whereLCD.y);
				}
				else if (_currentButton != kElement_None)
				{
					// Nothing to do...
				}
				else
				{
					if (what == kElement_Touchscreen)
					{
						PointType		whereLCD = ::SkinWindowToTouchscreen (where);
						_doc->pen (true, whereLCD.x, whereLCD.y);
						_penTracking = true;
					}
					else if (what != kElement_Frame)
					{
						_doc->button (true, what);
						_currentButton = what;
					}
				}
			}

			return 1;

		case FL_RELEASE:
			if (_doc != NULL)
			{
				if (_penTracking)
				{
					PointType	whereLCD = ::SkinWindowToTouchscreen (where);
					_doc->pen (false, whereLCD.x, whereLCD.y);
					_penTracking = false;
				}
				else if (_currentButton != kElement_None)
				{
					_doc->button (false, _currentButton);
					_currentButton = kElement_None;
				}
			}
			return 1;

		case FL_FOCUS:
			// always accept the focus
			return 1;

		case FL_KEYBOARD:
			if (Fl::event_state (FL_ALT|FL_META))
			{
				// reserved for shortcuts
				return 0;
			}

			if (_doc != NULL)
			{
				int c = Fl::event_key();

				// handle printable characters
				if (strlen (Fl::event_text()) > 0)
				{
					_doc->putkey (Fl::event_text()[0]);
					return 1;
				}

				// handle all other characters
				SkinElementType	whichButton = kElement_None;
				int				whichKey	= 0;
				switch (c)
				{
					case FL_Enter:
					case FL_KP_Enter:
						whichKey = '\n';
						break;

					case FL_F1:
						whichButton = kElement_App1Button;
						break;

					case FL_F2:
						whichButton = kElement_App2Button;
						break;

					case FL_F3:
						whichButton = kElement_App3Button;
						break;

					case FL_F4:
						whichButton = kElement_App4Button;
						break;

					case FL_F9:
						whichButton = kElement_PowerButton;
						break;

					case FL_F10:
						_parent->popupMenu();
						return 1;

					case FL_PageUp:
						whichButton=kElement_UpButton;
						break;

					case FL_PageDown:
						whichButton=kElement_DownButton;
						break;

					case FL_LeftArrow:
						whichKey = leftArrowChr;
						break;

					case FL_RightArrow:
						whichKey = rightArrowChr;
						break;

					case FL_UpArrow:
						whichKey = upArrowChr;
						break;

					case FL_DownArrow:
						whichKey = downArrowChr;
						break;

					default:
						return 0;
				}

				if (whichButton != kElement_None)
				{
					_doc->button (true, whichButton);
					_doc->button (false, whichButton);
				}

				if (whichKey != 0)
				{
					_doc->putkey (whichKey);
				}
			}

			return 1;

		default:
			return Fl_Double_Window::handle (event);
	}
}

// -----------------------------------------------------------------------------
// draw the case and the inner LCD screen
void LCD_window::draw()
{
	// redraw everything if anything other than just the lcd damage bit is on
	bool drawAll = (damage() & ~LCD_DAMAGE_BIT) != 0;
	drawAll = true;

	fl_color (FL_WHITE);
	if (drawAll && _case)
	{
		_case->draw (0, 0);
	}
	else
	{
		fl_rectf (0, 0, this->w (), this->h ());

		_message->damage (FL_DAMAGE_ALL);
		this->update_child (*_message);
	}

	// only draw the LCD if we have a document & the LCD is on
	if (_lcdOn && (_doc != NULL))
	{
		EmSessionStopper stopper (gSession, kStopNow);
		if (!stopper.Stopped())
			return;

		void* buffer = this->updateBuffer (drawAll);
		if (buffer != NULL)
		{
			EmRect lcdRect;
			::PrvGetLCDBounds (lcdRect);

			int x = lcdRect.fLeft;
			int y = lcdRect.fTop;
			int w = lcdRect.Width ();
			int h = lcdRect.Height ();

			fl_draw_image ((uchar*) buffer, x, y, w, h);

			Platform::DisposeMemory (buffer);
		}

		// Draw any LEDs.

		uint16 ledState = EmHAL::GetLEDState ();
		if (ledState)
			::PrvPaintLED (ledState);
	}
}

// -----------------------------------------------------------------------------
// turn the lcd off or on
void LCD_window::setOn( bool on )
{
	if (!_lcdOn && on)
	{
		// turn on lcd
		_lcdOn = true;
		Fl::add_timeout (REDRAW_FREQ, handlePeriodic, this);
		redraw();
	}
	else if (_lcdOn && !on)
	{
		// turn lcd off
		_lcdOn = false;

		redraw();
	}
}

// -----------------------------------------------------------------------------
// reset configurable parameters to default values.  called during construction.

void LCD_window::resetToDefault()
{
	_currentButton = kElement_None;
	_penTracking = false;

	this->ResetSkin ();
}

// -----------------------------------------------------------------------------
// load the graphics for the case.

void LCD_window::ResetSkin (void)
{
	if (!gSession)
	{
		this->size (_parent->w (), _parent->h ());

		_message->show ();
		_message->position (
			(this->w () - _message->w ()) / 2,
			(this->h () - _message->h ()) / 3);
		return;
	}

	_message->hide ();

	// Initialize the Skin engine with the preference settings.
	::SkinSetSkin ();

	delete _case;
	_case = NULL;

	EmStream*	stream = ::SkinGetSkinStream ();

	const void*	bmpData;
	int			width;
	int			height;
	int			rowBytes;

	if (stream != NULL)
	{
		::JPEGToBits (*stream, bmpData, width, height, rowBytes);
		_case = new CaseImage (new Dynamic_Image (bmpData, width, height, rowBytes));

		delete stream;
		stream = NULL;
	}
	else
	{
		Preference<ScaleType>		scalePref (kPrefKeyScale);
		char** xpm = *scalePref == 1 ? DefaultSmall : DefaultLarge;

		_case = new CaseImage (new Fl_Pixmap (xpm));
	}

	// resize this window to the size of the surrounding case

	int w, h;
	_case->measure (w, h);
	_parent->size (w, h);
	this->size (w, h);	
}


// -----------------------------------------------------------------------------
// invalidate the entire lcd area.

void LCD_window::damageLcdArea ()
{
	EmRect bounds;
	::PrvGetLCDBounds (bounds);
	this->damage (LCD_DAMAGE_BIT,
				  bounds.fLeft, bounds.fTop,
				  bounds.Width (), bounds.Height ());
}


// -----------------------------------------------------------------------------
// private methods
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// update the offscreen lcd buffer from the emulator's screen RAM. Returns true
// if the offscreen buffer data has been changed, false otherwise. Note that
// you must own the buffer lock to call this method.

void* LCD_window::updateBuffer (int always)
{
	// Get information about the screen.  Return immediately if nothing's changed.

	if (always)
		EmScreen::InvalidateAll ();

	EmScreenUpdateInfo	info;
	Bool	drawLCD = EmScreen::GetBits (info);

	if (!drawLCD || !info.fLCDOn)
		return NULL;

	// ======================================================================
	// Create a bitmap for blitting from.
	// ======================================================================

	EmPoint	before (1, 1);
	EmPoint	after = ::SkinScaleUp (before);

	EmPoint size = info.fImage.GetSize ();
	size = ::SkinScaleUp (size);

	void* buffer = Platform::AllocateMemory (size.fX * size.fY * 3);

	if (buffer)
	{
		::ConvertPixMapToHost (info.fImage, buffer,
							   info.fFirstLine, info.fLastLine,
							   before.fX != after.fX);
	}

	return buffer;
}


void	PrvGetLCDBounds	(EmRect& result)
{
	int				index = 0;
	SkinElementType	type;
	EmRect			bounds;

	while (::SkinGetElementInfo (index, type, bounds))
	{
		if (type == kElement_LCD)
		{
			result = bounds;
			return;
		}

		++index;
	}

	assert (false);
}


void PrvGetLEDBounds (EmRect& result)
{
	SkinElementType	type = kElement_None;
	int				ii = 0;
	while (::SkinGetElementInfo (ii++, type, result))
	{
		if (type == kElement_PowerButton)
			break;
	}

	if (type == kElement_PowerButton)
	{
		// Create a square centered in the power button.

		if (result.Width () > result.Height ())
		{
			result.fLeft = result.fLeft + (result.Width () - result.Height ()) / 2;
			result.fRight = result.fLeft + result.Height ();
		}
		else
		{
			result.fTop = result.fTop + (result.Height () - result.Width ()) / 2;
			result.fBottom = result.fTop + result.Width ();
		}

		// Inset it a little -- looks better this way.

		result.Inset (2, 2);
	}
	else
	{
		result = EmRect (10, 10, 20, 20);
	}
}


void PrvPaintLED (uint16 ledState)
{
	// Get the color for the LED.

	RGBType rgb;

	if ((ledState & (kLEDRed | kLEDGreen)) == (kLEDRed | kLEDGreen))
	{
		// Orange
		rgb = RGBType (255, 128, 0);
	}
	else if ((ledState & kLEDRed) == kLEDRed)
	{
		// Red
		rgb = RGBType (255, 0, 0);
	}
	else
	{
		// Green
		rgb = RGBType (0, 255, 0);
	}

	EmRect b;
	::PrvGetLEDBounds (b);

	fl_color (rgb.fRed, rgb.fGreen, rgb.fBlue);
	fl_pie (b.fLeft, b.fTop, b.Width (), b.Height (), 0, 360);
}
